package com.webssa.library.galleryviewpager;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.drawable.BitmapDrawable;
import android.graphics.drawable.Drawable;
import android.graphics.drawable.TransitionDrawable;
import android.net.Uri;
import android.os.Build;
import android.text.TextUtils;
import android.util.Log;
import android.view.Gravity;
import android.view.View;
import android.view.ViewGroup;
import android.widget.ImageView;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.viewpager.widget.PagerAdapter;

import com.webssa.library.mypicasso.LruCache;
import com.webssa.library.mypicasso.MemoryPolicy;
import com.webssa.library.mypicasso.Picasso;
import com.webssa.library.mypicasso.Target;
import com.webssa.library.utils.AsyncTaskCompat;
import com.webssa.library.utils.BitmapUtils;
import com.webssa.pda.sunmi.PDA_Application;
import com.webssa.pda.sunmi.GlobalProperties;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;


/**
 * Created by stanislav.perchenko on 7/4/2018
 */
public class GalleryPagerAdapter extends PagerAdapter {

    private static final boolean D = GlobalProperties.D;
    private static final int DEBUG_MIN_LOAD_TIME = 0;

    public static final String LOG_TAG = "G_P_A";

    public interface OnPageClickListener {
        void onClick(View v, int position);
    }


    private Drawable placeholderDrawable;
    private int pageBackgroundColor;
    private int cropGravity = Gravity.CENTER; // START / CENTER / END
    private int fadeDuration = 400;

    private final Context mContext;
    private final List<ImageReference> mData = new ArrayList<>();

    private final LruCache bmpMemCache;

    private OnPageClickListener onPageClickListener;


    public GalleryPagerAdapter(Context ctx, @NonNull Collection<ImageReference> data) {
        mContext = ctx;
        mData.addAll(data);
        bmpMemCache = PDA_Application.getPicassoMemCache();
    }

    public void setPlaceholderDrawable(Drawable dr) {
        placeholderDrawable = dr;
    }

    public void setPageBackgroundColor(int pageBackgroundColor) {
        this.pageBackgroundColor = pageBackgroundColor;
    }

    public void setOnPageClickListener(OnPageClickListener onPageClickListener) {
        this.onPageClickListener = onPageClickListener;
    }

    public void setCropGravity(int cropGravity) {
        if (this.cropGravity != cropGravity) {
            if (cropGravity != Gravity.START && cropGravity != Gravity.CENTER && cropGravity != Gravity.END) {
                throw new IllegalArgumentException("Crop Gravity must be one of START, CENTER, END");
            }
            this.cropGravity = cropGravity;
            notifyDataSetChanged();
        }
    }

    public void setFadeDuration(int fadeDuration) {
        this.fadeDuration = fadeDuration;
    }

    public void clear() {
        if (mData.size() > 0) {
            mData.clear();
            notifyDataSetChanged();
        }
    }

    public boolean isSameContent(@NonNull Collection<? extends ImageReference> nItems) {
        if (mData.size() != nItems.size()) return false;

        for (Iterator<? extends ImageReference> itr1 = mData.iterator(), itr2 = nItems.iterator(); itr1.hasNext() & itr2.hasNext(); ) {
            ImageReference item1 = itr1.next();
            ImageReference item2 = itr2.next();
            String p1 = (item1 == null) ? null : item1.getImagePath();
            String p2 = (item2 == null) ? null : item2.getImagePath();
            if (!TextUtils.equals(p1, p2)) return false;
        }
        return true;
    }

    public void setData(@NonNull Collection<? extends ImageReference> data) {
        mData.clear();
        mData.addAll(data);
        notifyDataSetChanged();
    }

    public Collection<ImageReference> getData() {
        Collection<ImageReference> ret = new ArrayList<>(mData.size());
        ret.addAll(mData);
        return ret;
    }

    public void addDataItem(ImageReference dataItem) {
        mData.add(dataItem);
        notifyDataSetChanged();
    }

    public ImageReference removeItem(int position) {
        if (position < 0 || position >= mData.size()){
            throw new IllegalArgumentException("Wrong position - "+position+". Size="+mData.size());
        }
        try {
            return mData.remove(position);
        } finally {
            notifyDataSetChanged();
        }
    }

    @Override
    public int getCount() {
        return mData.size();
    }

    @Override
    public int getItemPosition(Object object) {
        // For the notifyDataSetChanged() to cause immediate ViewPager update.
        return POSITION_NONE;
    }

    @Override
    public boolean isViewFromObject(View view, Object object) {
        return view.getTag() == object;
    }

    private final Map<Integer, AdapterObject> mapAdapterObjects = new HashMap<>();
    private final Set<AdapterObject> attachedAdapterObjects = new HashSet<>();
    private final Queue<AdapterObject> recycledAdapterObjects = new LinkedList<>();
    private Object primaryAdapterObject;
    private boolean hasContainerSize;
    private int lastContainerW, lastContainerH;

    @Override
    public AdapterObject instantiateItem(ViewGroup container, int position) {
        if (D) Log.d(LOG_TAG, "===>  Instantiate item - "+position);

        final int newContW = container.getMeasuredWidth();
        final int newContH = container.getMeasuredHeight();
        if (newContW == 0 || newContH == 0) {
            throw new IllegalStateException("Try to instantiate adapter item for the non-measured host container");
        } else if (!hasContainerSize || newContW != lastContainerW || newContH  != lastContainerH) {
            lastContainerW = newContW;
            lastContainerH = newContH;
            hasContainerSize = true;
        }

        AdapterObject itemObj = mapAdapterObjects.get(position);
        if (itemObj == null) {
            if (recycledAdapterObjects.size() > 0) {
                if (D) Log.d(LOG_TAG, String.format("Got item for position %d from SCRAP", position));
                itemObj = recycledAdapterObjects.remove();
            } else {
                if (D) Log.d(LOG_TAG, String.format("Build new item for position %d", position));
                itemObj = new AdapterObject(new ImageView(mContext));
                itemObj.imgView.setScaleType(ImageView.ScaleType.CENTER_INSIDE);
                itemObj.imgView.setBackgroundColor(pageBackgroundColor);
                itemObj.imgView.setLayoutParams(new ViewGroup.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.MATCH_PARENT));
            }
            mapAdapterObjects.put(position, itemObj);
        } else {
            if (D) Log.d(LOG_TAG, String.format("Got item for position %d from HEAP", position));
        }

        itemObj.imgView.setOnClickListener(v -> {
            if (onPageClickListener != null) onPageClickListener.onClick(v, position);
        });
        container.addView(itemObj.imgView);
        attachedAdapterObjects.add(itemObj);


        String itemPath = mData.get(position).getImagePath();
        itemObj.setImagePath(itemPath);
        itemObj.setContainerSize(container.getMeasuredWidth(), container.getMeasuredHeight());
        itemObj.setCropGravity(cropGravity);
        if (itemObj.isDirty()) {
            loadItemObject(position, itemObj);
        }

        return itemObj;
    }

    @Override
    public void destroyItem(ViewGroup container, int position, Object object) {
        if (D) Log.e(LOG_TAG, "<~~~~  Destroy item - "+position);

        AdapterObject itemObj = (AdapterObject) object;

        container.removeView(itemObj.imgView);
        itemObj.imgView.setOnClickListener(null);
        attachedAdapterObjects.remove(itemObj);
        boolean isHere = false;
        for (ImageReference ir : mData) {
            if (ir.getImagePath().equals(itemObj.imgPath)) {
                isHere = true; break;
            }
        }
        if (!isHere) {
            itemObj = mapAdapterObjects.remove(position);
            itemObj.setImagePath(null);
            recycledAdapterObjects.add(itemObj);
        }
    }

    @Override
    public void setPrimaryItem(ViewGroup container, int position, Object object) {
        primaryAdapterObject = object;
    }



    /**********************************************************************************************/
    private class AdapterObject {
        String imgPath;
        final ImageView imgView;
        int containerW, containerH;
        int cropGravity;

        boolean dirty;

        AdapterObject(ImageView imgView) {
            this.imgView = imgView;
            imgView.setTag(this);
        }

        public boolean isDirty() {
            return dirty;
        }

        void setImagePath(String path) {
            if (!TextUtils.equals(path, imgPath)) {
                imgPath = path;
                imgView.setImageBitmap(null);

                dirty = true;
            }
        }

        void setContainerSize(int w, int h) {
            if (containerW != w) {
                containerW = w;
                dirty = true;
            }
            if (containerH != h) {
                containerH = h;
                dirty = true;
            }
        }

        void setCropGravity(int cropGravity) {
            if (this.cropGravity != cropGravity) {
                this.cropGravity =  cropGravity;
                dirty = true;
            }
        }

        void setLoadedImageDrawable(Drawable dr) {
            imgView.setImageDrawable(dr);
            dirty = false;
        }

        String getHash() {
            return String.format("%dx%dx%d:%s", containerW, containerH, cropGravity, imgPath);
        }
    }


    /**********************************************************************************************/


    private void loadItemObject(final int position, AdapterObject io) {
        Bitmap bmp = bmpMemCache.get(io.getHash());
        if (bmp != null) {
            if (D) Log.d(LOG_TAG, String.format("    <---- Got image for position %d from CACHE", position));
            io.imgView.setImageBitmap(bmp);
        } else if (!inProgressLoadingTasks.containsKey(io.getHash())) {
            if (D) Log.d(LOG_TAG, "    ----> Start loading task for position - "+position);
            io.imgView.setImageDrawable(placeholderDrawable);
            (new ImageLoadingTask(io, bmpMemCache) {
                @Override
                public void onImageLoaded(String itemHash, @Nullable Bitmap bmpStart, @NonNull Bitmap bmpEnd) {

                    int nUpdated = 0;
                    if (bmpStart == null) {

                        //----  Set result bitmap with no transition  ----
                        for (AdapterObject itemObj : mapAdapterObjects.values()) {
                            if (itemObj.getHash().equals(itemHash)) {
                                itemObj.setLoadedImageDrawable(buildDrawableFromBitmap(bmpEnd));
                                nUpdated ++;
                            }
                        }

                    } else {

                        //---- Build result transition  ----
                        for (AdapterObject itemObj : mapAdapterObjects.values()) {
                            if (itemObj.getHash().equals(itemHash)) {

                                if (attachedAdapterObjects.contains(itemObj)) {

                                    TransitionDrawable drResult = new TransitionDrawable(new Drawable[]{
                                            buildDrawableFromBitmap(bmpStart),
                                            buildDrawableFromBitmap(bmpEnd)
                                    });

                                    //----  Set result and start animation  ----
                                    itemObj.setLoadedImageDrawable(drResult);
                                    drResult.startTransition((itemObj == primaryAdapterObject) ? fadeDuration : (fadeDuration / 2));
                                } else {
                                    itemObj.setLoadedImageDrawable(buildDrawableFromBitmap(bmpEnd));
                                }
                                nUpdated ++;
                            }
                        }

                    }

                    if (D) Log.d(LOG_TAG, String.format("    <---- Got image for position %d from PICASSO. %d pages were updated", position, nUpdated));
                }
            }).safeExecute();
        } else {
            io.imgView.setImageDrawable(placeholderDrawable);
            if (D) Log.d(LOG_TAG, String.format("    <---- Skip loading for position %d (proper loading task is in progress)", position));
        }
    }

    private BitmapDrawable buildDrawableFromBitmap(Bitmap bm) {
        return (Build.VERSION.SDK_INT >= 16) ? new BitmapDrawable(mContext.getResources(), bm) : new BitmapDrawable(bm);
    }


    private final Map<String, ImageLoadingTask> inProgressLoadingTasks = new HashMap<>();

    private abstract class ImageLoadingTask extends AsyncTaskCompat<Void, Void, Boolean> {

        private final AdapterObject adapterObj;
        private final LruCache memCache;

        final Object oLock = new Object();
        private boolean isLoading = true;
        private Bitmap loadedBmp;
        private Picasso.LoadedFrom loadedFrom;

        private Bitmap bmpPlaceholder;
        private Canvas placeholderCanv;

        public ImageLoadingTask(AdapterObject adObj, LruCache memCache) {
            this.adapterObj = adObj;
            this.memCache = memCache;
        }

        private Target loadingTarget = new Target() {
            @Override
            public void onBitmapLoaded(Bitmap bitmap, Picasso.LoadedFrom from) {
                System.out.println(from);
                synchronized (oLock) {
                    loadedBmp = bitmap;
                    loadedFrom= from;
                    isLoading = false;
                    oLock.notifyAll();
                }
            }

            @Override
            public void onBitmapFailed(Drawable errorDrawable) {
                System.out.println(errorDrawable);
            }

            @Override
            public void onPrepareLoad(Drawable placeHolderDrawable) { }
        };

        @Override
        protected void onPreExecute() {
            inProgressLoadingTasks.put(adapterObj.getHash(), this);

            //----  Start loading image  ----
            // Loading must be started from the UI Thread !!!
            Uri u = (adapterObj.imgPath.startsWith("/"))
                    ? Uri.fromFile(new File(adapterObj.imgPath))
                    : Uri.parse(adapterObj.imgPath);
            PDA_Application.getPicasso().load(u).memoryPolicy(MemoryPolicy.NO_CACHE, MemoryPolicy.NO_STORE).into(loadingTarget);
        }

        @Override
        protected Boolean doInBackground(Void... voids) {

            //----  Wait for load completion ----
            synchronized (oLock) {
                long tStart = System.currentTimeMillis();
                if (isLoading) {
                    do {
                        try {
                            oLock.wait(65);
                        } catch (InterruptedException e) {
                            // Probably task was cancelled
                            return false;
                        }
                    } while (isLoading && (System.currentTimeMillis() - tStart) < 15000);
                }


                //----  Debug protection delay  ----
                if (D) {
                    int dt = (int) (System.currentTimeMillis() - tStart);
                    if (dt < DEBUG_MIN_LOAD_TIME) {
                        try {
                            Thread.sleep(DEBUG_MIN_LOAD_TIME - dt);
                        } catch (InterruptedException e) {
                        }
                    }
                }

            }
            loadingTarget = null;

            if (loadedBmp == null) return false;


            //----  Resize result  ----



            if (loadedBmp.isRecycled()) return false;
            try {
                int nTry = 0;
                do {
                    try {
                        Bitmap scaledBmp = BitmapUtils.scaleBitmapAndCrop(loadedBmp, adapterObj.containerW, adapterObj.containerH, cropGravity);
                        if (scaledBmp != loadedBmp) {
                            loadedBmp.recycle();
                            loadedBmp = scaledBmp;
                        }
                        break;
                    } catch (OutOfMemoryError err) {
                        Thread.sleep(70);
                    }
                } while (++nTry < 10);


            } catch (Exception e) {
            }

            try {
                bmpPlaceholder = createBitmap(adapterObj.containerW, adapterObj.containerH, 15);
                placeholderCanv = new Canvas(bmpPlaceholder);
                return true;
            } catch (InterruptedException e) {
                return false;
            }
        }


        private Bitmap createBitmap(int w, int h, int nTries) throws InterruptedException {
            do {
                try {
                    return Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
                } catch (OutOfMemoryError err) {
                    Thread.sleep(70);
                }
            } while (--nTries > 0);
            throw new InterruptedException();
        }


        public abstract void onImageLoaded(String itemHash, @Nullable Bitmap bmpStart, @NonNull Bitmap bmpEnd);


        @Override
        protected void onPostExecute(Boolean result) {
            try {
                if (!result) return;
                String hash = adapterObj.getHash();

                //----  Cache result  ----
                memCache.set(hash, loadedBmp);

                if (loadedFrom == Picasso.LoadedFrom.MEMORY) {
                    onImageLoaded(hash, null, loadedBmp);
                } else {

                    //----  Extract full-size placeholder  ----
                    adapterObj.imgView.draw(placeholderCanv);

                    onImageLoaded(hash, bmpPlaceholder, loadedBmp);
                }
            } finally {
                onCancelled(result);
            }
        }



        @Override
        protected void onCancelled(Boolean result) {
            onCancelled();
        }

        @Override
        protected void onCancelled() {
            inProgressLoadingTasks.remove(adapterObj.getHash());
        }
    }

}
