package com.webssa.library.utils;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.BitmapShader;
import android.graphics.Canvas;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Shader;
import android.media.ExifInterface;
import android.util.Base64;
import android.view.Gravity;


import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Created by stanislav.perchenko on 8/22/2018
 */
public class BitmapUtils {

    /**
     * This loads required file into provided destination file calculating its hash in the process
     * It tries to calculate ShA-1 hash by default or MD5 in case of failure of the first one.
     * @param imageUrl
     * @param dstFile
     * @param maxSizeBytes maximum size allowed for download. Throws IOException is file exceeds this limit
     * @return base64-encode hash of the downloaded file
     */
    public static String loadImageToFileWithHash(@NonNull URL imageUrl, @NonNull File dstFile, int maxSizeBytes) throws IOException {
        MessageDigest digester = null;
        try {
            digester = MessageDigest.getInstance("SHA-1");
        } catch (NoSuchAlgorithmException e) {
            try {
                digester = MessageDigest.getInstance("MD5");
            } catch (NoSuchAlgorithmException e1) {
                throw new IOException("No algorithms for hash calculation: SHA-1, MD5");
            }
        }

        HttpURLConnection conn = (HttpURLConnection) imageUrl.openConnection();
        conn.setRequestMethod("GET");
        conn.setDoOutput(false);
        conn.connect();
        try {
            if (conn.getResponseCode() == HttpURLConnection.HTTP_OK) {
                int cLen = conn.getContentLength();
                if (cLen > maxSizeBytes) {
                    throw new IOException(String.format("Image content to large - %d bytes. Available max size - %d bites", cLen, maxSizeBytes));
                } else if (cLen == 0) {
                    throw new IOException("Content-Length is 0");
                } else {
                    try (OutputStream os = new BufferedOutputStream(new FileOutputStream(dstFile))) {
                        InputStream is = conn.getInputStream();
                        byte[] buff = new byte[1024];
                        int nRead, totalRead = 0;
                        while((nRead = is.read(buff)) > 0) {
                            totalRead += nRead;
                            if (totalRead > maxSizeBytes) throw new IOException("Unexpected too long content - "+totalRead);
                            digester.update(buff, 0, nRead);
                            os.write(buff, 0, nRead);
                        }
                        os.flush();
                    }
                    return Base64.encodeToString(digester.digest(), Base64.NO_WRAP | Base64.NO_PADDING | Base64.URL_SAFE);
                }
            } else {
                throw new IOException("Error HTTP response code - "+conn.getResponseCode());
            }
        } finally {
            conn.disconnect();
        }
    }

    @NonNull
    public static Bitmap loadBitmapFromNetwork(@Nullable URL url, int maxSizeBytes) throws IOException {
        if (url == null) throw new IOException("URL is null");
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setRequestMethod("GET");
        conn.setDoOutput(false);
        conn.connect();
        try {
            if (conn.getResponseCode() == HttpURLConnection.HTTP_OK) {
                int cLen = conn.getContentLength();
                if (cLen > maxSizeBytes) {
                    throw new IOException(String.format("Image content to large - %d bytes. Available max size - %d bites", cLen, maxSizeBytes));
                } else if (cLen == 0) {
                    throw new IOException("Content-Length is 0");
                } else {


                    InputStream is = conn.getInputStream();
                    byte[] buff = new byte[512];
                    ByteArrayOutputStream bos = new ByteArrayOutputStream(cLen + 1500);
                    int nRead;
                    while((nRead = is.read(buff)) > 0) bos.write(buff, 0, nRead);
                    byte[] data = bos.toByteArray();
                    is.close();
                    Bitmap bmp = BitmapFactory.decodeByteArray(data, 0, data.length);
                    if (bmp != null) {
                        return bmp;
                    } else {
                        throw new IOException("Loaded content is not an image");
                    }
                }
            } else {
                throw new IOException("Error HTTP response code - "+conn.getResponseCode());
            }
        } finally {
            conn.disconnect();
        }
    }


    public static Bitmap scaleBitmapAndCrop(@NonNull Bitmap loadedBmp, int containerW, int containerH, int cropGravity) {
        final float K = (float) containerW / containerH;
        float k = (float) loadedBmp.getWidth() / loadedBmp.getHeight();
        float resultScale = (K >= k) ? ((float) containerW / loadedBmp.getWidth()) : ((float) containerH / loadedBmp.getHeight());

        if (Math.abs(resultScale - 1f) < 0.05) return loadedBmp;

        int dstWidth = Math.round(resultScale * loadedBmp.getWidth());
        int dstHeight= Math.round(resultScale * loadedBmp.getHeight());
        Bitmap scaledBmp = Bitmap.createScaledBitmap(loadedBmp, dstWidth, dstHeight, true);

        if (scaledBmp != loadedBmp) {
            loadedBmp.recycle();

            //----  Crop result if necessary  ----
            if (K > (k + 0.02f)) {
                int dh = 0;
                if (cropGravity == Gravity.CENTER) {
                    dh = (scaledBmp.getHeight() - containerH) / 2;
                } else if (cropGravity == Gravity.END) {
                    dh = scaledBmp.getHeight() - containerH;
                }
                loadedBmp = Bitmap.createBitmap(scaledBmp, 0, dh, scaledBmp.getWidth(), containerH);
                scaledBmp.recycle();
            } else if (K < (k - 0.02)) {

                int dw = 0;
                if (cropGravity == Gravity.CENTER) {
                    dw = (scaledBmp.getWidth() - containerW) / 2;
                } else if (cropGravity == Gravity.END) {
                    dw = scaledBmp.getWidth() - containerW;
                }
                loadedBmp = Bitmap.createBitmap(scaledBmp, dw, 0, containerW, scaledBmp.getHeight());
                scaledBmp.recycle();
            } else {
                loadedBmp = scaledBmp;
            }
        }
        return loadedBmp;

    }

    public static Bitmap circleMask(Bitmap src, float radius) {
        Paint paint = new Paint();
        paint.setAntiAlias(true);
        paint.setShader(new BitmapShader(src, Shader.TileMode.CLAMP, Shader.TileMode.CLAMP));

        final int width = src.getWidth(), height = src.getHeight();
        Bitmap bitmap = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888);
        Canvas canvas = new Canvas(bitmap);
        canvas.drawCircle(width / 2f, height / 2f, radius, paint);
        return bitmap;
    }

    public static Size getImageSize(File fImage) {
        BitmapFactory.Options opts = new BitmapFactory.Options();
        opts.inJustDecodeBounds = true;
        BitmapFactory.decodeFile(fImage.getAbsolutePath(), opts);
        return new Size(opts.outWidth, opts.outHeight);
    }

    public static int getImageRotationExifConstant(String filename) throws IOException {
        String orientation = (new ExifInterface(filename)).getAttribute(ExifInterface.TAG_ORIENTATION);
        try {
            int rot = Integer.parseInt(orientation, 16);
            switch (rot) {
                case ExifInterface.ORIENTATION_UNDEFINED:
                    return ExifInterface.ORIENTATION_UNDEFINED;
                case ExifInterface.ORIENTATION_NORMAL:
                    return ExifInterface.ORIENTATION_NORMAL;
                case ExifInterface.ORIENTATION_FLIP_HORIZONTAL:
                    return ExifInterface.ORIENTATION_FLIP_HORIZONTAL;
                case ExifInterface.ORIENTATION_ROTATE_180:
                    return ExifInterface.ORIENTATION_ROTATE_180;
                case ExifInterface.ORIENTATION_FLIP_VERTICAL:
                    return ExifInterface.ORIENTATION_FLIP_VERTICAL;
                case ExifInterface.ORIENTATION_TRANSPOSE:
                    return ExifInterface.ORIENTATION_TRANSPOSE;
                case ExifInterface.ORIENTATION_ROTATE_90:
                    return ExifInterface.ORIENTATION_ROTATE_90;
                case ExifInterface.ORIENTATION_TRANSVERSE:
                    return ExifInterface.ORIENTATION_TRANSVERSE;
                case ExifInterface.ORIENTATION_ROTATE_270:
                    return ExifInterface.ORIENTATION_ROTATE_270;
                default:
                    throw new IOException("Orientation constant not supported - "+rot);
            }
        } catch (NumberFormatException e) {
            throw new IOException("Cannot parse Exif orientation - "+orientation);
        }
    }

    public static int getImageRotationDegrees(String filename) {
        try {
            switch (getImageRotationExifConstant(filename)) {
                case ExifInterface.ORIENTATION_ROTATE_90:
                    return 90;
                case ExifInterface.ORIENTATION_ROTATE_180:
                    return 180;
                case ExifInterface.ORIENTATION_ROTATE_270:
                    return 270;
                default:
                    return 0;
            }
        } catch (IOException e) {
            return 0;
        }
    }

    /**
     *
     * @param bmp original Bitmap
     * @param exifOrientation orientation constant as ExifInterface.ORIENTATION_?
     * @return
     */
    public static Bitmap fixBitmapOrientation(@NonNull Bitmap bmp, int exifOrientation) {
        int w = bmp.getWidth();
        int h = bmp.getHeight();
        if (w == 0 || h == 0) return bmp;
        Matrix mtx = new Matrix();
        final float px = w / 2f;
        final float py = h / 2f;
        switch (exifOrientation) {
            case ExifInterface.ORIENTATION_ROTATE_90:
                mtx.setRotate(90, px, py);
                break;
            case ExifInterface.ORIENTATION_ROTATE_180:
                mtx.setRotate(180, px, py);
                break;
            case ExifInterface.ORIENTATION_ROTATE_270:
                mtx.setRotate(270, px, py);
                break;
            case ExifInterface.ORIENTATION_FLIP_HORIZONTAL:
                mtx.setScale(-1, 1, px, py);
                break;
            case ExifInterface.ORIENTATION_FLIP_VERTICAL:
                mtx.setScale(1, -1, px, py);
                break;
            case ExifInterface.ORIENTATION_TRANSPOSE:
                mtx.postRotate(45, px, py);
                mtx.postScale(-1, 1, px, py);
                mtx.postRotate(-45, px, py);
                break;
            case ExifInterface.ORIENTATION_TRANSVERSE:
                mtx.postRotate(45, px, py);
                mtx.postScale(1, -1, px, py);
                mtx.postRotate(-45, px, py);
                break;
            default:
                return bmp;
        }
        return Bitmap.createBitmap(bmp, 0, 0, w, h, mtx, true);
    }



    public static BitmapFactory.Options decodeBitmapForOptions(String filePath) {
        BitmapFactory.Options opts = new BitmapFactory.Options();
        opts.inJustDecodeBounds = true;
        BitmapFactory.decodeFile(filePath, opts);
        return opts;
    }



    public static Bitmap scaleBitmapPreservingAspect(Bitmap inBmp, float scale) {
        Matrix mtrx = new Matrix();
        mtrx.postScale(scale, scale);
        return Bitmap.createBitmap(inBmp, 0, 0, inBmp.getWidth(), inBmp.getHeight(), mtrx, false);
    }

    public static String decodeImageFileForMimeType(String path) throws IOException {
        BitmapFactory.Options options = new BitmapFactory.Options();
        options.inJustDecodeBounds = true;
        BitmapFactory.decodeFile(path, options);
        if (options.outWidth==0 || options.outHeight==0 || options.outMimeType == null) {
            throw new IOException("Not an image");
        }
        return options.outMimeType;
    }

    public static Bitmap decodeBitmapMaxRezSafely(String path) {
        BitmapFactory.Options options = new BitmapFactory.Options();
        options.inSampleSize = 1;
        do {
            try {
                return BitmapFactory.decodeFile(path, options);
            } catch (OutOfMemoryError e) {
                e.printStackTrace();
                options.inSampleSize *= 2;
            }
        } while (options.inSampleSize < 17);
        return null;
    }

    public static Bitmap decodeBitmapMaxRezSafely(InputStream is) {
        is.mark(0);
        BitmapFactory.Options options = new BitmapFactory.Options();
        options.inSampleSize = 1;
        do {
            try {
                return BitmapFactory.decodeStream(is, null, options);
            } catch (OutOfMemoryError e) {
                options.inSampleSize *= 2;
                try {
                    is.reset();
                } catch (IOException e1) {
                    //This happens if the <is> is ton an ByteArrayInputStream.
                    return null;
                }
            }
        } while (options.inSampleSize < 17);
        return null;
    }


}
