package com.webssa.pda.sunmi.core;

import android.content.Context;
import android.content.SharedPreferences;
import android.net.Uri;
import android.text.TextUtils;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.squareup.moshi.JsonAdapter;
import com.squareup.moshi.Moshi;
import com.webssa.pda.sunmi.PDA_Application;
import com.webssa.pda.sunmi.GlobalProperties;
import com.webssa.pda.sunmi.rest.ApplicationRestService;
import com.webssa.pda.sunmi.rest.helper.ParserProvider;
import com.webssa.pda.sunmi.rest.model.LoginAnswer;
import com.webssa.pda.sunmi.rest.model.UserEntity;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import androidx.lifecycle.LiveData;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;

public final class SessionHolder {

    public static final boolean D = GlobalProperties.D;
    public static final String LOG_TAG = "SESSION_HOLDER";

    public interface OnSessionUpdateListener {
        void onSessionUpdated(Session s);
    }


    /**********************************************************************************************/
    /**********************************************************************************************/
    private static SessionHolder INSTANCE;

    public static SessionHolder getInstance() {
        if (INSTANCE == null) {
            synchronized (SessionHolder.class) {
                if (INSTANCE == null) {
                    INSTANCE = new SessionHolder();
                }
            }
        }
        return INSTANCE;
    }



    private SessionHolder() { }

    @Nullable
    private Session session;
    private UserEntity prevLoggedInUser;

    /**********************************************************************************************/
    /*****************************  Public Session static  API  ***********************************/
    /**********************************************************************************************/
    public static synchronized boolean isLoggedIn() {
        return getInstance().isLoggedIn_();
    }

    public static synchronized boolean isLoginExpired() {
        return getInstance().isLoginExpired_();
    }

    public static synchronized Session currentSession() {
        return getInstance().currentSession_();
    }

    public static synchronized Session login(@NonNull LoginAnswer loginAnswer) {
        return getInstance().login_(loginAnswer);
    }

    public static synchronized void logout() {
        getInstance().logout_();
    }

    @Nullable
    public static synchronized Session getSession() {
        return getInstance().session;
    }


    /**********************************************************************************************/
    /*****************************  Instance API implementation  **********************************/
    /**********************************************************************************************/

    private boolean isLoggedIn_() {
        return ((session != null) && (System.currentTimeMillis() <= (session.getTimeLoggedIn().getTime() + session.getExpiresIn()*1000L)));
    }

    private boolean isLoginExpired_() {
        return ((session != null) && (System.currentTimeMillis() > (session.getTimeLoggedIn().getTime() + session.getExpiresIn()*1000L)));
    }

    private Session currentSession_() {
        if (session == null) {
            restoreFromPreferences();
        }
        return session;
    }

    private Session login_(@NonNull LoginAnswer loginAnswer) {
        if (session != null) {
            throw new IllegalStateException("Already logged in - " + session);
        } else if (!loginAnswer.isSuccess()) {
            throw new IllegalArgumentException("Login was failed");
        } else {
            Uri uriShiftStart = TextUtils.isEmpty(loginAnswer.getShiftStart()) ? null : Uri.parse(loginAnswer.getShiftStart());
            Uri uriShiftEnd = TextUtils.isEmpty(loginAnswer.getShiftEnd()) ? null : Uri.parse(loginAnswer.getShiftEnd());
            Uri uriDieselPurchase = TextUtils.isEmpty(loginAnswer.getDieselPurchase()) ? null : Uri.parse(loginAnswer.getDieselPurchase());
            Session s = new Session(
                    loginAnswer.getUser(),
                    new Date(),
                    loginAnswer.getAccessToken(),
                    loginAnswer.getExpiresIn(),
                    loginAnswer.getRefreshToken(),
                    uriShiftStart,
                    uriShiftEnd,
                    uriDieselPurchase);
            setNonNullSession(s);
            saveCurrentSessionToPreferences();
            return session;
        }
    }

    private void logout_() {
        if (session != null) {
            prevLoggedInUser = session.getUser();
            session = null;
            removeCurrentSessionFromPreferences();
            savePreviousUserToPreferences();

        }
    }


    /**********************************************************************************************/
    /*****************************  Session update observer  **************************************/
    /**********************************************************************************************/
    private final Set<OnSessionUpdateListener> mObservers = new HashSet<>();

    void requestUpdates(OnSessionUpdateListener l) {
        boolean upd;
        synchronized (mObservers) {
            upd = mObservers.add(l);
        }
        if (upd && session != null) {
            l.onSessionUpdated(session);
        }
    }

    void removeUpdates(OnSessionUpdateListener l) {
        synchronized (mObservers) {
            mObservers.remove(l);
        }
    }


    public LiveData<Session> getSessionLiveData() {
        return new SessionLiveData();
    }


    public void updateMyProfileAsync() {
        synchronized (SessionHolder.class) {
            if(!isLoggedIn_()) return;
            ApplicationRestService.getInstance().getMyProfile().enqueue(new Callback<UserEntity>() {
                @Override
                public void onResponse(Call<UserEntity> call, Response<UserEntity> response) {
                    UserEntity me = response.body();
                    if (me != null && isLoggedIn_()) {
                        synchronized (SessionHolder.class) {
                            Session ns = new Session(me,
                                    session.getTimeLoggedIn(),
                                    session.getAccessToken(),
                                    session.getExpiresIn(),
                                    session.getRefreshToken(),
                                    session.getDocumentShiftStart(),
                                    session.getDocumentShiftEnd(),
                                    session.getDocumentDieselPurchase());
                            setNonNullSession(ns);
                        }
                    }
                }

                @Override
                public void onFailure(Call<UserEntity> call, Throwable t) {
                    //Not implemented
                }
            });
        }
    }


    private void setNonNullSession(@NonNull Session s) {
        if (!s.equals(session)) {
            session = s;
            List<OnSessionUpdateListener> obss;
            synchronized (mObservers) {
                obss = new ArrayList<>(mObservers.size());
                obss.addAll(mObservers);
            }
            for (OnSessionUpdateListener obs : obss) {
                obs.onSessionUpdated(s);
            }
        }
    }


    /**********************************************************************************************/
    /*****************************  Storage-related section  **************************************/
    /**********************************************************************************************/
    private static final String PREF_CURRENT_SESSION = "current_session";
    private static final String PREF_PREVIOUS_USER = "previous_user";

    private Moshi moshi;
    private SharedPreferences sPrefs;


    private Moshi getMoshi() {
        if (moshi == null) {
            synchronized (SessionHolder.class) {
                if (moshi == null) {
                    moshi = ParserProvider.getApplicationMoshi();
                }
            }
        }
        return moshi;
    }

    private SharedPreferences getPreferences() {
        if (sPrefs == null) {
            synchronized (SessionHolder.class) {
                if (sPrefs == null) {
                    sPrefs = PDA_Application.getStaticAppContext().getSharedPreferences(SessionHolder.class.getName(), Context.MODE_PRIVATE);
                }
            }
        }
        return sPrefs;
    }


    private void restoreFromPreferences() {
        if (D) Log.d(LOG_TAG, "---> Restoring current session from preferences");
        SharedPreferences prefs = getPreferences();
        String session_json = prefs.getString(PREF_CURRENT_SESSION, null);
        if (!TextUtils.isEmpty(session_json)) {
            try {
                setNonNullSession(getMoshi().adapter(Session.class).fromJson(session_json));
            }catch (IOException e) {
                System.err.println(e);
            }
        }

        String user_json = prefs.getString(PREF_PREVIOUS_USER, null);
        if (!TextUtils.isEmpty(user_json)) {
            try {
                prevLoggedInUser = getMoshi().adapter(UserEntity.class).fromJson(user_json);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private void removeCurrentSessionFromPreferences() {
        getPreferences().edit().remove(PREF_CURRENT_SESSION).commit();
    }

    private void saveCurrentSessionToPreferences() {
        if (session != null) {
            JsonAdapter<Session> adapter = getMoshi().adapter(Session.class);
            String json = adapter.toJson(session);
            getPreferences().edit().putString(PREF_CURRENT_SESSION, json).commit();
        }
    }

    private void savePreviousUserToPreferences() {
        if (prevLoggedInUser != null) {
            String json = getMoshi().adapter(UserEntity.class).toJson(prevLoggedInUser);
            getPreferences().edit().putString(PREF_PREVIOUS_USER, json).commit();
        }
    }

}
