package com.webssa.pda.sunmi.core;

import android.content.Context;
import android.content.SharedPreferences;
import android.text.TextUtils;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.PDA_Application;
import com.webssa.pda.sunmi.model.LocalCollectedGoodItem;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.model.UntiedLocalNote;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.PricedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;

import org.json.JSONArray;
import org.json.JSONException;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Created by stanislav.perchenko on 11/13/2019, 2:18 PM.
 */
public final class SingleVisitEditorStorage {

    private static SingleVisitEditorStorage INSTANCE;

    public static SingleVisitEditorStorage getInstance() {
        if (INSTANCE == null) {
            synchronized (SingleVisitEditorStorage.class) {
                if (INSTANCE == null) {
                    INSTANCE = new SingleVisitEditorStorage(PDA_Application.getStaticAppContext());
                }
            }
        }
        return INSTANCE;
    }

    private final SharedPreferences sPrefs;

    private static final String PREF_VISIT_ID = "visit_id";
    private static final String PREF_COLLECTED_GOODS = "coll_goods";
    private static final String PREF_CUSTOMER_NAME = "customer_name";
    private static final String PREF_CUSTOMER_SIGN = "customer_signature";
    private static final String PREF_OPT_DRIVER_NOTE = "opt_driver_note";
    private static final String PREF_WHITE_SACKS = "white_sacks";
    private static final String PREF_CLEAR_BAGS_COUNT = "clear_bags_count";


    private LongId<VisitEntity> visitId;
    private final Map<Long, LocalCollectedGoodItem> collectedGoods = new LinkedHashMap<>();
    private String customerName;
    private String customerSignature;
    private UntiedLocalNote optDriverNote;
    private boolean white_sacks;
    private int clear_bags_count;

    public SingleVisitEditorStorage(Context context) {
        this.sPrefs = context.getSharedPreferences(SingleVisitEditorStorage.class.getSimpleName(), Context.MODE_PRIVATE);
        long vis_id = sPrefs.getLong(PREF_VISIT_ID, 0);
        if (vis_id > 0) {
            try {
                visitId = new LongId<VisitEntity>(vis_id);
                String txtGoodsJson = sPrefs.getString(PREF_COLLECTED_GOODS, null);
                if (!TextUtils.isEmpty(txtGoodsJson)) {
                    JSONArray jItemArray = new JSONArray(txtGoodsJson);
                    for (int i = 0; i < jItemArray.length(); i++) {
                        LocalCollectedGoodItem item = LocalCollectedGoodItem.fromJsonObject(jItemArray.getJSONObject(i));
                        collectedGoods.put(item.getGoodcategoryId().getValue(), item);
                    }
                }
                customerName = sPrefs.getString(PREF_CUSTOMER_NAME, null);
                customerSignature = sPrefs.getString(PREF_CUSTOMER_SIGN, null);
                white_sacks = sPrefs.getBoolean(PREF_WHITE_SACKS, false);
                clear_bags_count = sPrefs.getInt(PREF_CLEAR_BAGS_COUNT, 0);
                String sNote = sPrefs.getString(PREF_OPT_DRIVER_NOTE, null);
                if (sNote != null) optDriverNote = UntiedLocalNote.fromJson(sNote);
            } catch (JSONException e) {
                visitId = null;
                collectedGoods.clear();
            }
        }
    }

    private void save() {
        if (visitId != null) {
            JSONArray jItemArray = new JSONArray();
            for (Iterator<Map.Entry<Long, LocalCollectedGoodItem>> itr = collectedGoods.entrySet().iterator(); itr.hasNext(); ) {
                jItemArray.put(itr.next().getValue().toJsonObject());
            }

            SharedPreferences.Editor edt = sPrefs.edit()
                    .putLong(PREF_VISIT_ID, visitId.getValue())
                    .putString(PREF_COLLECTED_GOODS, jItemArray.toString());
            if (customerName != null) edt.putString(PREF_CUSTOMER_NAME, customerName);
            if (customerSignature != null) edt.putString(PREF_CUSTOMER_SIGN, customerSignature);
            if (optDriverNote != null) edt.putString(PREF_OPT_DRIVER_NOTE, optDriverNote.toJson());
            edt.putBoolean(PREF_WHITE_SACKS, white_sacks);
            edt.putInt(PREF_CLEAR_BAGS_COUNT, clear_bags_count);
            edt.commit();
        } else {
            sPrefs.edit().clear().commit();
        }
    }

    @Nullable
    public LongId<VisitEntity> inEditVisitId() {
        return visitId;
    }

    public void setInEditVisit(LongId<VisitEntity> visitId) {
        if (this.visitId == null) {
            this.visitId = visitId;
        } else if ((this.visitId != null) && !this.visitId.equals(visitId)) {
            throw new IllegalStateException("Another Visit is being edited");
        }
    }

    /**
     * Clear all collected goods and remove current in-edit visit
     */
    public void clear() {
        if (inEditVisitId() != null) {
            visitId = null;
            collectedGoods.clear();
            customerName = null;
            customerSignature = null;
            optDriverNote = null;
            clear_bags_count = 0;
            white_sacks = false;
            save();
        }
    }

    public void addCollectedItem(@NonNull LongId<VisitEntity> visitId, LocalCollectedGoodItem collectedItem) {
        if (inEditVisitId() == null) {
            throw new IllegalStateException("No in-edit visit is set");
        } else if (!visitId.equals(inEditVisitId())) {
            throw new IllegalArgumentException("Wrong visit");
        }

        Long id = collectedItem.getGoodcategoryId().getValue();
        this.collectedGoods.remove(id);
        this.collectedGoods.put(id, collectedItem);
        save();
    }

    public void removeCollectedItem(@NonNull LongId<VisitEntity> visitId, LongId<PricedGoodEntity> goodcategoryId) {
        if (visitId.equals(inEditVisitId()) && (this.collectedGoods.remove(goodcategoryId.getValue()) != null)) {
            save();
        }
    }

    public int getCollectedGoodsCount() {
        return collectedGoods.size();
    }

    public Collection<LocalCollectedGoodItem> getAllCollectedGoods(@NonNull LongId<VisitEntity> visitId) {
        if (visitId.equals(inEditVisitId())) {
            List<LocalCollectedGoodItem> goods = new LinkedList<>();
            for (Iterator<Map.Entry<Long, LocalCollectedGoodItem>> itr = collectedGoods.entrySet().iterator(); itr.hasNext(); ) {
                goods.add(itr.next().getValue());
            }
            //Set<LocalCollectedGoodItem> result = new HashSet<>(collectedGoods.size());
            //for (LocalCollectedGoodItem item : collectedGoods.values()) result.add(item);
            return goods;
        } else {
            return new LinkedList<>();
        }
    }

    public boolean isHasCustomerData() {
        return !(TextUtils.isEmpty(customerName) || TextUtils.isEmpty(customerSignature));
    }

    public String getCustomerName() {
        return customerName;
    }

    public String getCustomerSignature() {
        return customerSignature;
    }

    public boolean isWhiteSacks() {
        return white_sacks;
    }

    public int getClearBagsCount() {
        return clear_bags_count;
    }

    public void setWhiteSacks(boolean white_sacks) {
        this.white_sacks = white_sacks;
        save();
    }
    public void setClearBagsCount(int clear_bags_count) {
        this.clear_bags_count = clear_bags_count;
        save();
    }


    public void setCustomerNameAndSignature(@NonNull LongId<VisitEntity> visitId, @NonNull String customerName, @Nullable String customerSignature) {
        if (inEditVisitId() == null) {
            throw new IllegalStateException("No in-edit visit is set");
        } else if (!visitId.equals(inEditVisitId())) {
            throw new IllegalArgumentException("Wrong visit");
        } else if (!customerName.equalsIgnoreCase(this.customerName) || !TextUtils.equals(this.customerSignature, customerSignature)) {
            this.customerName = customerName;
            this.customerSignature = customerSignature;
            save();
        }
    }

    public void clearCustomerSignature(@NonNull LongId<VisitEntity> visitId) {
        if (inEditVisitId() == null) {
            throw new IllegalStateException("No in-edit visit is set");
        } else if (!visitId.equals(inEditVisitId())) {
            throw new IllegalArgumentException("Wrong visit");
        } else if (customerSignature != null) {
            customerSignature = null;
            save();
        }
    }

    public UntiedLocalNote getOptDriverNote() {
        return optDriverNote;
    }

    public void setOptDriverNote(@NonNull LongId<VisitEntity> visitId, @NonNull UntiedLocalNote note) {
        if (inEditVisitId() == null) {
            throw new IllegalStateException("No in-edit visit is set");
        } else if (!visitId.equals(inEditVisitId())) {
            throw new IllegalArgumentException("Wrong visit");
        } else {
            optDriverNote = note;
            save();
        }
    }

    public void clearOptDriverNote(@NonNull LongId<VisitEntity> visitId) {
        if (inEditVisitId() == null) {
            throw new IllegalStateException("No in-edit visit is set");
        } else if (!visitId.equals(inEditVisitId())) {
            throw new IllegalArgumentException("Wrong visit");
        } else if (optDriverNote != null) {
            optDriverNote = null;
            save();
        }
    }

    @Nullable
    public LocalVisitResult buildVisitResult() {
        try {
            List<LocalCollectedGoodItem> goods = new LinkedList<>();
            for (Iterator<Map.Entry<Long, LocalCollectedGoodItem>> itr = collectedGoods.entrySet().iterator(); itr.hasNext(); ) {
                goods.add(itr.next().getValue());
            }

            final String myName = SessionHolder.getSession().getUser().getFullName();

            return LocalVisitResult.builder()
                    .setVisitId(visitId)
                    .setCollectedGoods(goods.isEmpty() ? null : Collections.unmodifiableList(goods))
                    .setDriverName(myName)
                    .setCustomerName(customerName)
                    .setCustomerSignature(customerSignature)
                    .setOptDriverNote(optDriverNote)
                    .setClearBags(clear_bags_count)
                    .setWhiteSacks(white_sacks)
                    .build();
        } catch (Exception e) {
            //If not all required data has been collected
            return null;
        }
    }
}
