package com.webssa.pda.sunmi.dataapi;

import android.content.Context;
import android.net.Uri;

import androidx.annotation.NonNull;

import com.squareup.moshi.JsonDataException;
import com.webssa.Loader;
import com.webssa.library.core.error.LocalError;
import com.webssa.library.core.processor.AbstractExecutable;
import com.webssa.pda.sunmi.core.LocalCompletedVisitsStorage;
import com.webssa.pda.sunmi.core.SingleVisitEditorStorage;
import com.webssa.pda.sunmi.model.LocalCollectedGoodItem;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.rest.ApplicationRestService;
import com.webssa.pda.sunmi.rest.UploadImagesRestRequest;
import com.webssa.pda.sunmi.rest.helper.HttpClientProvider;
import com.webssa.pda.sunmi.rest.model.CollectedGoodUploadEntity;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.StartVisitPutBody;
import com.webssa.pda.sunmi.rest.model.VisitResponse;
import com.webssa.pda.sunmi.rest.model.VisitResultPostBody;
import com.webssa.pda.sunmi.rest.model.VisitState;
import com.webssa.pda.sunmi.rest.model.impl.CollectedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.NoteEntity;
import com.webssa.pda.sunmi.rest.model.impl.PricedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitStateEntity;
import com.webssa.pda.sunmi.storage.dao.impl.CollectedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.NoteDAO;
import com.webssa.pda.sunmi.storage.dao.impl.PricedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.ShiftDAO;
import com.webssa.pda.sunmi.storage.dao.impl.VisitDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.PricedGoodDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import okhttp3.OkHttpClient;
import retrofit2.Call;
import retrofit2.Response;

/**
 * Created by stanislav.perchenko on 11/20/2019, 2:42 PM.
 */
public class ApiCompleteVisit extends AbstractExecutable<VisitDbModel> {


    private final Context context;
    private final LocalVisitResult visitResult;

    public static ApiCompleteVisit create(Context context, @NonNull LocalVisitResult visitResult) {
        return new ApiCompleteVisit(context, visitResult);
    }

    private ApiCompleteVisit(Context context, LocalVisitResult visitResult) {
        assert (context != null);
        assert (visitResult != null);
        this.context = context;
        this.visitResult = visitResult;
    }

    @Override
    public void executeSynchronously() throws Exception {
        DbSelectionCache cache = new DbSelectionCache();
        try (ShiftDAO shiftDao = new ShiftDAO(cache);  VisitDAO vDao = new VisitDAO(cache)) {
            VisitDbModel visit = vDao.selectEntryById(visitResult.visitId());
            if (visit == null) {
                setError(LocalError.Companion.create(context, "Visit not found - " + visitResult.visitId()));
            } else if (visit.state() != VisitState.IN_PROGRESS) {
                setError(LocalError.Companion.create(context, "Visit in wrong state - " + visit.state()));
            } else {
                processVisit(shiftDao, vDao, visit);
                SingleVisitEditorStorage.getInstance().clear();
            }
        }
    }

    private void processVisit(ShiftDAO shiftDao, VisitDAO vDao, VisitDbModel visit) {
        if (visit.state() != VisitState.IN_PROGRESS) throw new IllegalStateException("Visit is in wrong space");
        boolean success;
        if (visit.isAdHocVisit() && !visit.isRemoteCreated()) {
            visit = completeVisitLocal(vDao, visit);

            ApiSyncVisitsBack apiSync = new ApiSyncVisitsBack(visit.shiftId());
            VisitDbModel syncedVisit = apiSync.processVisit(shiftDao, vDao, visit);
            setResult((syncedVisit == null) ? visit : syncedVisit);
            return;

        } else if (!visit.isSynced()) {
            success = syncVisitBack(vDao, visit);
            if (!success) {
                setResult(completeVisitLocal(vDao, visit));
                return;
            }
        }

        success = completeVisitRemote(vDao, visit);
        if (success) {
            setResult(vDao.selectEntryById(visit.id()));
        } else {
            setResult(completeVisitLocal(vDao, visit));
        }
    }


    /******************  Remote complete Visit which is SYNCED and IN_PROGERSS  *******************/
    private boolean completeVisitRemote(VisitDAO vDao, VisitDbModel visit) {
        if (!visit.isSynced() && (visit.state() != VisitState.IN_PROGRESS)) {
            throw new IllegalStateException();
        }
        List<Uri> images;
        if ((visitResult.optDriverNote() == null) || visitResult.optDriverNote().getImportedImages().isEmpty()) {
            images = Collections.emptyList();
        } else {
            OkHttpClient httpClient = HttpClientProvider.getInstance().getApiHttpClient();
            final UploadImagesRestRequest imgUplRequest = UploadImagesRestRequest.create(visitResult.optDriverNote().getImportedImages());
            if (imgUplRequest.execute(httpClient)) {
                images = imgUplRequest.getResult();
            } else {
                return false;
            }
        }

        boolean result;
        String optNoteText = (visitResult.optDriverNote() == null) ? null : visitResult.optDriverNote().getNoteText();
        List<CollectedGoodUploadEntity> goods = new ArrayList<>(visitResult.collectedGoods().size());

        try (PricedGoodDAO goodsDao = new PricedGoodDAO(new DbSelectionCache())) {
            for (LocalCollectedGoodItem localGood : visitResult.collectedGoods()) {
                PricedGoodDbModel goodModel = goodsDao.selectGoodForSite(localGood.getSiteId(), localGood.getGoodcategoryId());
                if (goodModel != null) {
                    goods.add(new CollectedGoodUploadEntity(visitResult.visitId(), localGood.getGoodcategoryId(), goodModel.priceId(), localGood.getWeightKg()));
                }
            }
        }

        Loader.loadLocalNotes();

        Long em_id = (visitResult.optDriverNote() == null
                || visitResult.optDriverNote().getOptEmergency() == null ||
                visitResult.optDriverNote().getOptEmergency().getId().getValue()
                        == 0) ? null : visitResult.optDriverNote().getOptEmergency().getId().getValue();
        VisitResultPostBody body = new VisitResultPostBody(visitResult.visitId(), null, visitResult.customerName(),
                visitResult.customerSignature(), em_id,
                optNoteText, images, goods, visitResult.whiteSacks(), visitResult.clearBags());
        try {
            Call<VisitResponse> call = ApplicationRestService.getInstance().completeVisit(visitResult.visitId(), body);
            Response<VisitResponse> response = call.execute();
            if (response.isSuccessful()) {
                VisitResponse respData = response.body();
                if (respData == null) {
                    //setError(ParseRestError.Companion.create(context, call.request().url().toString()));
                    result = false;
                } else {
                    result = continueWithResponse(vDao, respData, body.getCollectedGoods(), visit.isAdHocVisit());
                }
            } else {
                result = false;
            }
        } catch (IOException e) {
            result = false;
        } catch (JsonDataException e) {
            result = false;
        }
        return result;
    }

    private boolean continueWithResponse(VisitDAO visDao, VisitResponse respData, List<CollectedGoodUploadEntity> postedGoods, boolean isAdHocVisit) {
        final VisitEntity visit = isAdHocVisit ? respData.getVisit().copyWithAdHocFlag() : respData.getVisit();
        final NoteEntity optNote = respData.getRelatedNote();
        final List<CollectedGoodEntity> finGoods = new ArrayList<>(16);
        for (CollectedGoodEntity g : respData.getCollectedGoods()) {
            if (g.getGoodcategoryId() == null) {
                int order = g.getOrder();
                LongId<PricedGoodEntity> goodId = postedGoods.get(order).getGoodcategoryId();
                CollectedGoodEntity ng = new CollectedGoodEntity(g.id(), g.getVisitId(), goodId, g.getGoodName(), g.getCollectedWeight(), g.getCollectedTotalAmount(), g.getCollectedTotalScale(), order);
                finGoods.add(ng);
            } else {
                finGoods.add(g);
            }
        }


        try (NoteDAO noteDao = new NoteDAO(); CollectedGoodDAO goodDao = new CollectedGoodDAO()) {
            noteDao.beginTransaction();
            try {
                if (optNote != null) {
                    noteDao.createOrUpdateInLocalDB(optNote);
                }
                for (CollectedGoodEntity g : finGoods) {
                    goodDao.createOrUpdateInLocalDB(g);
                }
                visDao.createOrUpdateInLocalDB(visit);
                noteDao.setTransactionSuccessful();
                return true;
            } catch (Exception e) {
                return false;
            } finally {
                noteDao.endTransaction();
            }
        }

    }


    /*********************  Complete visit locally  ***********************************************/
    private VisitDbModel completeVisitLocal(VisitDAO vDao, VisitDbModel visit) {
        if (visit.state() != VisitState.IN_PROGRESS) {
            throw new IllegalStateException();
        }
        vDao.completeVisitStateLocally(visit.id());
        LocalCompletedVisitsStorage.saveLocalVisitResult(visitResult);
        return vDao.selectEntryById(visit.id());
    }


    /********************  Try to sync visit back  ************************************************/
    //This visit can only be in the IN_PROGRESS state an not synced.
    private boolean syncVisitBack(VisitDAO vDao, VisitDbModel visit) {
        if ((visit.state() != VisitState.IN_PROGRESS) || visit.isSynced()) {
            String err = String.format("Wrong Visit state - %s. Must be IN_PROGRESS and NOT SYNCED", visit.id());
            throw new IllegalStateException(err);
        }

        boolean success = false;

        VisitStateEntity remoteState = getVisitStateRemote(visit.id());
        if (remoteState != null && remoteState.getState() == VisitState.NEW) {
            success = startVisitInPast(vDao, visit.id(), visit.actualStartTime());
        }
        return success;
    }

    private VisitStateEntity getVisitStateRemote(LongId<VisitEntity> visitId) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.INSTANCE_GOOD.getVisitState(visitId).execute();
            if (response.isSuccessful()) {
                return response.body();
            } else {
                return null;
            }
        } catch (IOException e) {
            return null;
        }
    }

    private boolean startVisitInPast(VisitDAO dao, LongId<VisitEntity> visitId, Date pastStartTime) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.getInstance().startVisit(visitId, new StartVisitPutBody(pastStartTime)).execute();
            if (response.code() == 200 && response.body() != null && visitId.equals(response.body().getId()) && (response.body().getState() == VisitState.IN_PROGRESS)) {
                VisitStateEntity new_state = response.body();
                dao.updateVisitStateFromRemote(new_state);
                return true;
            } else {
                return false;
            }
        } catch (IOException e) {
            return false;
        }
    }


}
