package com.webssa.pda.sunmi.dataapi;

import android.content.Context;
import android.net.Uri;

import com.squareup.moshi.JsonDataException;
import com.webssa.library.core.error.LocalError;
import com.webssa.library.core.processor.AbstractExecutable;
import com.webssa.pda.sunmi.core.LocalFailedVisitsStorage;
import com.webssa.pda.sunmi.core.SingleVisitEditorStorage;
import com.webssa.pda.sunmi.model.UntiedLocalNote;
import com.webssa.pda.sunmi.rest.ApplicationRestService;
import com.webssa.pda.sunmi.rest.UploadImagesRestRequest;
import com.webssa.pda.sunmi.rest.helper.HttpClientProvider;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.StartVisitPutBody;
import com.webssa.pda.sunmi.rest.model.VisitFailPostBody;
import com.webssa.pda.sunmi.rest.model.VisitState;
import com.webssa.pda.sunmi.rest.model.impl.EmergencyCategoryEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitStateEntity;
import com.webssa.pda.sunmi.storage.dao.impl.NoteDAO;
import com.webssa.pda.sunmi.storage.dao.impl.ShiftDAO;
import com.webssa.pda.sunmi.storage.dao.impl.VisitDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;

import java.io.IOException;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import okhttp3.OkHttpClient;
import retrofit2.Call;
import retrofit2.Response;

/**
 * Created by stanislav.perchenko on 11/20/2019, 11:39 AM.
 */
public class ApiFailVisit extends AbstractExecutable<VisitDbModel> {

    private final Context context;
    private final LongId<VisitEntity> visitId;
    private final UntiedLocalNote note;

    public static ApiFailVisit create(Context c, LongId<VisitEntity> visitId, UntiedLocalNote note) {
        return new ApiFailVisit(c, visitId, note);
    }

    private ApiFailVisit(Context c, LongId<VisitEntity> visitId, UntiedLocalNote note) {
        super();
        context = c;
        this.visitId = visitId;
        this.note = note;
    }

    @Override
    public void executeSynchronously() throws Exception {
        DbSelectionCache cache = new DbSelectionCache();
        try (ShiftDAO shiftDao = new ShiftDAO(cache);  VisitDAO vDao = new VisitDAO(cache)) {
            VisitDbModel visit = vDao.selectEntryById(visitId);
            if (visit == null) {
                setError(LocalError.Companion.create(context, "Visit not found - " + visitId));
            } else if (visit.state() != VisitState.IN_PROGRESS) {
                setError(LocalError.Companion.create(context, "Visit in wrong state - "+visit.state()));
            } else {
                processVisit(shiftDao, vDao, visit);
                SingleVisitEditorStorage.getInstance().clear();
            }
        }
    }

    private void processVisit(ShiftDAO shiftDao, VisitDAO vDao, VisitDbModel visit) {
        if (visit.state() != VisitState.IN_PROGRESS) throw new IllegalStateException("Visit is in wrong space");
        boolean success;
        if (visit.isAdHocVisit() && !visit.isRemoteCreated()) {
            visit = failVisitLocal(vDao, visit);

            ApiSyncVisitsBack apiSync = new ApiSyncVisitsBack(visit.shiftId());
            VisitDbModel syncedVisit = apiSync.processVisit(shiftDao, vDao, visit);
            setResult((syncedVisit == null) ? visit : syncedVisit);
            return;

        } else if (!visit.isSynced()) {
            success = syncVisitBack(vDao, visit);
            if (!success) {
                setResult(failVisitLocal(vDao, visit));
                return;
            }
        }



        success = failVisitRemote(vDao, visit);
        if (success) {
            setResult(vDao.selectEntryById(visit.id()));
        } else {
            setResult(failVisitLocal(vDao, visit));
        }
    }

    /******************  Remote fail Visit which is SYNCED and IN_PROGERSS  ***********************/
    private boolean failVisitRemote(VisitDAO vDao, VisitDbModel visit) {
        List<Uri> images;
        if (note.getImportedImages().isEmpty()) {
            images = Collections.emptyList();
        } else {
            OkHttpClient httpClient = HttpClientProvider.getInstance().getApiHttpClient();
            final UploadImagesRestRequest imgUplRequest = UploadImagesRestRequest.create(note.getImportedImages());
            if (imgUplRequest.execute(httpClient)) {
                images = imgUplRequest.getResult();
            } else {
                images = Collections.emptyList();
            }
        }




        LongId<EmergencyCategoryEntity> emergId = (note.getOptEmergency() == null) ? null : note.getOptEmergency().id();
        VisitFailPostBody body = new VisitFailPostBody(visit.id(), null, emergId, note.getNoteText(), images);
        boolean result;
        try {
            Call<VisitStateEntity> call = ApplicationRestService.getInstance().failVisit(visit.id(), body);
            Response<VisitStateEntity> response = call.execute();
            if (response.isSuccessful()) {
                VisitStateEntity respData = response.body();
                if (respData == null) {
                    result = false;
                } else {
                    result = continueWithResponse(vDao, respData);
                }
            } else {
                result = false;
            }
        } catch (IOException e) {
            result = false;
        } catch (JsonDataException e) {
            result = false;
        }
        return result;
    }

    private boolean continueWithResponse(VisitDAO visDao, VisitStateEntity respData) {
        try (NoteDAO noteDao = new NoteDAO()) {
            noteDao.beginTransaction();
            try {
                if (respData.getNote() != null) {
                    noteDao.createOrUpdateInLocalDB(respData.getNote());
                }
                visDao.updateVisitStateFromRemote(respData);
                VisitDbModel rv = visDao.selectEntryById(respData.id());
                if ((rv == null) || (rv.state() != VisitState.FAILED)) {
                    throw new Exception("No visit or wrong state - " + respData.id());
                }
                noteDao.setTransactionSuccessful();
                return true;
            } catch (Exception e) {
                return false;
            } finally {
                noteDao.endTransaction();
            }
        }
    }

    /*********************  Fail visit locally  ***************************************************/
    private VisitDbModel failVisitLocal(VisitDAO vDao, VisitDbModel visit) {
        if (visit.state() != VisitState.IN_PROGRESS) {
            throw new IllegalStateException();
        }
        vDao.failVisitStateLocally(visit.id());
        LocalFailedVisitsStorage.saveLocalVisitResult(visitId, note);
        return vDao.selectEntryById(visit.id());
    }


    /********************  Try to sync visit back  ************************************************/
    //This visit can only be in the IN_PROGRESS state an not synced.
    private boolean syncVisitBack(VisitDAO vDao, VisitDbModel visit) {
        if ((visit.state() != VisitState.IN_PROGRESS) || visit.isSynced()) {
            String err = String.format("Wrong Visit state - %s. Must be IN_PROGRESS and NOT SYNCED", visit.id());
            throw new IllegalStateException(err);
        }

        boolean success = false;
        VisitStateEntity remoteState = getVisitStateRemote(visit.id());
        if (remoteState != null && remoteState.getState() == VisitState.NEW) {
            success = startVisitInPast(vDao, visit.id(), visit.actualStartTime());
        }
        return success;
    }

    private VisitStateEntity getVisitStateRemote(LongId<VisitEntity> visitId) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.INSTANCE_GOOD.getVisitState(visitId).execute();
            if (response.isSuccessful()) {
                return response.body();
            } else {
                return null;
            }
        } catch (IOException e) {
            return null;
        }
    }

    private boolean startVisitInPast(VisitDAO dao, LongId<VisitEntity> visitId, Date pastStartTime) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.getInstance().startVisit(visitId, new StartVisitPutBody(pastStartTime)).execute();
            if (response.code() == 200 && response.body() != null && visitId.equals(response.body().getId()) && (response.body().getState() == VisitState.IN_PROGRESS)) {
                VisitStateEntity new_state = response.body();
                dao.updateVisitStateFromRemote(new_state);
                return true;
            } else {
                return false;
            }
        } catch (IOException e) {
            return false;
        }
    }
}
