package com.webssa.pda.sunmi.dataapi;

import android.net.Uri;

import androidx.annotation.Nullable;

import com.squareup.moshi.JsonDataException;
import com.webssa.Loader;
import com.webssa.library.core.processor.AbstractExecutable;
import com.webssa.pda.sunmi.core.LocalCompletedVisitsStorage;
import com.webssa.pda.sunmi.core.LocalFailedVisitsStorage;
import com.webssa.pda.sunmi.model.LocalCollectedGoodItem;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.model.UntiedLocalNote;
import com.webssa.pda.sunmi.model.VisitCreationModel;
import com.webssa.pda.sunmi.rest.ApplicationRestService;
import com.webssa.pda.sunmi.rest.UploadImagesRestRequest;
import com.webssa.pda.sunmi.rest.helper.HttpClientProvider;
import com.webssa.pda.sunmi.rest.model.CollectedGoodUploadEntity;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.StartVisitPutBody;
import com.webssa.pda.sunmi.rest.model.VisitFailPostBody;
import com.webssa.pda.sunmi.rest.model.VisitResponse;
import com.webssa.pda.sunmi.rest.model.VisitResultPostBody;
import com.webssa.pda.sunmi.rest.model.VisitState;
import com.webssa.pda.sunmi.rest.model.impl.CollectedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.EmergencyCategoryEntity;
import com.webssa.pda.sunmi.rest.model.impl.NoteEntity;
import com.webssa.pda.sunmi.rest.model.impl.PricedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitStateEntity;
import com.webssa.pda.sunmi.storage.dao.impl.CollectedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.NoteDAO;
import com.webssa.pda.sunmi.storage.dao.impl.PricedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.ShiftDAO;
import com.webssa.pda.sunmi.storage.dao.impl.VisitDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.PricedGoodDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.ShiftDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;
import com.webssa.pda.sunmi.util.DateNoTime;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import okhttp3.OkHttpClient;
import retrofit2.Call;
import retrofit2.Response;

/**
 * Created by stanislav.perchenko on 12/14/2019, 4:16 PM.
 */
public class ApiSyncVisitsBack extends AbstractExecutable<List<VisitDbModel>> {

    public static ApiSyncVisitsBack forShift(LongId<ShiftRestEntity> shiftId) {
        return new ApiSyncVisitsBack(shiftId);
    }


    private final LongId<ShiftRestEntity> shiftId;

    ApiSyncVisitsBack(LongId<ShiftRestEntity> shiftId) {
        this.shiftId = shiftId;
    }

    @Override
    public void executeSynchronously() throws Exception {
        long tStart = System.currentTimeMillis();
        List<VisitDbModel> resultVisits = new LinkedList<>();
        DbSelectionCache cache = new DbSelectionCache();
        try (ShiftDAO shiftDao = new ShiftDAO(cache); VisitDAO vDao = new VisitDAO(cache)) {
            List<VisitDbModel> visits = vDao.selectVisistForShiftSortedByOrder(shiftId);

            //We must evaluate Ad-Hoc visits first. In case they are created offline
            // we need to prevent shift from been closed early before ad-hoc creation.
            // Bug #10362
            Collections.sort(visits, (v1, v2) -> {
                if (v1.isAdHocVisit() && !v2.isAdHocVisit()) {
                    return -1;
                } else if (!v1.isAdHocVisit() && v2.isAdHocVisit()) {
                    return 1;
                } else {
                    return 0;
                }
            });

            for (VisitDbModel visit : visits) {
                if (!visit.isSynced()) {
                    VisitDbModel processedVisit = processVisit(shiftDao, vDao, visit);
                    if (processedVisit != null) {
                        resultVisits.add(processedVisit);
                    }
                }
            }
        }

        int dt = (int) (System.currentTimeMillis() - tStart);
        if (dt < 1100) Thread.sleep(1100 - dt);

        setResult(resultVisits);
    }

    @Nullable
    VisitDbModel processVisit(ShiftDAO shiftDao, VisitDAO dao, VisitDbModel localVisit) {
        VisitDbModel resultVisit = null;
        VisitState remoteState;

        if (localVisit.isAdHocVisit() && !localVisit.isRemoteCreated()) {
            resultVisit = createAdHocVisitRemote(localVisit, shiftDao.selectExecutionDateForShift(localVisit.shiftId()));
            if (resultVisit == null) {
                return null;
            } else {
                remoteState = resultVisit.state();
                LongId<VisitEntity> fakeOldVisitId = localVisit.id();
                localVisit = localVisit.withId(resultVisit.id()); //Replace ID of the argument visit

                dao.deleteEntryById(fakeOldVisitId);
                LocalVisitResult optResult = LocalCompletedVisitsStorage.getLocalVisitResult(fakeOldVisitId);
                if (optResult != null) {
                    LocalVisitResult updLocalResults = optResult.withVisitId(resultVisit.id());
                    LocalCompletedVisitsStorage.saveLocalVisitResult(updLocalResults);
                    LocalCompletedVisitsStorage.removeVisitResult(fakeOldVisitId);
                }

                UntiedLocalNote optFail = LocalFailedVisitsStorage.getLocalVisitResult(fakeOldVisitId);
                if (optFail != null) {
                    LocalFailedVisitsStorage.saveLocalVisitResult(resultVisit.id(), optFail);
                    LocalFailedVisitsStorage.removeVisitResult(fakeOldVisitId);
                }

            }
        } else {
            VisitStateEntity remVisit = getVisitStateRemote(localVisit.id());
            if (remVisit == null) {
                return null;
            } else {
                remoteState = remVisit.getState();
            }
        }

        if ((remoteState == VisitState.NEW) && (localVisit.state() != VisitState.NEW)) {
            resultVisit = startVisitInPast(dao, localVisit.id(), localVisit.actualStartTime());
            if (resultVisit == null) {
                return null;
            } else {
                remoteState = resultVisit.state();
            }
        }

        if ((remoteState == VisitState.IN_PROGRESS) && (localVisit.state() == VisitState.COMPLETED)) {
            LocalVisitResult visResult = LocalCompletedVisitsStorage.getLocalVisitResult(localVisit.id());
            if (visResult != null) {
                VisitDbModel updVisit = completeVisitInPast(dao, visResult, localVisit.isAdHocVisit(), localVisit.actualCompleteTime());
                if (updVisit != null) {
                    resultVisit = updVisit;
                    remoteState = updVisit.state();
                }
            }
        }

        if ((remoteState == VisitState.IN_PROGRESS) && (localVisit.state() == VisitState.FAILED)) {
            UntiedLocalNote failNote = LocalFailedVisitsStorage.getLocalVisitResult(localVisit.id());
            if (failNote != null) {
                VisitDbModel updVisit = failVisitInPast(dao, localVisit.id(), failNote, localVisit.actualCompleteTime());
                if (updVisit != null) {
                    resultVisit = updVisit;
                    remoteState = updVisit.state();
                }
            }
        }

        return resultVisit;
    }

    @Nullable
    VisitDbModel createAdHocVisitRemote(VisitDbModel argVisit, @Nullable DateNoTime shiftDate) {
        if (!argVisit.isAdHocVisit() || argVisit.isRemoteCreated()) {
            return argVisit;
        }

        VisitCreationModel vcm = new VisitCreationModel(argVisit.shiftId(), argVisit.site().id(), shiftDate, argVisit.suggestedTime());
        ApiCreateAdHocVisit apiCreate = ApiCreateAdHocVisit.create(vcm, false);
        try {
            apiCreate.executeSynchronously();
        } catch (Exception e) {
            return null;
        }
        if (apiCreate.getResult() != null) {
            VisitDbModel result = apiCreate.getResult();
            int deleted;
            try (VisitDAO vDao = new VisitDAO(null)) {
                deleted = vDao.deleteEntryById(argVisit.id());
            }

            return result;
        } else {
            return null;
        }
    }


    @Nullable
    private VisitStateEntity getVisitStateRemote(LongId<VisitEntity> visitId) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.INSTANCE_GOOD.getVisitState(visitId).execute();
            if (response.isSuccessful()) {
                return response.body();
            } else {
                return null;
            }
        } catch (IOException e) {
            return null;
        }
    }

    private VisitDbModel startVisitInPast(VisitDAO dao, LongId<VisitEntity> visitId, Date pastStartTime) {
        try {
            Response<VisitStateEntity> response = ApplicationRestService.getInstance()
                    .startVisit(visitId, new StartVisitPutBody(pastStartTime))
                    .execute();
            if (response.code() == 200 && response.body() != null && visitId.equals(response.body().getId()) && (response.body().getState() == VisitState.IN_PROGRESS)) {
                VisitStateEntity new_state = response.body();
                dao.updateVisitStateFromRemote(new_state);
                return dao.selectEntryById(visitId);
            } else {
                return null;
            }
        } catch (IOException e) {
            return null;
        }
    }


    /*********************************  Complete Visit in the Past  *******************************/
    @Nullable
    private VisitDbModel completeVisitInPast(VisitDAO vDao, LocalVisitResult visitResult, boolean isAdHocVisit, Date pastCompleteTime) {
        List<Uri> images;
        if ((visitResult.optDriverNote() == null) || visitResult.optDriverNote().getImportedImages().isEmpty()) {
            images = Collections.emptyList();
        } else {
            OkHttpClient httpClient = HttpClientProvider.getInstance().getApiHttpClient();
            final UploadImagesRestRequest imgUplRequest = UploadImagesRestRequest.create(visitResult.optDriverNote().getImportedImages());
            if (imgUplRequest.execute(httpClient)) {
                images = imgUplRequest.getResult();
            } else {
                return null;
            }
        }


        String optNoteText = (visitResult.optDriverNote() == null) ? null : visitResult.optDriverNote().getNoteText();


        List<CollectedGoodUploadEntity> goods = new ArrayList<>(visitResult.collectedGoods().size());
        try (PricedGoodDAO goodsDao = new PricedGoodDAO(new DbSelectionCache())) {
            for (LocalCollectedGoodItem localGood : visitResult.collectedGoods()) {
                PricedGoodDbModel goodModel = goodsDao.selectGoodForSite(localGood.getSiteId(), localGood.getGoodcategoryId());   //.selectEntryById(localGood.getGoodcategoryId());
                if (goodModel != null) {
                    goods.add(new CollectedGoodUploadEntity(visitResult.visitId(), localGood.getGoodcategoryId(), goodModel.priceId(), localGood.getWeightKg()));
                }
            }
        }

        Loader.loadLocalNotes();

        Long em_id = (visitResult.optDriverNote() == null
                || visitResult.optDriverNote().getOptEmergency() == null ||
                visitResult.optDriverNote().getOptEmergency().getId().getValue()
                        == 0) ? null : visitResult.optDriverNote().getOptEmergency().getId().getValue();
        VisitResultPostBody body = new VisitResultPostBody(visitResult.visitId(), pastCompleteTime, visitResult.customerName(),
                visitResult.customerSignature(), em_id, optNoteText, images, goods,
                visitResult.whiteSacks(), visitResult.clearBags());

        VisitDbModel result;
        try {
            Call<VisitResponse> call = ApplicationRestService.getInstance().completeVisit(visitResult.visitId(), body);
            Response<VisitResponse> response = call.execute();
            if (response.isSuccessful()) {
                VisitResponse respData = response.body();
                if (respData == null) {
                    result = null;
                } else {
                    result = continueCompleteWithResponse(vDao, respData, body.getCollectedGoods(), isAdHocVisit);
                }
            } else {
                result = null;
            }
        } catch (IOException e) {
            result = null;
        } catch (JsonDataException e) {
            result = null;
        } finally {
            LocalCompletedVisitsStorage.removeVisitResult(visitResult.visitId());
        }
        return result;
    }

    @Nullable
    private VisitDbModel continueCompleteWithResponse(VisitDAO visDao, VisitResponse respData, List<CollectedGoodUploadEntity> postedGoods, boolean isAdHocVisit) {
        final VisitEntity visit = isAdHocVisit ? respData.getVisit().copyWithAdHocFlag() : respData.getVisit();
        final NoteEntity optNote = respData.getRelatedNote();
        final List<CollectedGoodEntity> finGoods = new ArrayList<>(16);
        for (CollectedGoodEntity g : respData.getCollectedGoods()) {
            if (g.getGoodcategoryId() == null) {
                int order = g.getOrder();
                LongId<PricedGoodEntity> goodId = postedGoods.get(order).getGoodcategoryId();
                CollectedGoodEntity ng = new CollectedGoodEntity(g.id(), g.getVisitId(), goodId, g.getGoodName(), g.getCollectedWeight(), g.getCollectedTotalAmount(), g.getCollectedTotalScale(), order);
                finGoods.add(ng);
            } else {
                finGoods.add(g);
            }
        }

        try (NoteDAO noteDao = new NoteDAO(); CollectedGoodDAO goodDao = new CollectedGoodDAO()) {
            noteDao.beginTransaction();
            try {
                if (optNote != null) {
                    noteDao.createOrUpdateInLocalDB(optNote);
                }
                for (CollectedGoodEntity g : finGoods) {
                    goodDao.createOrUpdateInLocalDB(g);
                }
                visDao.createOrUpdateInLocalDB(visit);
                noteDao.setTransactionSuccessful();
            } catch (Exception e) {
                return null;
            } finally {
                noteDao.endTransaction();
            }
        }

        return visDao.selectEntryById(visit.id());
    }


    /*********************************  Fail Visit in the Past  ***********************************/
    @Nullable
    private VisitDbModel failVisitInPast(VisitDAO vDao, LongId<VisitEntity> visitId, UntiedLocalNote note, Date pastFailTime) {
        List<Uri> images;
        if (note.getImportedImages().isEmpty()) {
            images = Collections.emptyList();
        } else {
            OkHttpClient httpClient = HttpClientProvider.getInstance().getApiHttpClient();
            final UploadImagesRestRequest imgUplRequest = UploadImagesRestRequest.create(note.getImportedImages());
            if (imgUplRequest.execute(httpClient)) {
                images = imgUplRequest.getResult();
            } else {
                images = Collections.emptyList();
            }
        }


        LongId<EmergencyCategoryEntity> emergId = (note.getOptEmergency() == null) ? null : note.getOptEmergency().id();
        VisitFailPostBody body = new VisitFailPostBody(visitId, pastFailTime, emergId, note.getNoteText(), images);
        VisitDbModel result;
        try {
            Call<VisitStateEntity> call = ApplicationRestService.getInstance().failVisit(visitId, body);
            Response<VisitStateEntity> response = call.execute();
            if (response.isSuccessful()) {
                VisitStateEntity respData = response.body();
                if (respData == null) {
                    result = null;
                } else {
                    result = continueFailWithResponse(vDao, respData);
                }
            } else {
                result = null;
            }
        } catch (IOException e) {
            result = null;
        } catch (JsonDataException e) {
            result = null;
        } finally {
            LocalFailedVisitsStorage.removeVisitResult(visitId);
        }
        return result;
    }

    private VisitDbModel continueFailWithResponse(VisitDAO visDao, VisitStateEntity respData) {
        try (NoteDAO noteDao = new NoteDAO()) {
            noteDao.beginTransaction();
            try {
                if (respData.getNote() != null) {
                    noteDao.createOrUpdateInLocalDB(respData.getNote());
                }
                visDao.updateVisitStateFromRemote(respData);
                VisitDbModel rv = visDao.selectEntryById(respData.id());
                if ((rv == null) || (rv.state() != VisitState.FAILED)) {
                    throw new Exception("No visit or wrong state - " + respData.id());
                }
                noteDao.setTransactionSuccessful();
            } catch (Exception e) {
                return null;
            } finally {
                noteDao.endTransaction();
            }
            return visDao.selectEntryById(respData.id());
        }
    }
}
