package com.webssa.pda.sunmi.model;

import android.os.Parcel;
import android.os.Parcelable;
import android.text.TextUtils;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.library.attachment.model.CoverImportModel;
import com.webssa.pda.sunmi.core.EmergencyHolder;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.EmergencyCategoryEntity;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * This model encapsulates note text and locally-imported images CoverImportModel
 * It is used in the NoteCreateActivity and in the visit executor activity.
 *
 * Created by stanislav.perchenko on 11/5/2019, 11:49 AM.
 */
public class UntiedLocalNote implements Parcelable {

    private final String noteText;
    private final EmergencyCategoryEntity optEmergency;
    private final List<CoverImportModel> importedImages;

    private UntiedLocalNote(String noteText, EmergencyCategoryEntity optEmergency, List<CoverImportModel> importedImages) {
        this.noteText = noteText;
        this.optEmergency = optEmergency;
        this.importedImages = Collections.unmodifiableList(importedImages);
    }

    public String getNoteText() {
        return noteText;
    }

    public EmergencyCategoryEntity getOptEmergency() {
        return optEmergency;
    }

    public List<CoverImportModel> getImportedImages() {
        return importedImages;
    }

    @NonNull
    @Override
    public String toString() {
        return String.format("{text=%s, Nimages=%d, emergency=%s}", noteText, importedImages.size(), (optEmergency == null) ? "No" : optEmergency.toString());
    }

    /************************  Parcelable implementation  *****************************************/
    @Override
    public int describeContents() {
        return 0;
    }

    @Override
    public void writeToParcel(Parcel dest, int flags) {
        Object[] fieldArray = new Object[3];
        fieldArray[0] = noteText;
        fieldArray[1] = optEmergency;
        fieldArray[2] = importedImages;
        dest.writeArray(fieldArray);
    }

    protected UntiedLocalNote(Parcel in) {
        Object[] fa = in.readArray(UntiedLocalNote.class.getClassLoader());
        noteText = (String)fa[0];
        optEmergency = (EmergencyCategoryEntity) fa[1];
        importedImages = (List<CoverImportModel>) fa[2];
    }

    public static final Creator<UntiedLocalNote> CREATOR = new Creator<UntiedLocalNote>() {
        public UntiedLocalNote createFromParcel(Parcel in) {
            return new UntiedLocalNote(in);
        }

        public UntiedLocalNote[] newArray(int size) {
            return new UntiedLocalNote[size];
        }
    };


    /**************************************  JSON Conversion  *************************************/
    public static UntiedLocalNote fromJsonObject(JSONObject jObj) throws JSONException {
        UntiedLocalNote.Builder bld = UntiedLocalNote.builder().setNoteText(jObj.getString("noteText"));
        if (jObj.has("optEmergency")) {
            long em_id = jObj.getLong("optEmergency");
            EmergencyCategoryEntity emerg = EmergencyHolder.getInstance().getEmergencyById(new LongId<>(em_id));
            bld.setEmergency(emerg);
        }

        JSONArray jImages = jObj.getJSONArray("importedImages");
        for (int i=0; i<jImages.length(); i++) {
            CoverImportModel img = CoverImportModel.fromJsonObject(jImages.getJSONObject(i));
            bld.addImage(img);
        }

        return bld.build();
    }

    public static UntiedLocalNote fromJson(String json) {
        try {
            return fromJsonObject(new JSONObject(json));
        } catch (JSONException e) {
            throw new RuntimeException(e);
        }
    }


    public JSONObject toJsonObject() throws JSONException {
        JSONArray jImages = new JSONArray();
        for (CoverImportModel img : importedImages) {
            jImages.put(img.toJsonObject());
        }
        JSONObject jObj = new JSONObject();
        jObj.put("noteText", noteText);
        if (optEmergency != null) jObj.put("optEmergency", optEmergency.id().getValue());
        jObj.put("importedImages", jImages);
        return jObj;
    }

    public String toJson() {
        try {
            return toJsonObject().toString();
        } catch (JSONException e) {
            throw new RuntimeException(e);
        }
    }

    /**************************************  Builder pattern  *************************************/
    public static Builder builder() {
        return new Builder();
    }


    public static class Builder {
        private String text;
        private EmergencyCategoryEntity emg;
        private List<CoverImportModel> images = new LinkedList<>();

        private Builder() {
            // Use static creator
        }

        public Builder setNoteText(String text) {
            this.text = text;
            return this;
        }

        public Builder setEmergency(@Nullable EmergencyCategoryEntity emg) {
            this.emg = emg;
            return this;
        }

        public Builder addImage(CoverImportModel img) {
            this.images.add(img);
            return this;
        }

        public UntiedLocalNote build() {
            if (TextUtils.isEmpty(text)) throw new IllegalStateException("No text");
            return new UntiedLocalNote(text, emg, images);
        }
    }

}
