package com.webssa.pda.sunmi.receipt;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.text.TextUtils;
import android.util.Base64;
import android.view.View;

import androidx.annotation.NonNull;

import com.webssa.library.widget.SignatureView;
import com.webssa.pda.sunmi.GlobalProperties;
import com.webssa.pda.sunmi.model.LocalCollectedGoodItem;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.receipt.model.ReceiptCollectedGoodModel;
import com.webssa.pda.sunmi.receipt.model.ReceiptCustomerModel;
import com.webssa.pda.sunmi.receipt.model.ReceiptDataMainModel;
import com.webssa.pda.sunmi.receipt.model.ReceiptPersonSignatureModel;
import com.webssa.pda.sunmi.receipt.model.ReceiptSiteModel;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.SiteEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.storage.dao.impl.PricedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.VisitDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.PricedGoodDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by stanislav.perchenko on 11/26/2019, 2:24 PM.
 */
public class ReceiptModelBuilder {

    private final Context context;
    private final LocalVisitResult visitResult;

    public ReceiptModelBuilder(Context context, @NonNull LocalVisitResult visitResult) {
        assert (context != null);
        assert (visitResult != null);
        this.context = context;
        this.visitResult = visitResult;
    }

    public ReceiptDataMainModel buildReceiptModel() throws IOException {
        VisitDbModel visit = loadVisit(visitResult.visitId());

        SiteEntity site = visit.site();
        StringBuilder sbSiteName = new StringBuilder(site.getSiteName());
        if (!TextUtils.isEmpty(site.getAddressLine1())) {
            sbSiteName.append(", ");
            sbSiteName.append(site.getAddressLine1());
        }
        if (!TextUtils.isEmpty(site.getAddressLine2())) {
            sbSiteName.append(", ");
            sbSiteName.append(site.getAddressLine2());
        }
        if (!TextUtils.isEmpty(site.getTown())) {
            sbSiteName.append(", ");
            sbSiteName.append(site.getTown());
        }
        if (!TextUtils.isEmpty(site.getCity())) {
            sbSiteName.append(", ");
            sbSiteName.append(site.getCity());
        }
        if (!TextUtils.isEmpty(site.getPostCode())) {
            sbSiteName.append(", ");
            sbSiteName.append(site.getPostCode());
        }

        final String resolvedShopCode = TextUtils.isEmpty(site.getShopCodeText()) ? (""+site.getCode()) : site.getShopCodeText();

        ReceiptSiteModel receiptSite = ReceiptSiteModel.builder()
                .setName(sbSiteName.toString())
                .setCode(resolvedShopCode)
                .build();


        ReceiptPersonSignatureModel signature = buildSignatureModelFromVectorPath(visitResult.customerSignature());

        ReceiptCustomerModel customer = ReceiptCustomerModel.builder()
                .setCustomerName(visitResult.customerName())
                .setCustomerSignature(signature)
                .build();

        List<ReceiptCollectedGoodModel> collectedGoods = new ArrayList<>(visitResult.collectedGoods().size());
        try (PricedGoodDAO dao = new PricedGoodDAO(new DbSelectionCache())) {
            for (LocalCollectedGoodItem good : visitResult.collectedGoods()) {
                ReceiptCollectedGoodModel rcgm = buildReceiptGood(dao, good);
                if (rcgm != null) collectedGoods.add(rcgm);
            }
        }

        String receiptNum = (visit.isAdHocVisit() && !visit.isRemoteCreated())
                ? String.format("%d-%s", visit.shiftId().getValue(), resolvedShopCode)
                : visit.id().toString();

        return ReceiptDataMainModel.builder()
                .setOrganization(GlobalProperties.ORGANIZATION)
                .setSite(receiptSite)
                .setReceiptNumber(receiptNum)
                .setReceiptTimestamp(System.currentTimeMillis())
                //.setPaymentMethod("ACC")
                .setPaymentMethod(site.getPaymentMethod())
                .setDriverName(visitResult.driverName())
                .setCustomer(customer)
                .setEwcCodes(GlobalProperties.EWC_CODES)
                .setCollectedGoods(collectedGoods)
                .build();
    }

    private ReceiptCollectedGoodModel buildReceiptGood(PricedGoodDAO dao, LocalCollectedGoodItem good) {
        PricedGoodDbModel goodCategory = dao.selectGoodForSite(good.getSiteId(), good.getGoodcategoryId());
        if (goodCategory == null) return null;

        return ReceiptCollectedGoodModel.builder()
                .setGoodName(goodCategory.categoryName())
                .setEwcCode(goodCategory.ewcCode().getCode())
                .setCollectedAmount(good.getWeightKg())
                .setPayedAmount(goodCategory.priceValue() * good.getWeightKg())
                .setPayedScale(goodCategory.priceScale())
                .build();
    }

    private VisitDbModel loadVisit(LongId<VisitEntity> visitId) {
        try (VisitDAO dao = new VisitDAO(new DbSelectionCache())) {
            return dao.selectEntryById(visitId);
        }
    }

    private ReceiptPersonSignatureModel buildSignatureModelFromVectorPath(String vectorPath) throws IOException {
        SignatureView sv = new SignatureView(context);
        sv.setSignColor(Color.BLACK);
        sv.setSignStrokeWidth(GlobalProperties.RECEIPT_SIGNATURE_STROKE_WIDTH);
        sv.setSignature(vectorPath);
        final int origW = sv.getOriginalSignatureWidth();
        final int origH = sv.getOriginalSignatureHeight();
        final int targW = GlobalProperties.RECEIPT_SIGNATURE_WIDTH_PX;
        final int targH = Math.round((float)(targW * origH) / origW);
        final Bitmap targBmp = Bitmap.createBitmap(targW, targH, Bitmap.Config.ARGB_8888);
        sv.measure(View.MeasureSpec.makeMeasureSpec(targW, View.MeasureSpec.EXACTLY), View.MeasureSpec.makeMeasureSpec(targH, View.MeasureSpec.EXACTLY));
        sv.layout(0, 0, targW, targH);
        Canvas canv = new Canvas(targBmp);
        canv.drawColor(Color.WHITE);
        sv.draw(canv);

        ByteArrayOutputStream bos = new ByteArrayOutputStream(1024);
        OutputStream os = new BufferedOutputStream(bos, 512);
        targBmp.compress(Bitmap.CompressFormat.PNG, 0/*ignored*/, os);
        os.flush();
        final byte[] encodedSignBitmap = Base64.encode(bos.toByteArray(), Base64.NO_WRAP | Base64.NO_PADDING);

        return ReceiptPersonSignatureModel.builder()
                .setW(targW)
                .setH(targH)
                .setSignData(new String(encodedSignBitmap, "UTF-8"))
                .build();
    }
}
