package com.webssa.pda.sunmi.rest;

import androidx.annotation.Nullable;

import com.webssa.library.attachment.model.CoverImportModel;
import com.webssa.library.core.error.AppError;
import com.webssa.library.core.error.IORestError;
import com.webssa.library.core.error.ParseRestError;
import com.webssa.library.core.error.ServerRestError;
import com.webssa.pda.sunmi.PDA_Application;
import com.webssa.pda.sunmi.EnvironmentRelatedConstants;
import com.webssa.pda.sunmi.model.UntiedLocalNote;
import com.webssa.pda.sunmi.rest.helper.HttpClientUtils;
import com.webssa.pda.sunmi.rest.helper.MyUrlBuilder;
import com.webssa.pda.sunmi.rest.helper.ParserProvider;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.EmergencyCategoryEntity;
import com.webssa.pda.sunmi.rest.model.impl.NoteEntity;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.storage.dbmodel.impl.LocalNoteDbModel;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.List;

import okhttp3.MediaType;
import okhttp3.MultipartBody;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;

/**
 * Created by stanislav.perchenko on 11/5/2019, 5:27 PM.
 */
public class PostNoteForShiftRestRequest {

    public static PostNoteForShiftRestRequest create(LongId<ShiftRestEntity> shiftId, UntiedLocalNote noteData) {
        LongId<EmergencyCategoryEntity> em_id = (noteData.getOptEmergency() == null) ? null : noteData.getOptEmergency().id();
        PostNoteForShiftRestRequest req = new PostNoteForShiftRestRequest(shiftId, noteData.getNoteText(), em_id, noteData.getImportedImages());
        req.setCreationDate(null);
        return req;
    }

    public static PostNoteForShiftRestRequest create(LongId<ShiftRestEntity> shiftId, LocalNoteDbModel localNote) {
        LongId<EmergencyCategoryEntity> em_id = (localNote.optEmergencyId() == null || localNote.optEmergencyId().getValue()
                == 0) ? null : localNote.optEmergencyId();
        PostNoteForShiftRestRequest req = new PostNoteForShiftRestRequest(shiftId, localNote.noteText(), em_id, localNote.attachedImages());
        req.setCreationDate(localNote.createdAt());
        return req;
    }

    private final LongId<ShiftRestEntity> shiftId;
    private final String text;
    private final LongId<EmergencyCategoryEntity> emergId;
    private final List<CoverImportModel> images;
    private Date optCreationDate;


    private NoteEntity result;
    private AppError lastError;


    private PostNoteForShiftRestRequest(LongId<ShiftRestEntity> shiftId, String text, LongId<EmergencyCategoryEntity> emerg, List<CoverImportModel> images) {
        this.shiftId = shiftId;
        this.text = text;
        this.emergId = emerg;
        this.images = images;
    }

    private synchronized void setCreationDate(@Nullable Date d) {
        optCreationDate = d;
    }

    public synchronized boolean execute(OkHttpClient httpClient) {
        MyUrlBuilder urlBld = new MyUrlBuilder(EnvironmentRelatedConstants.BACKEND_API_URL, "/note");

        Request request = new Request.Builder()
                .url(urlBld.build())
                .post(getBody())
                .build();

        //--- Execute request ---
        try {
            Response okHttpResponse = httpClient.newCall(request).execute();
            int httpCode = okHttpResponse.code();
            if (httpCode == 200) {
                String bodyText = okHttpResponse.body().string();
                try {
                    result = ParserProvider.getApplicationMoshi().adapter(NoteEntity.class).fromJson(bodyText);
                    return true;
                } catch (Exception e) {
                    lastError = ParseRestError.Companion.create(PDA_Application.getStaticAppContext(), e);
                }
            } else {
                lastError = ServerRestError.Companion.create(PDA_Application.getStaticAppContext(), httpCode, okHttpResponse.message());
            }
        } catch (IOException e) {
            lastError = IORestError.Companion.create(PDA_Application.getStaticAppContext(), e);
        }
        return false;
    }

    private RequestBody getBody() {
        MultipartBody.Builder bodyBld = new MultipartBody.Builder()
                .setType(MultipartBody.FORM)
                .addPart(MultipartBody.Part.createFormData("note_data", createNotePostJson(null)));

        for (CoverImportModel img : images) {
            MediaType imgMediaType = MediaType.parse(img.mimeType());
            File imgFile = new File(img.localFile());
            bodyBld.addPart(MultipartBody.Part.createFormData("note_image", img.relativeHashName(), RequestBody.create(imgMediaType, imgFile)));
        }
        return bodyBld.build();
    }

    private String createNotePostJson(@Nullable String partName) {
        try {
            JSONObject json = new JSONObject();
            json.put("text", text);
            json.put("shift_id", shiftId.getValue());
            if (emergId != null) {
                json.put("opt_emergency_id", emergId.getValue());
            }
            if (optCreationDate != null) {
                json.put("creation_time", HttpClientUtils.dateToServerTime(optCreationDate));
            }
            return (partName == null) ? json.toString() : String.format("%s:%s", partName, json.toString());
        } catch (JSONException e) {
            throw new RuntimeException(e);
        }
    }

    @Nullable
    public synchronized AppError getError() {
        return lastError;
    }

    @Nullable
    public synchronized NoteEntity getResultNote() {
        return result;
    }

}
