package com.webssa.pda.sunmi.storage.dao;

import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.rest.model.IdProvidingModel;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.storage.DbUtils;
import com.webssa.pda.sunmi.storage.dbhelper.DatabaseManager;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbhelper.RemoteEntrySaver;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Created by stanislav.perchenko on 10/18/2019.
 */
public abstract class BaseDAO</*Remote type*/ RT extends IdProvidingModel, /*Local type*/ LT extends IdProvidingModel> implements RemoteEntrySaver<RT>, Closeable {

    @org.jetbrains.annotations.Contract("_ -> fail")
    public static void createTables(SQLiteDatabase db) {
        throw new UnsupportedOperationException("This method must be re-implemented by ancestors");
    }

    @org.jetbrains.annotations.Contract("_ -> fail")
    public static void dropTables(SQLiteDatabase db) {
        throw new UnsupportedOperationException("This method must be re-implemented by ancestors");
    }

    protected final DbSelectionCache selectionCache;

    public abstract LT buildDbModel(Cursor c);
    protected abstract String getTableName();
    protected abstract String[] getSelectionColumns();
    protected abstract String getIdColumnName();

    private SQLiteDatabase writeableDatabase;
    private final List<BaseDAO> childrenDAO = new ArrayList<>();

    /**
     * No-args constructor to instantiate subclasses via reflection
     */
    public BaseDAO(){
        this(null);
    }

    public BaseDAO(@Nullable DbSelectionCache selectionCache) {
        this.selectionCache = selectionCache;
        writeableDatabase = DatabaseManager.getInstance().openDatabase();
    }

    protected synchronized SQLiteDatabase getWriteableDatabase() {
        return writeableDatabase;
    }

    public void beginTransaction() {
        getWriteableDatabase().beginTransaction();
    }

    public void setTransactionSuccessful() {
        getWriteableDatabase().setTransactionSuccessful();
    }

    public void endTransaction() {
        getWriteableDatabase().endTransaction();
    }

    /**
     * Call this from subclasses when DAO for internal objects are created
     * @param dao
     */
    protected synchronized void addChildDAO(BaseDAO dao) {
        childrenDAO.add(dao);
    }

    @Override
    public synchronized void close() {
        for (BaseDAO dao : childrenDAO) {
            dao.close();
        }
        childrenDAO.clear();
        writeableDatabase = null;
        DatabaseManager.getInstance().closeDatabase();
    }

    /**
     * This method is called from the deleteEntryById() before actual entry deleting.
     * This gives a chance for an actual implementation to delete dependencies;
     * @param id
     */
    protected void onEntryDeleted(LongId<RT> id) {
        //Override this method in subclases if necessary
    }

    public int removeAll() {
        return deleteEntryByWhere(null);
    }

    @Override
    public int deleteEntryById(LongId<RT> id) {
        onEntryDeleted(id);
        String where = String.format("%s = %d", getIdColumnName(), id.getValue());
        return writeableDatabase.delete(getTableName(), where, null);
    }

    @Override
    public final int createOrUpdateMultipleEntries(Collection<RT> entries) {
        int ret = 0;
        if (entries != null) {
            for (RT model : entries) {
                LongId<RT> id = createOrUpdateInLocalDB(model);
                if (id != null) ret ++;
            }
        }
        return ret;
    }

    public final LT selectEntryById(LongId<RT> id) {
        String where = String.format("%s = %d", getIdColumnName(), id.getValue());
        List<LT> models = selectEntriesForWhereAndSort(where, null, 0);
        return (models.size() > 0) ? models.get(0) : null;
    }

    //This is not used in this project
    /*public final List<LT> selectEntriesByIdsNoSort(LongId<RT>... ids) {
        long[] l_ids = new long[ids.length];
        for (int i=0; i<ids.length; i++) l_ids[i] = ids[i].getValue();
        String where = DbUtils.buildWhereSectionIN(getIdColumnName(), l_ids);
        return selectEntriesForWhereAndSort(where, null, 0);
    }

    public final List<LT> selectEntriesByIdsNoSort(List<LongId<RT>> ids) {
        long[] l_ids = new long[ids.size()];
        for (int i=0; i<ids.size(); i++) l_ids[i] = ids.get(i).getValue();
        String where = DbUtils.buildWhereSectionIN(getIdColumnName(), l_ids);
        return selectEntriesForWhereAndSort(where, null, 0);
    }

    public final List<LT> selectAllEntriesWithSort(String sort) {
        return selectEntriesForWhereAndSort(null, sort, 0);
    }*/

    /**
     *
     * @param where
     * @param sort
     * @param limit 0 and negative values means no limit
     * @return
     */
    protected final List<LT> selectEntriesForWhereAndSort(String where, String sort, int limit) {
        try (Cursor c = writeableDatabase.query(getTableName(), getSelectionColumns(), where, null, null, null, sort, (limit > 0) ? ""+limit : null)) {
            List<LT> selection = new ArrayList<>();
            if (c.moveToFirst()) {
                do {
                    selection.add(buildDbModel(c));
                } while (c.moveToNext());
            }
            return selection;
        }
    }

    protected final int getEntryNumberForWhere(@NonNull String where) {
        Cursor c = writeableDatabase.query(getTableName(), getSelectionColumns(), where, null, null, null, null);
        try {
            return (c.moveToFirst()) ? c.getCount() : 0;
        } finally {
            c.close();
        }
    }

    protected final int deleteEntryByWhere(@NonNull String where) {
        return writeableDatabase.delete(getTableName(), where, null);
    }

    protected final int deleteAll() {
        return writeableDatabase.delete(getTableName(), null, null);
    }

}
