package com.webssa.pda.sunmi.storage.dao.impl;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.EmergencyCategoryEntity;
import com.webssa.pda.sunmi.rest.model.impl.NoteEntity;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.storage.dao.BaseDAO;
import com.webssa.pda.sunmi.storage.dbmodel.impl.NoteDbModel;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Set;

/**
 * Created by stanislav.perchenko on 11/6/2019, 10:10 AM.
 */
public class NoteDAO extends BaseDAO<NoteEntity, NoteDbModel> {

    public static final String TABLE_NAME = "Notes";

    public static final String COLUMN_ID = "_id";
    public static final String COLUMN_SHIFT_ID = "shift_id";
    public static final String COLUMN_VISIT_ID = "visit_id";
    public static final String COLUMN_SEQ_NUM = "seq_num";
    public static final String COLUMN_AUTHOR = "author";
    public static final String COLUMN_CREATED_AT = "created_at";
    public static final String COLUMN_TEXT = "text";
    public static final String COLUMN_EMERG_ID = "emerg_id";
    public static final String COLUMN_IMAGES = "images";

    public static final String[] SELECTION_COLUMNS = {COLUMN_ID, COLUMN_SHIFT_ID, COLUMN_VISIT_ID, COLUMN_SEQ_NUM, COLUMN_AUTHOR, COLUMN_CREATED_AT, COLUMN_TEXT, COLUMN_EMERG_ID, COLUMN_IMAGES};

    private static final String[] CREATIONS = new String[]{
            String.format("CREATE TABLE %s (%s INTEGER PRIMARY KEY, %s INTEGER, %s INTEGER, %s INTEGER, %s TEXT, %s INTEGER, %s TEXT, %s INTEGER, %s TEXT);",
                    TABLE_NAME,
                    COLUMN_ID,
                    COLUMN_SHIFT_ID,
                    COLUMN_VISIT_ID,
                    COLUMN_SEQ_NUM,
                    COLUMN_AUTHOR,
                    COLUMN_CREATED_AT,
                    COLUMN_TEXT,
                    COLUMN_EMERG_ID,
                    COLUMN_IMAGES)
    };

    private static final String[] DELETIONS = new String[] {
            String.format("DROP TABLE IF EXISTS %s", TABLE_NAME)
    };

    public static void createTables(SQLiteDatabase db) {
        for (String cr : CREATIONS) db.execSQL(cr);
    }

    public static void dropTables(SQLiteDatabase db) {
        for (String del : DELETIONS) db.execSQL(del);
    }

    public NoteDAO() {
        super();
    }

    @Override
    protected String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected String[] getSelectionColumns() {
        return SELECTION_COLUMNS;
    }

    @Override
    protected String getIdColumnName() {
        return COLUMN_ID;
    }

    @Override
    public LongId<NoteEntity> createOrUpdateInLocalDB(NoteEntity entry) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_SHIFT_ID, entry.getShiftId().getValue());
        //if (entry.getVisitId() != null) vvv.put(COLUMN_VISIT_ID, entry.getVisitId().getValue());
        vvv.put(COLUMN_VISIT_ID, (entry.getVisitId() == null) ? 0 : entry.getVisitId().getValue());
        vvv.put(COLUMN_SEQ_NUM, entry.getSequenceNumber());
        vvv.put(COLUMN_AUTHOR, entry.getAuthorId());
        vvv.put(COLUMN_CREATED_AT, entry.getTimeCreated().getTime());
        vvv.put(COLUMN_TEXT, entry.getText());
        vvv.put(COLUMN_IMAGES, getImageJson(entry.getImagesUrlSet()));
        if (entry.getOptEmergencyId() != null) vvv.put(COLUMN_EMERG_ID, entry.getOptEmergencyId().getValue());

        String where = COLUMN_ID + " = " + entry.id().getValue();
        if (getWriteableDatabase().update(TABLE_NAME, vvv, where, null) == 0) {
            vvv.put(COLUMN_ID, entry.id().getValue());
            long id = getWriteableDatabase().insert(TABLE_NAME, null, vvv);
            return (id > 0) ? LongId.Companion.valueOf(id) : null;
        } else {
            return entry.id();
        }
    }

    private String getImageJson(Set<String> images) {
        JSONArray json = new JSONArray();
        for (String img : images) {
            json.put(img);
        }
        return json.toString();
    }

    @Override
    public int deleteEntry(NoteEntity entry) {
        return super.deleteEntryById(entry.id());
    }

    @Override
    public NoteDbModel buildDbModel(Cursor c) {
        LongId<NoteEntity> id = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_ID)));
        LongId<ShiftRestEntity> shiftId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_SHIFT_ID)));
        LongId<VisitEntity> optVisitId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_VISIT_ID)));
        LongId<EmergencyCategoryEntity> optEmergencyId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_EMERG_ID)));

        List<String> images;
        try {
            JSONArray json = new JSONArray(c.getString(c.getColumnIndex(COLUMN_IMAGES)));
            String[] sss = new String[json.length()];
            for (int i = 0; i < sss.length; i++) sss[i] = json.getString(i);
            images = Arrays.asList(sss);
        } catch (JSONException e) {
            images = Arrays.asList();
        }

        return NoteDbModel.builder()
                .setId(id)
                .setShiftId(shiftId)
                .setOptVisitId(optVisitId)
                .setSequenceNumber(c.getInt(c.getColumnIndex(COLUMN_SEQ_NUM)))
                .setAuthorId(c.getString(c.getColumnIndex(COLUMN_AUTHOR)))
                .setCreatedAt(new Date(c.getLong(c.getColumnIndex(COLUMN_CREATED_AT))))
                .setNoteText(c.getString(c.getColumnIndex(COLUMN_TEXT)))
                .setOptEmergencyId(optEmergencyId)
                .setAttachedImages(images)
                .build();
    }


    public List<NoteDbModel> selectNotesForShiftOnly(LongId<ShiftRestEntity> shiftId) {
        String where = String.format("%s = %d AND %s < 1", COLUMN_SHIFT_ID, shiftId.getValue(), COLUMN_VISIT_ID);
        return selectEntriesForWhereAndSort(where, COLUMN_CREATED_AT+" ASC", 0);
    }

    @Nullable
    public NoteDbModel selectNoteForVisit(LongId<ShiftRestEntity> shiftId, LongId<VisitEntity> visitId) {
        String where = String.format("(%s = %d) AND (%s = %d)", COLUMN_SHIFT_ID, shiftId.getValue(), COLUMN_VISIT_ID, visitId.getValue());
        List<NoteDbModel> data = selectEntriesForWhereAndSort(where, null, 1);
        return data.isEmpty() ? null : data.get(0);
    }

    public void clearAllNotes() {
        //todo
        //deleteAll();
    }
}
