package com.webssa.pda.sunmi.storage.dao.impl;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.EWCCodeEntity;
import com.webssa.pda.sunmi.rest.model.impl.PriceEntity;
import com.webssa.pda.sunmi.rest.model.impl.PricedGoodEntity;
import com.webssa.pda.sunmi.rest.model.impl.SiteEntity;
import com.webssa.pda.sunmi.storage.DbUtils;
import com.webssa.pda.sunmi.storage.dao.BaseDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.PricedGoodDbModel;

import java.util.Date;
import java.util.List;

/**
 * Created by stanislav.perchenko on 10/18/2019.
 */
public class PricedGoodDAO extends BaseDAO<PricedGoodEntity, PricedGoodDbModel> {
    public static final String TABLE_NAME = "Pricedgoods";

    public static final String COLUMN_ID = "_id";                       //This is "goodcategory_id" from the API
    public static final String COLUMN_SITE_ID = "site_id";
    public static final String COLUMN_CATEGORY_NAME = "good_name";
    public static final String COLUMN_CATEGORY_ORDER = "good_order";
    public static final String COLUMN_PRICE_ID = "price_id";
    public static final String COLUMN_PRICE_VALUE = "price_value";
    public static final String COLUMN_PRICE_SCALE = "price_scale";
    public static final String COLUMN_PRICE_SET_DATE = "price_set_date";
    public static final String COLUMN_EWC_ID = "ewc_id";
    public static final String COLUMN_MEASURE_UNIT = "m_unit";

    public static final String[] SELECTION_COLUMNS = {
            COLUMN_ID, COLUMN_SITE_ID, COLUMN_CATEGORY_NAME,
            COLUMN_CATEGORY_ORDER, COLUMN_PRICE_ID, COLUMN_PRICE_VALUE,
            COLUMN_PRICE_SCALE, COLUMN_PRICE_SET_DATE, COLUMN_EWC_ID,
            COLUMN_MEASURE_UNIT };

    private static final String[] CREATIONS = new String[]{
    //                                        good_categ_id     site_id       good_name   good_order    price_id      price_value   price_scale   price_set_date    ewc_id         m_unit
            String.format("CREATE TABLE %1$s (%2$s INTEGER,     %3$s INTEGER,  %4$s TEXT, %5$s INTEGER, %6$s INTEGER, %7$s INTEGER, %8$s INTEGER, %9$s INTEGER,     %10$s INTEGER, %11$s TEXT, PRIMARY KEY(%2$s, %3$s));",
                    TABLE_NAME,
                    COLUMN_ID,
                    COLUMN_SITE_ID,
                    COLUMN_CATEGORY_NAME,
                    COLUMN_CATEGORY_ORDER,
                    COLUMN_PRICE_ID,
                    COLUMN_PRICE_VALUE,
                    COLUMN_PRICE_SCALE,
                    COLUMN_PRICE_SET_DATE,
                    COLUMN_EWC_ID,
                    COLUMN_MEASURE_UNIT)
    };

    private static final String[] DELETIONS = new String[]{
            String.format("DROP TABLE IF EXISTS %s", TABLE_NAME)
    };

    public static void createTables(SQLiteDatabase db) {
        for (String cr : CREATIONS) db.execSQL(cr);
    }

    public static void dropTables(SQLiteDatabase db) {
        for (String del : DELETIONS) db.execSQL(del);
    }

    public PricedGoodDAO(@Nullable DbSelectionCache selectionCache) {
        super(selectionCache);
    }

    @Override
    protected String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected String[] getSelectionColumns() {
        return SELECTION_COLUMNS;
    }

    @Override
    protected String getIdColumnName() {
        return COLUMN_ID;
    }

    @Override
    public LongId<PricedGoodEntity> createOrUpdateInLocalDB(PricedGoodEntity entry) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_CATEGORY_NAME, entry.getCategoryName());
        vvv.put(COLUMN_CATEGORY_ORDER, entry.getCategoryOrder());
        vvv.put(COLUMN_PRICE_ID, entry.getPriceId().getValue());
        vvv.put(COLUMN_PRICE_VALUE, entry.getPriceValue());
        vvv.put(COLUMN_PRICE_SCALE, entry.getPriceScale());
        vvv.put(COLUMN_PRICE_SET_DATE, entry.getPriceSetDate().getTime());
        vvv.put(COLUMN_EWC_ID, entry.getEwcCodeId().getValue());
        if (entry.getUnitOfMeasure() != null) vvv.put(COLUMN_MEASURE_UNIT, entry.getUnitOfMeasure());

        String where = String.format("(%s = %d AND %s = %d)", COLUMN_ID, entry.id().getValue(), COLUMN_SITE_ID, entry.getSiteId().getValue());
        if (getWriteableDatabase().update(TABLE_NAME, vvv, where, null) == 0) {
            vvv.put(COLUMN_ID, entry.id().getValue());
            vvv.put(COLUMN_SITE_ID, entry.getSiteId().getValue());
            long id = getWriteableDatabase().insert(TABLE_NAME, null, vvv);
            return (id > 0) ? LongId.Companion.valueOf(id) : null;
        } else {
            return entry.id();
        }
    }

    @Override
    public int deleteEntry(PricedGoodEntity entry) {
        return super.deleteEntryById(entry.id());
    }

    @Override
    public PricedGoodDbModel buildDbModel(Cursor c) {
        LongId<PricedGoodEntity> id = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_ID)));
        LongId<PriceEntity> priceId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_PRICE_ID)));
        LongId<EWCCodeEntity> ewcId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_EWC_ID)));
        long time = c.getLong(c.getColumnIndex(COLUMN_PRICE_SET_DATE));

        return PricedGoodDbModel.builder()
                .setId(id)
                .setCategoryName(c.getString(c.getColumnIndex(COLUMN_CATEGORY_NAME)))
                .setCategoryOrder(c.getInt(c.getColumnIndex(COLUMN_CATEGORY_ORDER)))
                .setPriceId(priceId)
                .setPriceValue(c.getInt(c.getColumnIndex(COLUMN_PRICE_VALUE)))
                .setPriceScale(c.getInt(c.getColumnIndex(COLUMN_PRICE_SCALE)))
                .setPriceSetDate(new Date(time))
                .setEwcCode(selectEwcCode(ewcId))
                .setUnitOfMeasure(c.getString(c.getColumnIndex(COLUMN_MEASURE_UNIT)))
                .setSiteId(LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_SITE_ID))))
                .build();
    }

    private EWCCodeEntity selectEwcCode(LongId<EWCCodeEntity> ewcId) {
        EWCCodeEntity ewc = (selectionCache == null) ? null : selectionCache.getEwcCodeEntity(ewcId);
        if (ewc == null) {
            ewc = getEwcDao().selectEntryById(ewcId);
            if ((selectionCache != null) && (ewc != null)) selectionCache.putDbModel(ewc);
        }
        return ewc;
    }

    private EWCCodeDAO ewcDao;

    private EWCCodeDAO getEwcDao() {
        if (ewcDao == null) addChildDAO(ewcDao = new EWCCodeDAO());
        return ewcDao;
    }

    public List<PricedGoodDbModel> selectGoodsBySiteId(LongId<SiteEntity> siteId) {
        final String where = String.format("%s = %d", COLUMN_SITE_ID, siteId.getValue());
        final String sort = COLUMN_CATEGORY_ORDER + " ASC";
        return selectEntriesForWhereAndSort(where, sort, 0);
    }

    @Nullable
    public PricedGoodDbModel selectGoodForSite(LongId<SiteEntity> siteId, LongId<PricedGoodEntity> goodcategoryId) {
        final String where = String.format("(%s = %d) AND (%s = %d)", COLUMN_SITE_ID, siteId.getValue(), COLUMN_ID, goodcategoryId.getValue());
        final String sort = COLUMN_CATEGORY_ORDER + " ASC";
        List<PricedGoodDbModel> rrr = selectEntriesForWhereAndSort(where, sort, 0);
        return rrr.isEmpty() ? null : rrr.get(0);
    }

    public List<PricedGoodDbModel> selectGoodsByIdsForSite(LongId<SiteEntity> siteId, List<LongId<PricedGoodEntity>> goodIds) {
        long[] l_ids = new long[goodIds.size()];
        for (int i=0; i<goodIds.size(); i++) l_ids[i] = goodIds.get(i).getValue();
        String where_in = DbUtils.buildWhereSectionIN(getIdColumnName(), l_ids);
        String where = String.format("(%s = %d) AND %s", COLUMN_SITE_ID, siteId.getValue(), where_in);
        return selectEntriesForWhereAndSort(where, null, 0);
    }
}
