package com.webssa.pda.sunmi.storage.dao.impl;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.ShiftState;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.storage.dao.BaseDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.ShiftDbModel;
import com.webssa.pda.sunmi.util.DateNoTime;
import com.webssa.pda.sunmi.util.DurationHMS;
import com.webssa.pda.sunmi.util.TimeOfDayHMS;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * Created by stanislav.perchenko on 10/21/2019.
 */
public class ShiftDAO extends BaseDAO<ShiftRestEntity, ShiftDbModel> {

    public static final String TABLE_NAME = "Shifts";

    public static final String COLUMN_ID = "_id";
    public static final String COLUMN_DRIVER_ID = "driver_id";
    public static final String COLUMN_STATE = "state";
    public static final String COLUMN_EXEC_DATE = "exec_date";
    public static final String COLUMN_SCHEDULE_T_START = "schedul_t_start";
    public static final String COLUMN_SCHEDULE_DURATION = "schedul_duration";
    public static final String COLUMN_N_VISITS = "n_visits";
    public static final String COLUMN_N_COMPL_VISITS = "n_compl_visits";
    public static final String COLUMN_N_FAILED_VISITS = "n_failed_visits";
    public static final String COLUMN_ACT_T_START = "act_t_start";
    public static final String COLUMN_ACT_T_CLOSED = "act_t_closed";


    public static final String[] SELECTION_COLUMNS = new String[]{COLUMN_ID, COLUMN_DRIVER_ID, COLUMN_STATE, COLUMN_EXEC_DATE, COLUMN_SCHEDULE_T_START, COLUMN_SCHEDULE_DURATION, COLUMN_N_VISITS, COLUMN_N_COMPL_VISITS, COLUMN_N_FAILED_VISITS, COLUMN_ACT_T_START, COLUMN_ACT_T_CLOSED};

    private static final String[] CREATIONS = new String[]{
            //                                    id                driver_id  state    exec_date  t_start  duration  n_visits    n_compl_vis  n_failed_vis  act_t_start  act_t_closed
            String.format("CREATE TABLE %s (%s INTEGER PRIMARY KEY, %s TEXT,   %s TEXT, %s TEXT,   %s TEXT, %s TEXT,  %s INTEGER, %s INTEGER,  %s INTEGER,   %s INTEGER,  %s INTEGER);",
                    TABLE_NAME,
                    COLUMN_ID,
                    COLUMN_DRIVER_ID,
                    COLUMN_STATE,
                    COLUMN_EXEC_DATE,
                    COLUMN_SCHEDULE_T_START,
                    COLUMN_SCHEDULE_DURATION,
                    COLUMN_N_VISITS,
                    COLUMN_N_COMPL_VISITS,
                    COLUMN_N_FAILED_VISITS,
                    COLUMN_ACT_T_START,
                    COLUMN_ACT_T_CLOSED)
    };

    private static final String[] DELETIONS = new String[] {
            String.format("DROP TABLE IF EXISTS %s", TABLE_NAME)
    };

    public static void createTables(SQLiteDatabase db) {
        for (String cr : CREATIONS) db.execSQL(cr);
    }

    public static void dropTables(SQLiteDatabase db) {
        for (String del : DELETIONS) db.execSQL(del);
    }

    public ShiftDAO(@Nullable DbSelectionCache selectionCache) {
        super(selectionCache);
    }

    @Override
    protected String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected String[] getSelectionColumns() {
        return SELECTION_COLUMNS;
    }

    @Override
    protected String getIdColumnName() {
        return COLUMN_ID;
    }

    @Override
    public LongId<ShiftRestEntity> createOrUpdateInLocalDB(ShiftRestEntity entry) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_DRIVER_ID, entry.getDriverId());
        vvv.put(COLUMN_STATE, entry.getState().name());
        vvv.put(COLUMN_EXEC_DATE, entry.getExecutionDate().toString());
        vvv.put(COLUMN_SCHEDULE_T_START, (entry.getScheduledTimeStart() == null) ? null : entry.getScheduledTimeStart().toString());
        vvv.put(COLUMN_SCHEDULE_DURATION, (entry.getScheduledDuration() == null) ? null : entry.getScheduledDuration().toString());
        vvv.put(COLUMN_N_VISITS, entry.getNVisits());
        vvv.put(COLUMN_N_COMPL_VISITS, entry.getNCompletedVisits());
        vvv.put(COLUMN_N_FAILED_VISITS, entry.getNFailedVisits());
        vvv.put(COLUMN_ACT_T_START, (entry.getActualTimeStart() == null) ? 0 : entry.getActualTimeStart().getTime());
        vvv.put(COLUMN_ACT_T_CLOSED, (entry.getActualTimeClosed() == null) ? 0 : entry.getActualTimeClosed().getTime());

        String where = COLUMN_ID + " = " + entry.id().getValue();
        if (getWriteableDatabase().update(TABLE_NAME, vvv, where, null) == 0) {
            vvv.put(COLUMN_ID, entry.id().getValue());
            long id = getWriteableDatabase().insert(TABLE_NAME, null, vvv);
            return (id > 0) ? LongId.Companion.valueOf(id) : null;
        } else {
            return entry.id();
        }
    }

    @Override
    public int deleteEntry(ShiftRestEntity entry) {
        return super.deleteEntryById(entry.id());
    }

    @Override
    public ShiftDbModel buildDbModel(Cursor c) {

        final LongId<ShiftRestEntity> id = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_ID)));
        final long actStart = c.getLong(c.getColumnIndex(COLUMN_ACT_T_START));
        final long actClosed = c.getLong(c.getColumnIndex(COLUMN_ACT_T_CLOSED));
        final String schedTStart = c.getString(c.getColumnIndex(COLUMN_SCHEDULE_T_START));
        final String schedDur = c.getString(c.getColumnIndex(COLUMN_SCHEDULE_DURATION));

        try {
            return ShiftDbModel
                    .builder()
                    .setId(id)
                    .setDriverId(c.getString(c.getColumnIndex(COLUMN_DRIVER_ID)))
                    .setState(ShiftState.valueOf(c.getString(c.getColumnIndex(COLUMN_STATE))))
                    .setExecutionDate(DateNoTime.parseDateNoTime(c.getString(c.getColumnIndex(COLUMN_EXEC_DATE))))
                    .setScheduledTimeStart(TimeOfDayHMS.parseTime(schedTStart))
                    .setScheduledDuration(DurationHMS.parseDurationHMS(schedDur))
                    .setNTotalVisitsBackend(c.getInt(c.getColumnIndex(COLUMN_N_VISITS)))
                    .setNCompletedVisits(c.getInt(c.getColumnIndex(COLUMN_N_COMPL_VISITS)))
                    .setNFailedVisits(c.getInt(c.getColumnIndex(COLUMN_N_FAILED_VISITS)))
                    .setActualTimeStart((actStart > 0) ? new Date(actStart) : null)
                    .setActualTimeClosed((actClosed > 0) ? new Date(actClosed) : null)
                    .setVisits(getVisitDao().selectVisistForShiftSortedByOrder(id))
                    .setRemoteNotes(getNoteDao().selectNotesForShiftOnly(id))
                    .setLocalNotes(getLocalNoteDao().selectNotesForShiftOnly(id))
                    .build();
        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }


    private VisitDAO visDao;
    private VisitDAO getVisitDao() {
        if (visDao == null) addChildDAO(visDao = new VisitDAO(selectionCache));
        return visDao;
    }

    private NoteDAO noteDao;
    private NoteDAO getNoteDao() {
        if (noteDao == null) addChildDAO(noteDao = new NoteDAO());
        return noteDao;
    }

    private LocalNoteDAO localNoteDao;
    private LocalNoteDAO getLocalNoteDao() {
        if (localNoteDao == null) addChildDAO(localNoteDao = new LocalNoteDAO());
        return localNoteDao;
    }



    @Nullable
    public ShiftDbModel getShiftForDate(DateNoTime date) {
        final String where = String.format("%s = '%s'", COLUMN_EXEC_DATE, date.toString());
        final String sort  = COLUMN_ID + " DESC";
        List<ShiftDbModel> sss = selectEntriesForWhereAndSort(where, sort, 1);
        return sss.isEmpty() ? null : sss.get(0);
    }

    @Nullable
    public LongId<ShiftRestEntity> getActiveShiftId(DateNoTime date) {
        final String sort  = COLUMN_ID + " DESC";
        try (Cursor c = getWriteableDatabase().query(TABLE_NAME, new String[]{COLUMN_ID, COLUMN_STATE}, null, null, null, null, sort)) {
            if (c.moveToFirst()) {
                do {
                    ShiftState state = ShiftState.valueOf(c.getString(c.getColumnIndex(COLUMN_STATE)));
                    if (state == ShiftState.ACCEPTED) {
                        return LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_ID)));
                    }
                } while (c.moveToNext());
            }
        }
        return null;
    }

    @Nullable
    public DateNoTime selectExecutionDateForShift(LongId<ShiftRestEntity> shiftId) {
        final String where = String.format("%s = %d", COLUMN_ID, shiftId.getValue());
        try (Cursor c = getWriteableDatabase().query(TABLE_NAME, new String[]{COLUMN_EXEC_DATE}, where, null, null, null, null)) {
            if (c.moveToFirst()) {
                return DateNoTime.parseDateNoTime(c.getString(0));
            } else {
                return null;
            }
        } catch (ParseException e) {
            return null;
        }
    }
}
