package com.webssa.pda.sunmi.storage.dao.impl;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

import androidx.annotation.Nullable;

import com.webssa.pda.sunmi.model.VisitCreationModel;
import com.webssa.pda.sunmi.rest.helper.ParserProvider;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.VisitState;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.rest.model.impl.SiteEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitStateEntity;
import com.webssa.pda.sunmi.storage.dao.BaseDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;
import com.webssa.pda.sunmi.util.TimeOfDayHMS;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.IOException;
import java.text.ParseException;
import java.util.Date;
import java.util.List;

/**
 * Created by stanislav.perchenko on 10/21/2019.
 */
public class VisitDAO extends BaseDAO<VisitEntity, VisitDbModel> {

    public static final String TABLE_NAME = "Visits";

    public static final String COLUMN_ID = "_id";
    public static final String COLUMN_SHIFT_ID = "shift_id";
    public static final String COLUMN_SITE_ID = "site_id";
    public static final String COLUMN_ORDER = "o_r_d_e_r";
    public static final String COLUMN_STATE = "state";
    public static final String COLUMN_SUGGEST_TIME = "suggest_time";
    public static final String COLUMN_ACTUAL_START_TIME = "act_start_t";
    public static final String COLUMN_ACTUAL_COMPLETE_TIME = "act_compl_t";
    public static final String COLUMN_CUSTOMER_NAME = "customer_name";
    public static final String COLUMN_CUSTOMER_SIGNATURE = "customer_sign";
    public static final String COLUMN_CLEAR_BAGS = "CLEAR_BAGS";
    public static final String COLUMN_WHITE_SACKS = "white_sacks";
    public static final String COLUMN_TOT_PAY_AMOUNT = "tot_pay_amount";
    public static final String COLUMN_TOT_PAY_SCALE = "tot_pay_scale";
    public static final String COLUMN_IS_SYNCED = "is_synced";
    public static final String COLUMN_IS_AD_HOC = "is_ad_hoc";
    public static final String COLUMN_IS_REMOTE_CREATED = "is_remote_created";

    private static final String COLUMN_SUGGEST_TIME_MILLIS = "suggest_time_millis"; //This column is used internally to order query result. It is backed by the COLUMN_SUGGEST_TIME

    public static final String[] SELECTION_COLUMNS = new String[]{COLUMN_ID,
            COLUMN_SHIFT_ID, COLUMN_SITE_ID, COLUMN_ORDER, COLUMN_STATE,
            COLUMN_SUGGEST_TIME, COLUMN_ACTUAL_START_TIME,
            COLUMN_ACTUAL_COMPLETE_TIME, COLUMN_CUSTOMER_NAME,
            COLUMN_CUSTOMER_SIGNATURE, COLUMN_WHITE_SACKS, COLUMN_CLEAR_BAGS,
            COLUMN_TOT_PAY_AMOUNT, COLUMN_TOT_PAY_SCALE,
            COLUMN_IS_SYNCED, COLUMN_IS_AD_HOC, COLUMN_IS_REMOTE_CREATED, COLUMN_SUGGEST_TIME_MILLIS };

    private static final String[] CREATIONS = new String[] {
            //                                    id                shift_id    site_id     order       state    suggest_t   act_start_t  act_compl_t  cust_name   cust_sign  tot_amount  tot_scale   is_synced   clead_bags  white_sacks  is_ad_hoc  is_rem_created   suggest_t_ms
            String.format("CREATE TABLE %s (%s INTEGER PRIMARY KEY, %s INTEGER, %s INTEGER, %s INTEGER, %s TEXT, %s TEXT,    %s INTEGER,  %s INTEGER,  %s TEXT,    %s TEXT,   %s INTEGER, %s INTEGER, %s INTEGER, %s INTEGER, %s INTEGER, %s INTEGER, %s INTEGER,      %s INTEGER );",
                    TABLE_NAME,
                    COLUMN_ID,
                    COLUMN_SHIFT_ID,
                    COLUMN_SITE_ID,
                    COLUMN_ORDER,
                    COLUMN_STATE,
                    COLUMN_SUGGEST_TIME,
                    COLUMN_ACTUAL_START_TIME,
                    COLUMN_ACTUAL_COMPLETE_TIME,
                    COLUMN_CUSTOMER_NAME,
                    COLUMN_CUSTOMER_SIGNATURE,
                    COLUMN_TOT_PAY_AMOUNT,
                    COLUMN_TOT_PAY_SCALE,
                    COLUMN_IS_SYNCED,
                    COLUMN_CLEAR_BAGS,
                    COLUMN_WHITE_SACKS,
                    COLUMN_IS_AD_HOC,
                    COLUMN_IS_REMOTE_CREATED,
                    COLUMN_SUGGEST_TIME_MILLIS)
    };

    private static final String[] DELETIONS = new String[]{
            String.format("DROP TABLE IF EXISTS %s", TABLE_NAME)
    };

    public static void createTables(SQLiteDatabase db) {
        for (String cr : CREATIONS) db.execSQL(cr);
    }

    public static void dropTables(SQLiteDatabase db) {
        for (String del : DELETIONS) db.execSQL(del);
    }

    public VisitDAO(@Nullable DbSelectionCache selectionCache) {
        super(selectionCache);
    }

    @Override
    protected String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected String[] getSelectionColumns() {
        return SELECTION_COLUMNS;
    }

    @Override
    protected String getIdColumnName() {
        return COLUMN_ID;
    }

    @Override
    public LongId<VisitEntity> createOrUpdateInLocalDB(VisitEntity entry) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_SHIFT_ID, entry.getShiftId().getValue());
        vvv.put(COLUMN_SITE_ID, entry.getSiteId().getValue());
        vvv.put(COLUMN_ORDER, entry.getOrder());
        vvv.put(COLUMN_STATE, entry.getState().name());
        if (entry.getSuggestedTime() != null) {
            vvv.put(COLUMN_SUGGEST_TIME, entry.getSuggestedTime().toString());
            vvv.put(COLUMN_SUGGEST_TIME_MILLIS, entry.getSuggestedTime().getTimeMillis());
        } else {
            vvv.put(COLUMN_SUGGEST_TIME_MILLIS, -1);
        }
        vvv.put(COLUMN_ACTUAL_START_TIME, (entry.getActualStartTime() == null) ? 0 : entry.getActualStartTime().getTime());
        vvv.put(COLUMN_ACTUAL_COMPLETE_TIME, (entry.getActualCompleteTime() == null) ? 0 : entry.getActualCompleteTime().getTime());
        vvv.put(COLUMN_CUSTOMER_NAME, entry.getCustomerName());
        vvv.put(COLUMN_CUSTOMER_SIGNATURE, entry.getCustomerSignature());
        vvv.put(COLUMN_CLEAR_BAGS, entry.getClear_bags());
        vvv.put(COLUMN_WHITE_SACKS, entry.getWhite_sacks());
        vvv.put(COLUMN_TOT_PAY_AMOUNT, entry.getTotalPayedAmount());
        vvv.put(COLUMN_TOT_PAY_SCALE, entry.getTotalPayedScale());
        vvv.put(COLUMN_IS_SYNCED, ((entry.isLocalCreated() == null) || !entry.isLocalCreated()) ? 1 : 0);
        vvv.put(COLUMN_IS_AD_HOC, entry.isAdHocVisit());
        vvv.put(COLUMN_IS_REMOTE_CREATED, ((entry.isLocalCreated() == null) || !entry.isLocalCreated()) ? 1 : 0);

        String where = COLUMN_ID + " = " + entry.id().getValue();
        if (getWriteableDatabase().update(TABLE_NAME, vvv, where, null) == 0) {
            vvv.put(COLUMN_ID, entry.id().getValue());
            long id = getWriteableDatabase().insert(TABLE_NAME, null, vvv);
            return (id > 0) ? LongId.Companion.valueOf(id) : null;
        } else {
            return entry.id();
        }
    }


    public VisitDbModel createAdHocVisitLocally(VisitCreationModel creationModel) throws JSONException, IOException {
        long fakeVisitId = (creationModel.getShiftId().getValue() << 32) | (creationModel.getSiteId().getValue() & 0xFFFFFFL);
        int order = 1 + selectMaxVisitOrderForShift(creationModel.getShiftId());
        JSONObject jVisit = new JSONObject();
        jVisit.put("visit_id", fakeVisitId);
        jVisit.put("shift_id", creationModel.getShiftId().getValue());
        jVisit.put("site_id", creationModel.getSiteId().getValue());
        jVisit.put("order", order);
        jVisit.put("suggest_time", creationModel.getSuggestTime().toString());
        jVisit.put("start_time", null);
        jVisit.put("complete_time", null);
        jVisit.put("state", "new");
        jVisit.put("customer_name", null);
        jVisit.put("customer_signature", null);
        jVisit.put("total_payed_amount", 0);
        jVisit.put("total_payed_scale", 0);
        jVisit.put("white_sacks", false);
        jVisit.put("clear_bags", 0);
        jVisit.put("is_ad_hoc", true);
        jVisit.put("is_local_created", true);
        VisitEntity visEnt = ParserProvider.getApplicationMoshi().adapter(VisitEntity.class).fromJson(jVisit.toString());
        createOrUpdateInLocalDB(visEnt);
        VisitDbModel result = super.selectEntryById(LongId.Companion.valueOf(fakeVisitId));

        return result;
    }






    @Override
    public int deleteEntry(VisitEntity entry) {
        return super.deleteEntryById(entry.id());
    }

    @Override
    public VisitDbModel buildDbModel(Cursor c) {
        try {
            LongId<VisitEntity> visitId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_ID)));
            LongId<ShiftRestEntity> shiftId = LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_SHIFT_ID)));
            VisitState state = VisitState.valueOf(c.getString(c.getColumnIndex(COLUMN_STATE)));
            long actStart = c.getLong(c.getColumnIndex(COLUMN_ACTUAL_START_TIME));
            long actCompl = c.getLong(c.getColumnIndex(COLUMN_ACTUAL_COMPLETE_TIME));
            String suggTimeText = c.getString(c.getColumnIndex(COLUMN_SUGGEST_TIME));
            SiteEntity site = selectSite(LongId.Companion.valueOf(c.getLong(c.getColumnIndex(COLUMN_SITE_ID))));
            VisitDbModel.Builder bld = VisitDbModel.builder()
                    .setId(visitId)
                    .setState(state)
                    .setShiftId(shiftId)
                    .setSite(site)
                    .setOrder(c.getInt(c.getColumnIndex(COLUMN_ORDER)))
                    .setSuggestedTime((suggTimeText == null) ? null : TimeOfDayHMS.parseTime(suggTimeText))
                    .setActualStartTime((actStart > 0) ? new Date(actStart) : null)
                    .setActualCompleteTime((actCompl > 0) ? new Date(actCompl) : null)
                    .setIsSynced(c.getInt(c.getColumnIndex(COLUMN_IS_SYNCED)) > 0)
                    .setIsAdHocVisit(c.getInt(c.getColumnIndex(COLUMN_IS_AD_HOC)) > 0)
                    .setIsRemoteCreated(c.getInt(c.getColumnIndex(COLUMN_IS_REMOTE_CREATED)) > 0);
            if (state == VisitState.COMPLETED) {
                bld.setCustomerName(c.getString(c.getColumnIndex(COLUMN_CUSTOMER_NAME)))
                        .setCustomerSignature(c.getString(c.getColumnIndex(COLUMN_CUSTOMER_SIGNATURE)))
                        .setWhiteSacks(c.getInt(c.getColumnIndex(COLUMN_WHITE_SACKS)) == 1)
                        .setClearBags(c.getInt(c.getColumnIndex(COLUMN_CLEAR_BAGS)))
                        .setTotalPayedAmount(c.getInt(c.getColumnIndex(COLUMN_TOT_PAY_AMOUNT)))
                        .setTotalPayedScale(c.getInt(c.getColumnIndex(COLUMN_TOT_PAY_SCALE)))
                        .setOptNote(getNoteDao().selectNoteForVisit(shiftId, visitId));
            } else if (state == VisitState.FAILED) {
                bld.setOptNote(getNoteDao().selectNoteForVisit(shiftId, visitId));
            }

            return bld
                    .setCollectedGoods(getGoodDao().selectCollectedGoodsForVisit(visitId))
                    .build();
        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }

    private SiteEntity selectSite(LongId<SiteEntity> siteId) {
        SiteEntity site = (selectionCache == null) ? null : selectionCache.getSite(siteId);
        if (site == null) {
            site = getSiteDao().selectEntryById(siteId);
            if ((selectionCache != null) && (site != null)) selectionCache.putDbModel(site);
        }
        return site;
    }

    private SitesDAO siteDao;

    private SitesDAO getSiteDao() {
        if (siteDao == null) addChildDAO(siteDao = new SitesDAO());
        return siteDao;
    }

    private CollectedGoodDAO goodDao;

    private CollectedGoodDAO getGoodDao() {
        if (goodDao == null) addChildDAO(goodDao = new CollectedGoodDAO());
        return goodDao;
    }

    private NoteDAO noteDao;

    private NoteDAO getNoteDao() {
        if (noteDao == null) addChildDAO(noteDao = new NoteDAO());
        return noteDao;
    }

    public List<VisitDbModel> selectVisistForShiftSortedByOrder(LongId<ShiftRestEntity> shiftId) {
        final String where = String.format("%s = %d", COLUMN_SHIFT_ID, shiftId.getValue());
        final String sort = COLUMN_ORDER + " ASC";
        return selectEntriesForWhereAndSort(where, sort, 0);
    }


    @Nullable
    public VisitState getVisitState(LongId<VisitEntity> visitId) {
        final String where = String.format("%s = %d", COLUMN_ID, visitId.getValue());
        try (Cursor c = getWriteableDatabase().query(TABLE_NAME, new String[]{COLUMN_STATE}, where, null, null, null, null, "1")) {
            if (c.moveToFirst()) {
                return VisitState.valueOf(c.getString(0));
            } else {
                return null;
            }
        }
    }

    public void updateVisitStateFromRemote(VisitStateEntity stateResp) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_STATE, stateResp.getState().name());
        vvv.put(COLUMN_ACTUAL_START_TIME, (stateResp.getActualStartTime() == null) ? 0 : stateResp.getActualStartTime().getTime());
        vvv.put(COLUMN_ACTUAL_COMPLETE_TIME, (stateResp.getActualCompleteTime() == null) ? 0 : stateResp.getActualCompleteTime().getTime());
        vvv.put(COLUMN_IS_SYNCED, 1);
        String where = COLUMN_ID + " = " + stateResp.id().getValue();
        getWriteableDatabase().update(TABLE_NAME, vvv, where, null);
    }

    public void clear(LongId<VisitEntity> visitId) {
        getGoodDao().clear(visitId);
        getNoteDao().clearAllNotes();
    }

    public void startVisitStateLocally(LongId<VisitEntity> visitId) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_STATE, VisitState.IN_PROGRESS.name());
        vvv.put(COLUMN_ACTUAL_START_TIME, System.currentTimeMillis());
        vvv.put(COLUMN_IS_SYNCED, 0);
        String where = COLUMN_ID + " = " + visitId.getValue();
        getWriteableDatabase().update(TABLE_NAME, vvv, where, null);
    }

    public void completeVisitStateLocally(LongId<VisitEntity> visitId) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_STATE, VisitState.COMPLETED.name());
        vvv.put(COLUMN_ACTUAL_COMPLETE_TIME, System.currentTimeMillis());
        vvv.put(COLUMN_IS_SYNCED, 0);
        String where = COLUMN_ID + " = " + visitId.getValue();
        getWriteableDatabase().update(TABLE_NAME, vvv, where, null);
    }

    public void failVisitStateLocally(LongId<VisitEntity> visitId) {
        ContentValues vvv = new ContentValues();
        vvv.put(COLUMN_STATE, VisitState.FAILED.name());
        vvv.put(COLUMN_ACTUAL_COMPLETE_TIME, System.currentTimeMillis());
        vvv.put(COLUMN_IS_SYNCED, 0);
        String where = COLUMN_ID + " = " + visitId.getValue();
        getWriteableDatabase().update(TABLE_NAME, vvv, where, null);
    }

    @Nullable
    public TimeOfDayHMS selectLatestTimeOfVisit(LongId<ShiftRestEntity> shiftId) {
        final String where = String.format("%s = %d", COLUMN_SHIFT_ID, shiftId.getValue());
        try (Cursor c = getWriteableDatabase().query(TABLE_NAME, new String[]{COLUMN_SUGGEST_TIME}, where, null, null, null, COLUMN_SUGGEST_TIME_MILLIS+" DESC", "1")) {
            if (c.moveToFirst()) {
                String suggTimeText = c.getString(0);
                return (suggTimeText == null) ? null : TimeOfDayHMS.parseTime(suggTimeText);
            } else {
                return null;
            }
        } catch (ParseException e) {
            return null;
        }
    }

    public int selectMaxVisitOrderForShift(LongId<ShiftRestEntity> shiftId) {
        final String where = String.format("%s = %d", COLUMN_SHIFT_ID, shiftId.getValue());
        try (Cursor c = getWriteableDatabase().query(TABLE_NAME, new String[]{COLUMN_ORDER}, where, null, null, null, COLUMN_SUGGEST_TIME_MILLIS+" DESC", "1")) {
            if (c.moveToFirst()) {
                return c.getInt(0);
            } else {
                return 0;
            }
        }
    }
}
