package com.webssa.pda.sunmi.ui.activity;

import android.Manifest;
import android.app.AlertDialog;
import android.content.Context;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.telephony.TelephonyManager;
import android.text.TextUtils;
import android.view.View;
import android.widget.Checkable;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.databinding.BaseObservable;
import androidx.databinding.Bindable;
import androidx.databinding.DataBindingUtil;

import com.webssa.library.core.error.AppError;
import com.webssa.pda.sunmi.BR;
import com.webssa.pda.sunmi.BuildConfig;
import com.webssa.pda.sunmi.GlobalProperties;
import com.webssa.pda.sunmi.R;
import com.webssa.pda.sunmi.core.SessionHolder;
import com.webssa.pda.sunmi.databinding.ActivityLoginBinding;
import com.webssa.pda.sunmi.rest.model.DriverLoginRequest;
import com.webssa.pda.sunmi.rest.model.LoginMode;
import com.webssa.pda.sunmi.rest.model.UserEntity;
import com.webssa.pda.sunmi.ui.NavigationHelper;
import com.webssa.pda.sunmi.ui.presenter.LoginPresenter;
import com.webssa.pda.sunmi.util.StringUtils;


public class LoginActivity extends AppCompatActivity implements LoginActivityView {

    private View vProgressOverlay;

    private LoginDisplayModel mDisplModel;

    private boolean isStarted;

    private LoginPresenter mPresenter;

    private LoginMode currentLoginMode = LoginMode.BY_EMAIL;

    @Override
    protected void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        final ActivityLoginBinding binding = DataBindingUtil.setContentView(this, R.layout.activity_login);
        binding.setViewModel(mDisplModel = new LoginDisplayModel());
        binding.setClickerLogin(v -> {

            if (!mPresenter.isLoginInProgress() && (ContextCompat.checkSelfPermission(this, Manifest.permission.READ_PHONE_STATE) == PackageManager.PERMISSION_GRANTED)) {
                String imei = ((TelephonyManager) this.getSystemService(Context.TELEPHONY_SERVICE)).getDeviceId();
                if(imei == null || imei.isEmpty()) {
                    imei = "no_imei";
                }
                mPresenter.startLogin(new DriverLoginRequest(currentLoginMode, imei, mDisplModel.getLogin(), mDisplModel.getPassword()));
            }
            updateProgressVisibility();
        });


        binding.selectLoginEmail.setOnClickListener(v -> {
            if (!((Checkable) v).isChecked()) {
                if (BuildConfig.DEBUG) {
                    prefillLoginEmail();
                } else {
                    clearCreeds();
                }
                mDisplModel.setLoginMode(currentLoginMode = LoginMode.BY_EMAIL);
                binding.selectLoginEmail.setChecked(true);
                binding.selectLoginPhone.setChecked(false);
            }
        });
        binding.selectLoginPhone.setOnClickListener(v -> {
            if (!((Checkable) v).isChecked()) {
                if (BuildConfig.DEBUG) {
                    prefillLoginPhone();
                } else {
                    clearCreeds();
                }
                mDisplModel.setLoginMode(currentLoginMode = LoginMode.BY_PHONE_NUMBER);
                binding.selectLoginEmail.setChecked(false);
                binding.selectLoginPhone.setChecked(true);
            }
        });

        mPresenter = new LoginPresenter(this);
        mPresenter.initializeView();
        vProgressOverlay = binding.progress;
        updateProgressVisibility();
    }

    @Override
    protected void onPostCreate(@Nullable Bundle savedInstanceState) {
        super.onPostCreate(savedInstanceState);
        if (BuildConfig.DEBUG) {
            vProgressOverlay.postDelayed(() -> prefillLoginEmail(), 250);
        }
    }

    private void prefillLoginEmail() {
        mDisplModel.setLogin("craig.ihezie@gmail.com");
        mDisplModel.setPassword("123456");
    }

    private void prefillLoginPhone() {
        mDisplModel.setLogin("+380501230000");
        mDisplModel.setPassword("123456");
    }

    private void clearCreeds() {
        mDisplModel.setLogin("");
        mDisplModel.setPassword("");
    }

    private static final int REQUEST_IMEI_ACCESS = 101;

    @Override
    protected void onStart() {
        super.onStart();
        isStarted = true;
        if (SessionHolder.isLoggedIn()) {
            NavigationHelper.goMainScreen(this, true);
            finish();
        } else if (ContextCompat.checkSelfPermission(this, Manifest.permission.READ_PHONE_STATE) != PackageManager.PERMISSION_GRANTED) {
            showPermissionAlert();
        }
    }

    private void showPermissionAlert() {
        AlertDialog ad = new AlertDialog.Builder(this)
                .setTitle("Need permission")
                .setMessage("This application needs access to the device IMEI in order to operate. Please grant the requesting permission.")
                .setCancelable(false).setPositiveButton(android.R.string.ok, (dialog, which) -> ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.READ_PHONE_STATE}, REQUEST_IMEI_ACCESS))
                .setNegativeButton(android.R.string.no, (dialog, which) -> finish())
                .create();
        ad.setCanceledOnTouchOutside(false);
        ad.show();
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {
        if (requestCode == REQUEST_IMEI_ACCESS) {
            if (grantResults.length == 0 || grantResults[0] != PackageManager.PERMISSION_GRANTED) {
                if (ActivityCompat.shouldShowRequestPermissionRationale(this, Manifest.permission.READ_PHONE_STATE)) {
                    showPermissionAlert();
                } else {
                    //Ends-up with never ask again
                    NavigationHelper.openApplicationSettings(this);
                }
            }
        }
    }

    @Override
    protected void onStop() {
        isStarted = false;
        super.onStop();
    }

    @Override
    public void onBackPressed() {
        if (!mPresenter.isLoginInProgress()) {
            super.onBackPressed();
        }
    }

    private void updateProgressVisibility() {
        vProgressOverlay.setVisibility(mPresenter.isLoginInProgress() ? View.VISIBLE : View.GONE);
    }


    /**********************************************************************************************/
    /************************  Activity View interface implementation  ****************************/
    /**********************************************************************************************/

    @Override
    public void onLoginCompleted(@NonNull UserEntity myProfile) {
        if (BuildConfig.DEBUG) {
            Toast.makeText(this, getString(R.string.msg_login_result_ok, myProfile.getFullName()), Toast.LENGTH_SHORT).show();
        }
        if (isStarted) {
            NavigationHelper.goMainScreen(this, true);
            finish();
        }
    }

    @Override
    public void onLoginFailed(@Nullable AppError reason) {
        updateProgressVisibility();
        if (mDisplModel != null) mDisplModel.setPassword("");
        if (isStarted) {
            String msg = getResources().getString(R.string.msg_login_result_failed);
            if (reason != null) {
                msg = BuildConfig.DEBUG
                        ? String.format("%s: %s (%s)", msg, reason.getUserMessage(), reason.getInternalMessage())
                        : String.format("%s - %s", msg, reason.getUserMessage());
            }
            Toast.makeText(this, msg, Toast.LENGTH_LONG).show();
        }
    }

    @NonNull
    @Override
    public Context getContext() {
        return this;
    }

    @Override
    public void close() {
        finish();
    }



    /**********************************************************************************************/
    /*************************  Bindable display model  *******************************************/
    /**********************************************************************************************/
    public static class LoginDisplayModel extends BaseObservable {
        private LoginMode loginMode = LoginMode.BY_EMAIL;
        private String login, password;

        @Bindable
        public boolean isDataValid() {
            return StringUtils.validatePassword(password) && StringUtils.validateLogin(login, loginMode);
        }

        @Bindable
        public LoginMode getLoginMode() {
            return loginMode;
        }

        public void setLoginMode(LoginMode loginMode) {
            if (this.loginMode != loginMode) {
                this.loginMode = loginMode;
                notifyPropertyChanged(BR.loginMode);
                notifyPropertyChanged(BR.dataValid);
            }
        }

        @Bindable
        public String getLogin() {
            return login;
        }

        public void setLogin(String login) {
            if (!TextUtils.equals(this.login, login)) {
                this.login = login;
                notifyPropertyChanged(BR.login);
                notifyPropertyChanged(BR.dataValid);
            }
        }

        @Bindable
        public String getPassword() {
            return password;
        }

        public void setPassword(String password) {
            if (!TextUtils.equals(this.password, password)) {
                this.password = password;
                notifyPropertyChanged(BR.password);
                notifyPropertyChanged(BR.dataValid);
            }
        }
    }
}
