package com.webssa.pda.sunmi.ui.adapter;

import android.content.Context;
import android.location.Location;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.library.util.OnAdapterItemClickListener;
import com.webssa.pda.sunmi.GlobalProperties;
import com.webssa.pda.sunmi.R;
import com.webssa.pda.sunmi.rest.model.IdProvidingModel;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.ShiftState;
import com.webssa.pda.sunmi.rest.model.VisitState;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.storage.dbmodel.BaseNoteDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.ScheduledModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.ShiftDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;
import com.webssa.pda.sunmi.ui.widget.NoteListItemView;
import com.webssa.pda.sunmi.ui.widget.ShiftListItemDesignedView;
import com.webssa.pda.sunmi.ui.widget.VisitListItemView;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Created by stanislav.perchenko on 10/23/2019.
 */
public class ShiftAdapter extends BaseAdapter {

    public interface OnCreateAdHocVisitListener {
        void onCreateAdHocVisit(LongId<ShiftRestEntity> shiftId);
    }

    private static final int NUM_ITEM_TYPES = 4;
    private static final int ITEM_TYPE_SHIFT = 0;
    private static final int ITEM_TYPE_VISIT = 1;
    private static final int ITEM_TYPE_NOTE = 2;
    private static final int ITEM_TYPE_AD_HOC_VISIT_SLOT = 3;

    private final LayoutInflater inflater;
    private Location location;
    private ShiftDbModel shift;

    private int size = 0;
    private final List<ScheduledModel> mItems;

    private OnAdapterItemClickListener onItemClickListener;
    private OnAdapterItemClickListener onItemLongClickListener;
    private ShiftListItemDesignedView.OnSearchNewTasksListener onSearchNewTasksListener;
    private OnCreateAdHocVisitListener onCreateAdHocVisitListener;

    public ShiftAdapter(Context context, @Nullable ShiftDbModel shift) {
        this.inflater = LayoutInflater.from(context);
        mItems = new ArrayList<>(20);
        if (shift != null) setShiftInternal(shift);
    }


    public void setOnItemClickListener(OnAdapterItemClickListener l) {
        this.onItemClickListener = l;
    }

    public void setOnItemLongClickListener(OnAdapterItemClickListener l) {
        this.onItemLongClickListener = l;
    }

    public void setOnSearchNewTasksListener(ShiftListItemDesignedView.OnSearchNewTasksListener l) {
        this.onSearchNewTasksListener = l;
    }

    public void setOnCreateAdHocVisitListener(OnCreateAdHocVisitListener l) {
        this.onCreateAdHocVisitListener = l;
    }

    public void setLocation(@Nullable Location location) {
        this.location = location;
        notifyDataSetChanged();
    }

    public void setShift(@NonNull ShiftDbModel shift) {
        if (!shift.equals(this.shift)) setShiftInternal(shift);
    }

    @Nullable
    public LongId<ShiftRestEntity> getShiftId() {
        return (shift == null) ? null : shift.id();
    }

    @Nullable
    public ShiftState getShiftState() {
        return (shift == null) ? null : shift.state();
    }

    public boolean isHasInProgressVisit() {
        for (ScheduledModel item : mItems) {
            if (item instanceof VisitDbModel) {
                VisitDbModel v = (VisitDbModel) item;
                if (v.state() == VisitState.IN_PROGRESS) return true;
            }
        }
        return false;
    }

    public boolean isHasUnsyncedData() {
        for (ScheduledModel item : mItems) {
            if (item instanceof VisitDbModel) {
                if (!((VisitDbModel) item).isSynced()) return true;
            }
        }
        return false;
    }

    private void setShiftInternal(ShiftDbModel shift) {
        this.shift = shift;
        mItems.clear();
        mItems.add(shift);
        mItems.addAll(shift.visits());
        mItems.addAll(shift.remoteNotes());
        mItems.addAll(shift.localNotes());
        Collections.sort(mItems, (o1, o2) -> (o1.scheduledTimeMillis() - o2.scheduledTimeMillis()) );



        //----  Add slots for ad-hoc-visits  ----
        boolean hasTime = (shift.visits().size() > 0) && ( (mItems.get(mItems.size()-1).scheduledTimeMillis()/60000) <= (22*60+15) );
        hasTime = true;
        if (shift.state() == ShiftState.ACCEPTED && hasTime) {
            int nAdHoc = 0;
            for (VisitDbModel vis : shift.visits()) {
                if (vis.isAdHocVisit()) nAdHoc++;
            }
            if (nAdHoc < GlobalProperties.MAX_NUM_AD_HOC_VISITS) {
                int nSlots = GlobalProperties.MAX_NUM_AD_HOC_VISITS - nAdHoc;
                int t = mItems.get(mItems.size() - 1).scheduledTimeMillis();
                for (int order = 1; order <= nSlots; order++) {
                    t += 1000;
                    mItems.add(new AdHocVisitSlotModel(order, t));
                }
            }
        }
        this.size = mItems.size();
        notifyDataSetChanged();
    }

    @Override
    public int getViewTypeCount() {
        return NUM_ITEM_TYPES;
    }

    @Override
    public int getCount() {
        return size;
    }

    @Override
    public int getItemViewType(int position) {
        ScheduledModel item = getItem(position);
        if (item instanceof ShiftDbModel) {
            return ITEM_TYPE_SHIFT;
        } else if (item instanceof VisitDbModel) {
            return ITEM_TYPE_VISIT;
        } else if (item instanceof BaseNoteDbModel) {
            return ITEM_TYPE_NOTE;
        } else if (item instanceof AdHocVisitSlotModel) {
            return ITEM_TYPE_AD_HOC_VISIT_SLOT;
        } else {
            throw new RuntimeException("Item type not supported - "+item.getClass().getName());
        }
    }

    @Override
    public ScheduledModel getItem(int position) {
        return mItems.get(position);
    }

    @Override
    public long getItemId(int position) {
        ScheduledModel item = getItem(position);
        return (item instanceof IdProvidingModel) ? ((IdProvidingModel) item).id().getValue() : 0;
    }

    @Override
    public View getView(int position, View convertView, ViewGroup parent) {
        View row = convertView;
        if (row == null) {
            ViewHolder holder = null;
            switch (getItemViewType(position)) {
                case ITEM_TYPE_SHIFT:
                    holder = new ShiftViewHolder(inflater, (ShiftDbModel) getItem(position));
                    break;
                case ITEM_TYPE_VISIT:
                    holder = new VisitViewHolder(inflater, position, (VisitDbModel) getItem(position));
                    break;
                case ITEM_TYPE_NOTE:
                    holder = new NoteViewHolder(inflater, position, (BaseNoteDbModel) getItem(position));
                    break;
                case ITEM_TYPE_AD_HOC_VISIT_SLOT:
                    holder = new AdHocSlotViewHolder(inflater, ((AdHocVisitSlotModel) getItem(position)).slotOrder);
                    break;
            }
            row = holder.getBaseView();
            row.setTag(holder);
        } else {
            switch (getItemViewType(position)) {
                case ITEM_TYPE_SHIFT:
                    ((ShiftViewHolder) row.getTag()).bindItem((ShiftDbModel) getItem(position));
                    break;
                case ITEM_TYPE_VISIT:
                    ((VisitViewHolder) row.getTag()).bindItem(position, (VisitDbModel) getItem(position));
                    break;
                case ITEM_TYPE_NOTE:
                    ((NoteViewHolder) row.getTag()).bindItem(position, (BaseNoteDbModel) getItem(position));
                    break;
                case ITEM_TYPE_AD_HOC_VISIT_SLOT:
                    ((AdHocSlotViewHolder) row.getTag()).bindItem(((AdHocVisitSlotModel) getItem(position)).slotOrder);
                    break;
            }
        }
        return row;
    }


    private interface ViewHolder {
        View getBaseView();
    }

    private class ShiftViewHolder implements ViewHolder {
        private final ShiftListItemDesignedView view;

        ShiftViewHolder(LayoutInflater inflater, ShiftDbModel shift) {
            view = new ShiftListItemDesignedView(inflater);
            view.setOnSearchNewTasksListener(v -> {
                if (onSearchNewTasksListener != null) onSearchNewTasksListener.onStartSearching(v);
            });
            bindItem(shift);
        }

        @Override
        public View getBaseView() {
            return view;
        }

        void bindItem(ShiftDbModel shift) {
            view.setData(shift);
        }
    }

    private class VisitViewHolder implements ViewHolder {
        private final VisitListItemView view;

        VisitViewHolder(LayoutInflater inflater, int position, VisitDbModel visit) {
            view = new VisitListItemView(inflater, (pos, vis) -> {
                if (onItemClickListener != null) onItemClickListener.onItemClicked(ShiftAdapter.this, pos);
            });
            view.setOnLongClickListener((pos, vis) -> {
                if (onItemLongClickListener != null) onItemLongClickListener.onItemClicked(ShiftAdapter.this, pos);
            });
            bindItem(position, visit);
        }

        @Override
        public View getBaseView() {
            return view;
        }

        void bindItem(int position, VisitDbModel visit) {
            view.setData(visit, position);
            view.setLocation(location);
        }
    }

    private class NoteViewHolder implements ViewHolder {
        private final NoteListItemView view;

        NoteViewHolder(LayoutInflater inflater, int position, BaseNoteDbModel note) {
            view = new NoteListItemView(inflater, pos -> {
                if (onItemClickListener != null) onItemClickListener.onItemClicked(ShiftAdapter.this, pos);
            });
            view.setOnLongClickListener(position1 -> {
                if (onItemLongClickListener != null) onItemLongClickListener.onItemClicked(ShiftAdapter.this, position);
            });
            bindItem(position, note);
        }

        @Override
        public View getBaseView() {
            return view;
        }

        void bindItem(int position, BaseNoteDbModel note) {
            view.setData(note, position);
        }
    }

    private class AdHocSlotViewHolder implements ViewHolder {
        private final View view;
        private final TextView vTxtTitle;
        private final String txt_title_template;

        AdHocSlotViewHolder(LayoutInflater inflater, int order) {
            view = inflater.inflate(R.layout.ad_hoc_slot_list_item, null, false);
            vTxtTitle = view.findViewById(R.id.txt_title);
            view.findViewById(R.id.click_container).setOnClickListener(v -> {
                if (onCreateAdHocVisitListener != null) onCreateAdHocVisitListener.onCreateAdHocVisit(shift.id());
            });
            txt_title_template = inflater.getContext().getResources().getString(R.string.ad_hoc_visit_slot_title);
            bindItem(order);
        }

        @Override
        public View getBaseView() {
            return view;
        }

        void bindItem(int order) {
            vTxtTitle.setText(String.format(txt_title_template, order));
        }
    }




}
