package com.webssa.pda.sunmi.ui.dialog;

import android.app.Dialog;
import android.content.Context;
import android.graphics.Color;
import android.graphics.drawable.ColorDrawable;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.os.Bundle;
import android.util.Log;
import android.view.KeyEvent;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.Window;
import android.view.WindowManager;
import android.widget.ProgressBar;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.fragment.app.DialogFragment;

import com.alperez.sunmi.pos.receiptengine.escpos.ESCUtils;
import com.alperez.sunmi.pos.receiptengine.parammapper.ParameterValueMapper;
import com.alperez.sunmi.pos.receiptengine.print.PosPrinter;
import com.alperez.sunmi.pos.receiptengine.print.PrintResultCallback;
import com.alperez.sunmi.pos.receiptengine.template.ReceiptTemplate;
import com.squareup.moshi.Moshi;
import com.webssa.library.utils.AsyncTaskCompat;
import com.webssa.pda.sunmi.R;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.receipt.ReceiptModelBuilder;
import com.webssa.pda.sunmi.receipt.ReceiptModelsMoshiJsonFactory;
import com.webssa.pda.sunmi.receipt.model.ReceiptDataMainModel;
import com.webssa.pda.sunmi.util.FileUtils;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;

/**
 * Created by stanislav.perchenko on 12/5/2019, 11:56 AM.
 */
public class PrintReceiptDialogFragment extends DialogFragment {

    public interface OnCompleteListener {
        void onPrintDone();
        void onPrintError();
    }


    public static PrintReceiptDialogFragment newInstance(LocalVisitResult visitResult) {
        Bundle args = new Bundle();
        args.putParcelable("visit_result", visitResult);
        PrintReceiptDialogFragment f = new PrintReceiptDialogFragment();
        f.setArguments(args);
        return f;
    }


    private OnCompleteListener mListener;
    private LocalVisitResult visitResult;


    @Override
    public void onAttach(Context activity) {
        super.onAttach(activity);


        if (activity instanceof OnCompleteListener) {
            mListener = (OnCompleteListener) activity;
        }
    }

    @Override
    public void onDetach() {
        super.onDetach();
        mListener = null;
    }


    @Override
    public void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        Bundle args = getArguments();
        visitResult = (args == null) ? null : args.getParcelable("visit_result");
    }

    @Override
    public Dialog onCreateDialog(Bundle savedInstanceState) {
        Drawable bgTransparent = new ColorDrawable(Color.TRANSPARENT);

        Dialog dialog = new Dialog(getActivity(), getTheme()) {
            @Override
            public boolean onKeyUp(int keyCode, KeyEvent event) {
                if (keyCode == KeyEvent.KEYCODE_BACK && event.isTracking() && !event.isCanceled()) {
                    // No back action!
                    // No auto-dismiss because of not calling super
                    return true;
                }
                return super.onKeyUp(keyCode, event);
            }
        };



        dialog.setCancelable(false);
        dialog.setCanceledOnTouchOutside(false);
        dialog.getWindow().requestFeature(Window.FEATURE_NO_TITLE);
        dialog.getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN, WindowManager.LayoutParams.FLAG_FULLSCREEN);
        dialog.getWindow().setBackgroundDrawable(bgTransparent);

        //----  The hack to hide title-content divider on some phones  ----
        int dividerId = getActivity().getResources().getIdentifier("android:id/titleDivider", null, null);
        if(dividerId > 0) {
            View vDiv = dialog.findViewById(dividerId);
            if(vDiv != null) {
                if(Build.VERSION.SDK_INT >= 16) {
                    vDiv.setBackground(bgTransparent);
                } else {
                    vDiv.setBackgroundDrawable(bgTransparent);
                }
            }
        }
        return dialog;
    }


    private View vContent;
    private ProgressBar vProgress;

    @Nullable
    @Override
    public View onCreateView(@NonNull LayoutInflater inflater, @Nullable ViewGroup container, @Nullable Bundle savedInstanceState) {
        if (vContent == null) {
            vContent = getLayoutInflater().inflate(R.layout.print_receipt_dialog, container, false);
            vProgress = vContent.findViewById(R.id.progress);
            vProgress.setMax(0);
            vProgress.setProgress(0);
        } else {
            container.removeView(vContent);
        }
        return vContent;
    }

    private LoadReceiptTask loadReceiptTask;

    @Override
    public void onActivityCreated(@Nullable Bundle savedInstanceState) {
        super.onActivityCreated(savedInstanceState);
        if (visitResult != null) {
            (loadReceiptTask = new LoadReceiptTask()).safeExecute(visitResult);
        } else {
            if (mListener != null) mListener.onPrintError();
            dismiss();
        }
    }


    private class LoadReceiptTask extends AsyncTaskCompat<LocalVisitResult, Void, ReceiptTemplate> {
        private Context context;

        @Override
        protected void onPreExecute() {
            context = getContext();
        }

        @Override
        protected ReceiptTemplate doInBackground(LocalVisitResult... localVisitResults) {
            try {
                ReceiptModelBuilder receiptBuilder = new ReceiptModelBuilder(context, localVisitResults[0]);
                ReceiptDataMainModel receiptModel = receiptBuilder.buildReceiptModel();

                final String receiptJson = (new Moshi.Builder()).add(ReceiptModelsMoshiJsonFactory.create()).build().adapter(ReceiptDataMainModel.class).toJson(receiptModel);
                ParameterValueMapper mapper = ParameterValueMapper.getInstance(new JSONObject(receiptJson));

                String templateJson = FileUtils.loadAsset(context, "receipt_template.json");
                JSONObject jTemplate = new JSONObject(templateJson);

                ReceiptTemplate receiptTemplate = ReceiptTemplate.fromJson(jTemplate, mapper);
                return receiptTemplate;
            } catch (JSONException | IOException e) {
                return null;
            }
        }

        @Override
        protected void onPostExecute(ReceiptTemplate receiptTemplate) {
            if (receiptTemplate != null) {
                print(context, receiptTemplate);
            } else {
                dismiss();
                if (mListener != null) mListener.onPrintError();
            }
            loadReceiptTask = null;
        }

        @Override
        protected void onCancelled(ReceiptTemplate receiptTemplate) {
            dismiss();
            if (mListener != null) mListener.onPrintError();
            loadReceiptTask = null;
        }
    }


    int cnt_prn_all_tasks, cnt_prn_ok, cnt_prn_err;

    private void print(Context context, ReceiptTemplate receiptTemplate) {
        cnt_prn_all_tasks = cnt_prn_ok = cnt_prn_err = 0;

        PosPrinter printer = PosPrinter.getSunmiInternalBTPrinter(context);

        for (Iterator<Collection<byte[]>> itr = receiptTemplate.getRawPrintDataIterator(printer.getPrinterParams()); itr.hasNext(); ) {
            printer.print(itr.next(), printCallback);
            vProgress.setMax(++cnt_prn_all_tasks);
            logPrinting("add task");
        }
        //noinspection ArraysAsListWithZeroOrOneArgument
        printer.print(Arrays.asList(ESCUtils.nextLine(3)), null);
    }

    private final PrintResultCallback printCallback = new PrintResultCallback() {
        @Override
        public void onComplete() {
            cnt_prn_ok ++;
            vProgress.setProgress(cnt_prn_ok + cnt_prn_err);
            logPrinting("result OK");
            checkForCompletion();
        }

        @Override
        public void onError(@NonNull String reason, @Nullable Throwable t) {
            cnt_prn_err ++;
            vProgress.setProgress(cnt_prn_ok + cnt_prn_err);
            logPrinting("result ERR");
            checkForCompletion();
        }
    };

    private void checkForCompletion() {
        if (cnt_prn_ok + cnt_prn_err == cnt_prn_all_tasks) {
            if (mListener != null && (cnt_prn_err > 0))  {
                mListener.onPrintError();
            } else if (mListener != null) {
                mListener.onPrintDone();
            }
            dismiss();
        }
    }

    private void logPrinting(String type) {
        Log.e("PRN_PROGR", String.format("%s: cnt1 = %d; cnt_ok = %d; cnt_err = %d", type, cnt_prn_all_tasks, cnt_prn_ok, cnt_prn_err));
    }
}
