package com.webssa.pda.sunmi.ui.presenter;

import android.annotation.SuppressLint;
import android.os.Looper;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.library.core.error.AppError;
import com.webssa.library.core.error.ServerRestError;
import com.webssa.library.core.processor.BackgroundExecutor;
import com.webssa.library.core.processor.model.Callback;
import com.webssa.library.core.processor.model.Disposable;
import com.webssa.library.utils.AsyncTaskCompat;
import com.webssa.pda.sunmi.broadcasting.BroadcastController;
import com.webssa.pda.sunmi.dataapi.ApiAcceptShift;
import com.webssa.pda.sunmi.dataapi.ApiCreateAdHocVisit;
import com.webssa.pda.sunmi.dataapi.ApiFindTodayNewShift;
import com.webssa.pda.sunmi.dataapi.ApiGetAllSites;
import com.webssa.pda.sunmi.dataapi.ApiLoadShiftFromBackend;
import com.webssa.pda.sunmi.dataapi.ApiRejectShift;
import com.webssa.pda.sunmi.dataapi.ApiSelectSavedShiftForDate;
import com.webssa.pda.sunmi.dataapi.ApiStartVisit;
import com.webssa.pda.sunmi.dataapi.ApiSyncVisitsBack;
import com.webssa.pda.sunmi.model.VisitCreationModel;
import com.webssa.pda.sunmi.rest.ApplicationRestService;
import com.webssa.pda.sunmi.rest.model.FindNewShiftResponse;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.ShiftState;
import com.webssa.pda.sunmi.rest.model.impl.ShiftRestEntity;
import com.webssa.pda.sunmi.rest.model.impl.SiteEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.storage.dao.impl.CollectedGoodDAO;
import com.webssa.pda.sunmi.storage.dao.impl.ShiftDAO;
import com.webssa.pda.sunmi.storage.dbhelper.DbSelectionCache;
import com.webssa.pda.sunmi.storage.dbmodel.impl.ShiftDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;
import com.webssa.pda.sunmi.ui.activity.ScannerActivityView;
import com.webssa.pda.sunmi.ui.activity.ShiftActivityView;
import com.webssa.pda.sunmi.ui.presenter.base.BasePresenter;
import com.webssa.pda.sunmi.util.DateNoTime;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Created by stanislav.perchenko on 10/18/2019.
 */
public class ShiftActivityPresenter extends BasePresenter<ShiftActivityView> {

    private BackgroundExecutor exec;

    private boolean isShiftDataLoading;

    public ShiftActivityPresenter(@NonNull ShiftActivityView view) {
        super(view);
    }

    @Override
    public void initializeView() {
        cancelAllInProgressRequests();
        if (exec == null) exec = new BackgroundExecutor(Looper.getMainLooper());
    }

    @Override
    public synchronized void release() {
        super.release();
        if (exec != null) {
            exec.release();
            exec = null;
        }
        if (localLoadTask != null) {
            localLoadTask.cancel(true);
            localLoadTask = null;
        }
    }

    private  AsyncTaskCompat<LongId<ShiftRestEntity>, Void, ShiftDbModel> localLoadTask;

    @SuppressLint("StaticFieldLeak")
    public void reloadAcceptedShiftLocally(LongId<ShiftRestEntity> shiftId) {
        if (localLoadTask != null) return;
        localLoadTask = new AsyncTaskCompat<LongId<ShiftRestEntity>, Void, ShiftDbModel>() {
            @Override
            protected ShiftDbModel doInBackground(LongId<ShiftRestEntity>... longIds) {
                try (ShiftDAO dao = new ShiftDAO(new DbSelectionCache())) {
                    ShiftDbModel shift = dao.selectEntryById(longIds[0]);
                    return ((shift != null) && (shift.state() == ShiftState.ACCEPTED)) ? shift : null;
                }
            }

            @Override
            protected void onPostExecute(@Nullable ShiftDbModel shift) {
                localLoadTask = null;
                ShiftActivityView v = getView();
                if ((shift != null) && !isReleased() && (v != null)) {
                    v.onShiftFound(shift);
                }
            }

            @Override
            protected void onCancelled(@Nullable ShiftDbModel shiftDbModel) {
                localLoadTask = null;
            }
        };
        localLoadTask.safeExecute(shiftId);
    }

    /**
     * At the 1-st stage this APi search for the locally-saved shift for the current day (which must be in the ACCEPTED state).
     * In case of the saved shift is found, this API returns to the {@Link ShiftActivityView#onShiftFound(ShiftDbModel)} callback
     *
     * In case of a saved Shift is not found this API proceed with searching new shift on the backend - stage 2.
     *
     * In case of some error in the REST call other then 404, this API returns to the {@Link ShiftActivityView#onNewShiftNotFound(AppError)}
     * callback with the non-null reason appError instance.
     *
     * In case of no new shift for today is found (no shift was returned of 404 error), this API
     * returns to the {@Link ShiftActivityView#onNewShiftNotFound(AppError)} callback
     * with null reason.
     *
     * In case of a new Shift was found, this API returns to the {@Link ShiftActivityView#onNewShiftFound(ShiftRestEntity, Date)}
     * with Shift in the NEW state and current server time.
     *
     * This method must be called from the UI thread.
     */
    public boolean loadTodayShift() {
        if (isShiftDataLoading) {
            return false;
        } else {
            isShiftDataLoading = true;
        }

        ApiSelectSavedShiftForDate.forDate(DateNoTime.today()).on(exec).executeAsync(new Callback<ShiftDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable ShiftDbModel result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onShiftFound(result);
                }
                isShiftDataLoading = false;
            }

            @Override
            public void onError(@NonNull AppError reason) {
                isShiftDataLoading = false;
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    // Continue with remote search
                    searchNewShiftForToday();
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
        return true;
    }

    public boolean searchNewShiftForToday() {
        if (isShiftDataLoading) {
            return false;
        } else {
            isShiftDataLoading = true;
        }

        ApiFindTodayNewShift.withContext(getView().getContext()).on(exec).executeAsync(new Callback<FindNewShiftResponse>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable FindNewShiftResponse result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    if (result.getOptFoundShift() != null) {
                        v.onNewShiftFound(result.getOptFoundShift(), result.getServerTime());
                    } else {
                        v.onNewShiftNotFound(null);
                    }
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onNewShiftNotFound(((reason instanceof ServerRestError) && (((ServerRestError) reason).getHttpCode() == 404)) ? null : reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                isShiftDataLoading = false;
                unregisterApiDisposable(d);
            }
        });
        return true;
    }


    public void acceptShift(LongId<ShiftRestEntity> shiftId) {
        ApiAcceptShift.create(getView().getContext(), shiftId).on(exec).executeAsync(new Callback<ShiftDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable ShiftDbModel result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onShiftFound(result);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onNewShiftAcceptFailed(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }

    public void loadAllSites() {
        ApiGetAllSites.create(getView().getContext()).on(exec).executeAsync(new Callback<List<SiteEntity>>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable List<SiteEntity> result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) v.onAllSitesLoaded((result == null) ? 0 : result.size());
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) v.onLoadAllSitesError(reason);
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }

    public void rejectShift(LongId<ShiftRestEntity> shiftId) {
        ApiRejectShift.create(getView().getContext(), shiftId).on(exec).executeAsync(new Callback<ShiftRestEntity>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable ShiftRestEntity result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onNewShiftRejected(result.id());
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onNewShiftRejectFailed(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }


    /**
     * This API reloads locally-saved shift from the backend.
     * If this shift of consisting Vists are not synced back with the backend, it tries to upload data first
     * and then download back full shift
     *
     * This API return to the {@Link ShiftActivityView#onShiftFound(ShiftDbModel)} in case of success
     * and to the {@Link ShiftActivityView#onShiftUpdateFailed(LongId<ShiftRestEntity>, AppError)} in case of failure
     *
     * @param shiftId
     */
    public void reSyncSavedShift(final LongId<ShiftRestEntity> shiftId) {

        ApiLoadShiftFromBackend.create(getView().getContext(), shiftId).on(exec).executeAsync(new Callback<ShiftDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable ShiftDbModel result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onShiftFound(result);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onShiftUpdateFailed(shiftId, reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }




    /************************** Start Visit (manually from long tap)  *****************************/
    private final Set<Long> startingVisits = new HashSet<>();

    public void startVisit(VisitDbModel visit) {
        final LongId<VisitEntity> visitId = visit.id();

        if (!startingVisits.add(visitId.getValue())) return;
        new ApiStartVisit(getView().getContext(), visitId).on(exec).executeAsync(new Callback<VisitDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
                getView().onStartVisitStartApi();
            }

            @Override
            public void onSuccess(@Nullable VisitDbModel result) {
                BroadcastController.broadcastVisitStateChanged(result);
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onVisitStarted(result);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onStartVisitFailed(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
                startingVisits.remove(visitId.getValue());
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onEndVisitStartApi();
                }
            }
        });
    }


    /**************************  Sync Visits back to the backend  *********************************/
    private boolean isVisitDataSyncing;
    public boolean syncVisitsBack(LongId<ShiftRestEntity> shiftId) {
        if (isVisitDataSyncing) {
            return false;
        } else {
            isVisitDataSyncing = true;
        }

        ApiSyncVisitsBack.forShift(shiftId).on(exec).executeAsync(new Callback<List<VisitDbModel>>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
                getView().onSyncVisitsStarted();
            }

            @Override
            public void onSuccess(@Nullable List<VisitDbModel> result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onSyncVisitsFinished(result.size());
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onSyncVisitsFinished(0);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
                isVisitDataSyncing = false;
            }
        });
        return true;
    }

    /**************************  Ad-Hoc Visit create  *********************************************/
    private boolean isAdHocCreating = false;
    public boolean createAdHocVisit(VisitCreationModel creationModel) {
        if (isAdHocCreating) {
            return false;
        } else {
            isAdHocCreating = true;
        }

        ApiCreateAdHocVisit.create(creationModel, true).on(exec).executeAsync(new Callback<VisitDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable VisitDbModel result) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onAdHocVisitCreated(result);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                ShiftActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onAdHocVisitNotCreated(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
                isAdHocCreating = false;
            }
        });
        return true;
    }


}
