package com.webssa.pda.sunmi.ui.presenter;

import android.os.Looper;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.webssa.library.core.error.AppError;
import com.webssa.library.core.processor.BackgroundExecutor;
import com.webssa.library.core.processor.model.Callback;
import com.webssa.library.core.processor.model.Disposable;
import com.webssa.pda.sunmi.broadcasting.BroadcastController;
import com.webssa.pda.sunmi.dataapi.ApiCompleteVisit;
import com.webssa.pda.sunmi.dataapi.ApiFailVisit;
import com.webssa.pda.sunmi.dataapi.ApiSelectAllGoodsFromLocalDB;
import com.webssa.pda.sunmi.dataapi.ApiSelectVisitLocally;
import com.webssa.pda.sunmi.model.LocalVisitResult;
import com.webssa.pda.sunmi.model.UntiedLocalNote;
import com.webssa.pda.sunmi.rest.model.LongId;
import com.webssa.pda.sunmi.rest.model.impl.SiteEntity;
import com.webssa.pda.sunmi.rest.model.impl.VisitEntity;
import com.webssa.pda.sunmi.storage.dbmodel.impl.PricedGoodDbModel;
import com.webssa.pda.sunmi.storage.dbmodel.impl.VisitDbModel;
import com.webssa.pda.sunmi.ui.activity.VisitEditorActivityView;
import com.webssa.pda.sunmi.ui.presenter.base.BasePresenter;

import java.util.List;

/**
 * Created by stanislav.perchenko on 11/15/2019, 11:02 AM.
 */
public class VisitEditorActivityPresenter extends BasePresenter<VisitEditorActivityView> {

    private final LongId<VisitEntity> visitId;
    private BackgroundExecutor exec;

    private boolean isVisitEditing;

    public VisitEditorActivityPresenter(@NonNull LongId<VisitEntity> visitId,
                                        @NonNull VisitEditorActivityView view) {
        super(view);
        assert (visitId != null);
        this.visitId = visitId;
    }

    @Override
    public void initializeView() {
        if (exec == null) exec = new BackgroundExecutor(Looper.getMainLooper());
        loadVisit();
    }

    @Override
    public synchronized void release() {
        super.release();
        if (exec != null) {
            exec.release();
            exec = null;
        }
    }

    private void loadVisit() {
        (new ApiSelectVisitLocally(getView().getContext(), visitId)).on(exec).executeAsync(new Callback<VisitDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable VisitDbModel result) {
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onVisitLoaded(result);
                    if(result != null) {
                        loadGoodCategories(result.site().id());
                    }
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onVisitLoaded(null);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }

    private void loadGoodCategories(LongId<SiteEntity> siteId) {
        (new ApiSelectAllGoodsFromLocalDB(siteId)).on(exec).executeAsync(new Callback<List<PricedGoodDbModel>>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable List<PricedGoodDbModel> goods) {
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null) && !goods.isEmpty()) {
                    v.onGoodsLoaded(goods);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                // Not implemented
                Log.e("TAG", reason.getUserMessage());
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
    }

    public boolean completeVisit(@NonNull final LocalVisitResult visitResult) {
        if (isVisitEditing) {
            return false;
        } else if (isReleased()) {
            throw new IllegalStateException("Already released");
        }
        ApiCompleteVisit.create(getView().getContext(), visitResult).on(exec).executeAsync(new Callback<VisitDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable VisitDbModel visit) {
                BroadcastController.broadcastVisitStateChanged(visit);
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onCompleteVisitCompleted(visitResult);
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onCompleteVisitFailed(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });

        return true;
    }


    public boolean failVisit(@NonNull LongId<VisitEntity> visitId, @NonNull UntiedLocalNote note) {
        if (isVisitEditing) {
            return false;
        } else if (isReleased()) {
            throw new IllegalStateException("Already released");
        }
        ApiFailVisit.create(getView().getContext(), visitId, note).on(exec).executeAsync(new Callback<VisitDbModel>() {
            @Override
            public void onSubscribe(@NonNull Disposable d) {
                registerApiDisposable(d);
            }

            @Override
            public void onSuccess(@Nullable VisitDbModel visit) {
                BroadcastController.broadcastVisitStateChanged(visit);
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onFailVisitCompleted();
                }
            }

            @Override
            public void onError(@NonNull AppError reason) {
                VisitEditorActivityView v = getView();
                if (!isReleased() && (v != null)) {
                    v.onFailVisitFailed(reason);
                }
            }

            @Override
            public void onComplete(@NonNull Disposable d) {
                unregisterApiDisposable(d);
            }
        });
        return true;
    }
}
