package com.webssa.pda.sunmi.ui.presenter.base;

import androidx.annotation.NonNull;

import com.webssa.library.core.processor.model.Disposable;
import com.webssa.library.core.processor.utils.CompositeDisposable;

import java.lang.ref.WeakReference;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import retrofit2.Call;

public abstract class BasePresenter<TView> {
    private WeakReference<TView> viewRef;

    private volatile boolean released;

    protected BasePresenter(@NonNull TView view) {
        if (view != null) {
            viewRef = new WeakReference<>(view);
        } else {
            throw new IllegalArgumentException("A valid instance of the View interface must be provided");
        }
    }

    protected TView getView() {
        return viewRef.get();
    }

    /**
     * This method starts initialization process of prepared UI.
     * It must be called from the Activity's onPostCreate()
     */
    public abstract void initializeView();

    /**
     * This method must be called from Activity's onDestroy()
     */
    public synchronized void release() {
        released = true;
        viewRef.clear();
        cancelAllInProgressRequests();
    }

    public synchronized boolean isReleased() {
        return released;
    }

    /**
     * This method checks if the presentor has been released already
     * and throws IllegalSztateException if so.
     * subclasses must call this method before start execution of any operation.
     */
    protected synchronized void checkReleased() {
        if (released) {
            throw new IllegalStateException("Already released!");
        }
    }














    private final List<Call> mRegisterredCalls = new LinkedList<>();
    private final CompositeDisposable mDisposableContainer = new CompositeDisposable();


    protected void registerApiDisposable(Disposable d) {
        synchronized (mRegisterredCalls) {
            mDisposableContainer.add(d);
        }
    }

    protected void unregisterApiDisposable(Disposable d) {
        synchronized (mRegisterredCalls) {
            mDisposableContainer.delete(d);
        }
    }


    protected void registerApiCall(Call<?> apiCall) {
        synchronized (mRegisterredCalls) {
            for (Call c : mRegisterredCalls) {
                if (c == apiCall) return;
            }
            mRegisterredCalls.add(apiCall);
        }
    }

    protected void unregisterApiCall(Call<?> apiCall) {
        synchronized (mRegisterredCalls) {
            for (Iterator<Call> itr = mRegisterredCalls.iterator(); itr.hasNext(); ) {
                if (itr.next() == apiCall) {
                    itr.remove();
                    return;
                }
            }
        }
    }

    protected void unregisterAndDismissApiCall(Call<?> apiCall) {
        synchronized (mRegisterredCalls) {
            for (Iterator<Call> itr = mRegisterredCalls.iterator(); itr.hasNext(); ) {
                if (itr.next() == apiCall) {
                    itr.remove();
                    if (!apiCall.isExecuted() && !apiCall.isCanceled()) {
                        apiCall.cancel();
                    }
                    return;
                }
            }
        }
    }

    protected void cancelAllInProgressRequests() {
        synchronized (mRegisterredCalls) {
            mDisposableContainer.clear();
            for (Call c : mRegisterredCalls) {
                if (!c.isExecuted() && !c.isCanceled()) {
                    c.cancel();
                }
            }
            mRegisterredCalls.clear();
        }
    }
}
