package com.webssa.pda.sunmi.util;

import android.os.Parcelable;

import androidx.annotation.NonNull;

import com.google.auto.value.AutoValue;

import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

@AutoValue
public abstract class DateNoTime implements Parcelable {
    public abstract int years();
    public abstract int months();
    public abstract int days();

    public static DateNoTime parseDateNoTime(String time) throws ParseException {
        try {
            String[] parts = time.split("-");
            return builder()
                    .setYears(Integer.parseInt(parts[0]))
                    .setMonths(Integer.parseInt(parts[1]))
                    .setDays(Integer.parseInt(parts[2]))
                    .build();
        } catch (Exception e) {
            String msg = "Error parsing Date. Must be yyyy-mm-dd. Reason - "+e.getMessage();
            System.err.println(msg);
            throw new ParseException(msg, 0);
        }
    }

    public static DateNoTime today() {
        Calendar now = Calendar.getInstance();
        return builder()
                .setYears(now.get(Calendar.YEAR))
                .setMonths(1+now.get(Calendar.MONTH)) // Month in calendar starts with 0
                .setDays(now.get(Calendar.DAY_OF_MONTH))
                .build();
    }

    public static Builder builder() {
        return new AutoValue_DateNoTime.Builder();
    }

    @AutoValue.Builder
    public abstract static class Builder {
        public abstract Builder setYears(int years);
        public abstract Builder setMonths(int months);
        public abstract Builder setDays(int days);

        abstract DateNoTime actualBuild();

        public DateNoTime build() {
            final DateNoTime ent = actualBuild();
            if (ent.years() < 2019 || ent.years() > 2119) {
                throw new IllegalStateException("Invalid years value - " + ent.years()+". Allowed range - [2019, 2119]");
            } else if (ent.months() < 1 || ent.months() > 12) {
                throw new IllegalStateException("Invalid months value - " + ent.months());
            } else if (ent.days() < 0 || ent.days() > 31) {
                throw new IllegalStateException("Invalid days value - " + ent.days());
            }
            ent.setupCalendarAndValidate();
            return ent;
        }
    }


    private Date cachedDate;

    public Date getDate() {
        if (cachedDate == null) {
            synchronized (this) {
                if (cachedDate == null) {
                    Calendar cal = Calendar.getInstance();
                    cal.set(years(), 1 + months(), days());
                    cachedDate = cal.getTime();
                }
            }
        }
        return new Date(cachedDate.getTime());
    }

    private String cachedTextForm;

    @NonNull
    @Override
    public String toString() {
        if (cachedTextForm == null) {
            synchronized (this) {
                if (cachedTextForm == null) {
                    cachedTextForm = String.format("%04d-%02d-%02d", years(), months(), days());
                }
            }
        }
        return cachedTextForm;
    }

    public String toStringDMY() {
        return String.format("%02d-%02d-%04d", days(), months(), years());
    }


    private Calendar dateCalendar;

    private void setupCalendarAndValidate() {
        dateCalendar = new GregorianCalendar(years(), months() - 1, 1);
        if (days() > dateCalendar.getActualMaximum(Calendar.DAY_OF_MONTH)) {
            throw new IllegalStateException("Day value is too large - " + toString());
        } else {
            dateCalendar.set(Calendar.DAY_OF_MONTH, days());
        }
    }
}
