package com.alperez.widget.picker;

import android.annotation.SuppressLint;
import android.content.Context;
import android.content.res.TypedArray;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Typeface;
import android.graphics.drawable.ColorDrawable;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.text.TextUtils;
import android.util.AttributeSet;
import android.view.MotionEvent;
import android.view.VelocityTracker;
import android.view.View;
import android.view.ViewConfiguration;
import android.view.ViewGroup;
import android.view.animation.DecelerateInterpolator;
import android.widget.OverScroller;

import androidx.annotation.Nullable;
import androidx.core.content.ContextCompat;

import com.alperez.R;

import java.util.ArrayList;
import java.util.List;

/**
 * Created by stanislav.perchenko on 2/7/2020, 12:56 PM.
 */
public class AlpNumberWheelPicker extends View {
    public interface OnValueChangeListener {

        /**
         * Called upon a change of the current value.
         *
         * @param picker The NumberPicker associated with this listener.
         * @param oldVal The previous value.
         * @param newVal The new value.
         */
        void onValueChange(AlpNumberWheelPicker picker, int oldVal, int newVal);
    }

    public interface OnScrollListener {

        /**
         * The view is not scrolling.
         */
        int SCROLL_STATE_IDLE = 0;

        /**
         * The user is scrolling using touch, and his finger is still on the screen.
         */
        int SCROLL_STATE_TOUCH_SCROLL = 1;

        /**
         * The user had previously been scrolling using touch and performed a fling.
         */
        int SCROLL_STATE_FLING = 2;

        /**
         * Callback invoked while the number picker scroll state has changed.
         *
         * @param picker The view whose scroll state is being reported.
         * @param scrollState The current scroll state. One of
         * [.SCROLL_STATE_IDLE],
         * [.SCROLL_STATE_TOUCH_SCROLL] or
         * [.SCROLL_STATE_IDLE].
         */
        void onScrollStateChange(AlpNumberWheelPicker picker, int scrollState);
    }

    public interface DisplayValueFormatter {
        String format(int value);
    }


    public static final float TOP_AND_BOTTOM_FADING_EDGE_STRENGTH = 0.9f;
    public static final int SNAP_SCROLL_DURATION = 300;
    public static final int SELECTOR_MAX_FLING_VELOCITY_ADJUSTMENT = 4;
    public static final int DEFAULT_ITEM_COUNT = 3;
    public static final int DEFAULT_TEXT_SIZE = 80;
    public static final float DEFAULT_SELECTED_TEXT_SCALE = 0.3f;

    private int mSelectorItemCount;
    private int mSelectorVisibleItemCount;
    private int mMinIndex;
    private int mMaxIndex;
    private Integer mMaxValidIndex;
    private Integer mMinValidIndex;
    private int mWheelMiddleItemIndex;
    private int mWheelVisibleItemMiddleIndex;
    private List<Integer> mSelectorItemIndices;
    private List<Boolean> mSelectorItemValidStatus;
    private int mCurrSelectedItemIndex;
    private boolean mWrapSelectorWheelPreferred;

    private final Paint mTextPaint = new Paint();
    private int mSelectedTextColor;
    private final int mUnSelectedTextColor;
    private final int mInvalidTextColor;
    private final int mTextSize;
    private String mTextAlign;

    private final OverScroller mOverScroller;
    private VelocityTracker mVelocityTracker;
    private final int mTouchSlop;
    private final int mMaximumFlingVelocity;
    private final int mMinimumFlingVelocity;

    private float mLastY = 0f;
    private boolean mIsDragging;
    private int mCurrentFirstItemOffset;
    private int mInitialFirstItemOffset = Integer.MIN_VALUE;
    private int mTextGapHeight;
    private int mItemHeight;
    private int mPreviousScrollerY;

    private OnValueChangeListener mOnValueChangeListener;
    private OnScrollListener mOnScrollListener;
    private DisplayValueFormatter mDisplayValueFormatter;

    private final boolean mFadingEdgeEnabled;
    private float mSelectedTextScale;
    private final int mTypefaceIndex;

    private String suggestedMaximumLengthText;


    private Drawable mSelectionDivider;
    private int mSelectionDividerHeight;

    /**
     * The distance between the two selection dividers.
     */
    private final int mSelectionDividersDistance;

    /**
     * The top of the top selection divider.
     */
    private int mTopSelectionDividerTop;

    /**
     * The bottom of the bottom selection divider.
     */
    private int mBottomSelectionDividerBottom;


    /**
     * The current scroll state of the number picker.
     */
    private int mScrollState = OnScrollListener.SCROLL_STATE_IDLE;

    public AlpNumberWheelPicker(Context context) {
        this(context, null, 0);
    }

    public AlpNumberWheelPicker(Context context, @Nullable AttributeSet attrs) {
        this(context, attrs, 0);
    }

    public AlpNumberWheelPicker(Context context, @Nullable AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);

        TypedArray a = context.obtainStyledAttributes(attrs, R.styleable.AlpNumberWheelPicker, defStyleAttr, 0);
        try {
            mSelectorVisibleItemCount = a.getInt(R.styleable.AlpNumberWheelPicker_wheelItemCount, DEFAULT_ITEM_COUNT);
            mSelectorItemCount = mSelectorVisibleItemCount + 2;
            mWheelMiddleItemIndex = (mSelectorItemCount - 1) / 2;
            mWheelVisibleItemMiddleIndex = (mSelectorVisibleItemCount - 1) / 2;
            mSelectorItemIndices = new ArrayList<>(mSelectorItemCount);
            mSelectorItemValidStatus = new ArrayList<>(mSelectorItemCount);

            mMinIndex = a.getInt(R.styleable.AlpNumberWheelPicker_min, Integer.MIN_VALUE);
            mMaxIndex = a.getInt(R.styleable.AlpNumberWheelPicker_max, Integer.MAX_VALUE);
            mMaxValidIndex = (a.hasValue(R.styleable.AlpNumberWheelPicker_maxValidIndex)) ? a.getInt(R.styleable.AlpNumberWheelPicker_maxValidIndex, 0) : null;
            mMinValidIndex = (a.hasValue(R.styleable.AlpNumberWheelPicker_minValidIndex)) ? a.getInt(R.styleable.AlpNumberWheelPicker_minValidIndex, 0) : null;

            mWrapSelectorWheelPreferred = a.getBoolean(R.styleable.AlpNumberWheelPicker_wrapSelectorWheel, false);
            mSelectedTextScale = a.getFloat(R.styleable.AlpNumberWheelPicker_selectedTextScale, DEFAULT_SELECTED_TEXT_SCALE);

            mOverScroller = new OverScroller(context, new DecelerateInterpolator(2.5f));

            ViewConfiguration configuration = ViewConfiguration.get(context);
            mTouchSlop = configuration.getScaledTouchSlop();
            mMaximumFlingVelocity = configuration.getScaledMaximumFlingVelocity() / SELECTOR_MAX_FLING_VELOCITY_ADJUSTMENT;
            mMinimumFlingVelocity = configuration.getScaledMinimumFlingVelocity();

            mInvalidTextColor = a.getColor(R.styleable.AlpNumberWheelPicker_invalidTextColor, ContextCompat.getColor(context, R.color.invalid_text_default_color));
            mSelectedTextColor = a.getColor(R.styleable.AlpNumberWheelPicker_selectedTextColor, ContextCompat.getColor(context, R.color.selected_text_default_color));
            mUnSelectedTextColor = a.getColor(R.styleable.AlpNumberWheelPicker_textColor, ContextCompat.getColor(context, R.color.unselected_text_default_color));

            mTextSize = a.getDimensionPixelSize(R.styleable.AlpNumberWheelPicker_textSize, DEFAULT_TEXT_SIZE);
            mTextAlign = convertTextAlignIntToString(a.getInt(R.styleable.AlpNumberWheelPicker_align, 1));

            mFadingEdgeEnabled = a.getBoolean(R.styleable.AlpNumberWheelPicker_fadingEdgeEnabled, true);
            mTypefaceIndex = a.getInt(R.styleable.AlpNumberWheelPicker_typeface, 0);

            suggestedMaximumLengthText = a.getString(R.styleable.AlpNumberWheelPicker_suggestedMaximumLengthText);

            mSelectionDividersDistance = a.getDimensionPixelSize(R.styleable.AlpNumberWheelPicker_selectionDividersDistance, 0);

            Drawable selectionDivider = a.getDrawable(R.styleable.AlpNumberWheelPicker_selectionDivider);
            if (selectionDivider == null && a.hasValue(R.styleable.AlpNumberWheelPicker_dividers_color)) {
                int divColor = a.getColor(R.styleable.AlpNumberWheelPicker_dividers_color, 0);
                selectionDivider = new ColorDrawable(divColor);
            }

            if (selectionDivider != null) {
                selectionDivider.setCallback(this);
                if (Build.VERSION.SDK_INT >= 23) selectionDivider.setLayoutDirection(getLayoutDirection());
                if (selectionDivider.isStateful()) {
                    selectionDivider.setState(getDrawableState());
                }
                mSelectionDividerHeight = a.getDimensionPixelSize(R.styleable.AlpNumberWheelPicker_dividers_thick, 1);
            }

            mSelectionDivider = selectionDivider;

            setupTextPaint();
        } finally {
            a.recycle();

            initializeSelectorWheelIndices();
        }

    }

    private String convertTextAlignIntToString(int textAlignInt) {
        switch (textAlignInt) {
            case 0:
                return "LEFT";
            case 2:
                return "RIGHT";
            case 1:
            default:
                return "CENTER";
        }
    }

    private void setupTextPaint() {
        mTextPaint.setAntiAlias(true);
        mTextPaint.setTextSize(mTextSize);
        mTextPaint.setTextAlign(Paint.Align.valueOf(mTextAlign));
        mTextPaint.setStyle(Paint.Style.FILL_AND_STROKE);
        switch (mTypefaceIndex) { // for the constant values please check the WheelPicker_typeface in the attrs.xml
            case 1:
                mTextPaint.setTypeface(Typeface.SANS_SERIF);
                break;
            case 2:
                mTextPaint.setTypeface(Typeface.SERIF);
                break;
            default:
                mTextPaint.setTypeface(Typeface.DEFAULT);
        }
    }

    @Override
    protected void onLayout(boolean changed, int left, int top, int right, int bottom) {
        super.onLayout(changed, left, top, right, bottom);
        if (changed) {

            int divDist = mSelectionDividersDistance;
            if (divDist <= 0 && mSelectorVisibleItemCount > 0) {
                divDist = (bottom - top) / mSelectorVisibleItemCount;
            }

            mTopSelectionDividerTop = (bottom - top - divDist) / 2 - mSelectionDividerHeight;
            mBottomSelectionDividerBottom = mTopSelectionDividerTop + divDist + 2 * mSelectionDividerHeight;


            // need to do all this when we know our size
            initializeSelectorWheel();
            initializeFadingEdges();
        }
    }

    @Override
    protected void drawableStateChanged() {
        super.drawableStateChanged();

        final Drawable selectionDivider = mSelectionDivider;
        if (selectionDivider != null && selectionDivider.isStateful() && selectionDivider.setState(getDrawableState())) {
            invalidateDrawable(selectionDivider);
        }
    }

    @Override
    public void jumpDrawablesToCurrentState() {
        super.jumpDrawablesToCurrentState();

        if (mSelectionDivider != null) {
            mSelectionDivider.jumpToCurrentState();
        }
    }

    @Override
    protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        // Try greedily to fit the max width and height.
        ViewGroup.LayoutParams lp = getLayoutParams();
        if (lp == null) lp = new ViewGroup.LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT);

        int width = calculateSize(getSuggestedMinimumWidth(), lp.width, widthMeasureSpec);
        int height = calculateSize(getSuggestedMinimumHeight(), lp.height, heightMeasureSpec);

        setMeasuredDimension(width, height);
    }

    @Override
    protected int getSuggestedMinimumWidth() {
        int width = super.getSuggestedMinimumWidth();
        if (mSelectorVisibleItemCount > 0) {
            width = Math.max(width, computeMaximumContentWidth() + getPaddingLeft() + getPaddingRight());
        }
        return width;
    }

    @Override
    protected int getSuggestedMinimumHeight() {
        int height = super.getSuggestedMinimumHeight();
        if (mSelectorVisibleItemCount > 0) {
            Paint.FontMetricsInt fmi = mTextPaint.getFontMetricsInt();
            int fontH = fmi.descent - fmi.ascent;
            height = Math.max(height, fontH * mSelectorVisibleItemCount);
        }
        return height;
    }

    private int computeMaximumContentWidth() {
        try {
            mTextPaint.setTextSize(mTextSize * 1.3f);
            float width = (TextUtils.isEmpty(suggestedMaximumLengthText))
                    ? Math.max(mTextPaint.measureText(getDisplayValue(mMinIndex)), mTextPaint.measureText(getDisplayValue(mMaxIndex)))
                    : mTextPaint.measureText(suggestedMaximumLengthText);
            return (int) Math.ceil(width);
        } finally {
            mTextPaint.setTextSize(mTextSize);
        }
    }

    private int calculateSize(int suggestedSize, int paramSize, int measureSpec) {
        int result = 0;
        final int size = MeasureSpec.getSize(measureSpec);
        final int mode = MeasureSpec.getMode(measureSpec);
        switch (MeasureSpec.getMode(mode)) {
            case MeasureSpec.AT_MOST:
                if (paramSize == ViewGroup.LayoutParams.WRAP_CONTENT)
                    result = Math.min(suggestedSize, size);
                else if (paramSize == ViewGroup.LayoutParams.MATCH_PARENT)
                    result = size;
                else {
                    result = Math.min(paramSize, size);
                }
                break;
            case MeasureSpec.EXACTLY:
                result = size;
                break;
            case MeasureSpec.UNSPECIFIED:
                result = suggestedSize;
                break;
        }
        return result;
    }

    private void initializeSelectorWheel() {
        mItemHeight = getItemHeight();
        final int textHeight = computeTextHeight();
        mTextGapHeight = getGapHeight();

        final int visibleMiddleItemPos = mItemHeight * mWheelVisibleItemMiddleIndex + (mItemHeight + textHeight) / 2;
        mCurrentFirstItemOffset = mInitialFirstItemOffset = visibleMiddleItemPos - mItemHeight * mWheelMiddleItemIndex;
    }

    private void initializeFadingEdges() {
        setVerticalFadingEdgeEnabled(mFadingEdgeEnabled);
        if (mFadingEdgeEnabled) {
            setFadingEdgeLength((getBottom() - getTop() - mTextSize) / 2);
        }
    }

    private void initializeSelectorWheelIndices() {
        mSelectorItemIndices.clear();
        mSelectorItemValidStatus.clear();
        final int selection = (mMinValidIndex == null || mMinValidIndex < mMinIndex)
                ? (mMinIndex <= 0) ? 0 : mMinIndex
                : (mMinValidIndex <= 0) ? 0 : mMinValidIndex;

        for (int i=0; i < mSelectorItemCount; i++) {
            int selectorIndex = selection - mWheelMiddleItemIndex + i;
            if (mWrapSelectorWheelPreferred) {
                selectorIndex = getWrappedSelectorIndex(selectorIndex);
            }
            mSelectorItemIndices.add(selectorIndex);
            mSelectorItemValidStatus.add(isValidValue(selectorIndex));
        }

        onSelectionChanged(mSelectorItemIndices.get(mWheelMiddleItemIndex));
    }

    private int getWrappedSelectorIndex(int selectorIndex) {
        if (selectorIndex > mMaxIndex) {
            return mMinIndex + (selectorIndex - mMaxIndex) % (mMaxIndex - mMinIndex + 1) - 1;
        } else if (selectorIndex < mMinIndex) {
            return mMaxIndex - (mMinIndex - selectorIndex) % (mMaxIndex - mMinIndex + 1) + 1;
        }
        return selectorIndex;
    }

    private int getItemHeight() {
        return getHeight() / (mSelectorItemCount - 2);
    }

    private int getGapHeight() {
        return getItemHeight() - computeTextHeight();
    }

    private int computeTextHeight() {
        Paint.FontMetricsInt metricsInt = mTextPaint.getFontMetricsInt();
        return Math.abs(metricsInt.bottom + metricsInt.top);
    }

    @Override
    protected float getTopFadingEdgeStrength() {
        return TOP_AND_BOTTOM_FADING_EDGE_STRENGTH;
    }

    @Override
    protected float getBottomFadingEdgeStrength() {
        return TOP_AND_BOTTOM_FADING_EDGE_STRENGTH;
    }


    //*********************************************************************************************/
    //**************************  Drawing content  ************************************************/
    //*********************************************************************************************/
    @Override
    protected void onDraw(Canvas canvas) {
        super.onDraw(canvas);
        if (mSelectorItemIndices.isEmpty()) return;

        final int itemHeight = getItemHeight();

        final float x;
        switch (mTextPaint.getTextAlign()) {
            case LEFT:
                x = getPaddingLeft();
                break;
            case RIGHT:
                x = getRight() - getLeft() - getPaddingRight();
                break;
            case CENTER:
            default:
                x = (getRight() - getLeft()) / 2f;
        }

        final int topIndexDiffToMid = mWheelVisibleItemMiddleIndex;
        final int bottomIndexDiffToMid = mSelectorVisibleItemCount - mWheelVisibleItemMiddleIndex - 1;
        final int maxIndexDiffToMid = Math.max(topIndexDiffToMid, bottomIndexDiffToMid);

        float y = mCurrentFirstItemOffset;
        for (int i=0; i<mSelectorItemIndices.size(); i++) {
            float offsetToMiddle = Math.abs(y - (mInitialFirstItemOffset + mWheelMiddleItemIndex * itemHeight));

            float scale = 1f;
            if (maxIndexDiffToMid != 0) {
                scale = mSelectedTextScale * (itemHeight * maxIndexDiffToMid - offsetToMiddle) / (itemHeight * maxIndexDiffToMid) + 1;
            }

            if (mSelectorItemValidStatus.get(i)) {
                mTextPaint.setColor((offsetToMiddle < mItemHeight / 2) ? mSelectedTextColor : mUnSelectedTextColor);
            } else {
                mTextPaint.setColor(mInvalidTextColor);
            }

            final int value = mSelectorItemIndices.get(i);

            canvas.save();
            canvas.scale(scale, scale, x, y);
            canvas.drawText(getDisplayValue(value), x, y, mTextPaint);
            canvas.restore();

            y += itemHeight;
        }

        /*if (mDrawSelectedValueDividers) {
            canvas.drawLine(0, mSelDividerTopY, canvas.getWidth(), mSelDividerTopY, mSelDividerPaint);
            canvas.drawLine(0, mSelDividerBotY, canvas.getWidth(), mSelDividerBotY, mSelDividerPaint);
        }*/

        // draw the selection dividers
        if (mSelectionDivider != null) {
            // draw the top divider
            int topOfTopDivider = mTopSelectionDividerTop;
            int bottomOfTopDivider = topOfTopDivider + mSelectionDividerHeight;
            mSelectionDivider.setBounds(0, topOfTopDivider, getRight(), bottomOfTopDivider);
            mSelectionDivider.draw(canvas);

            // draw the bottom divider
            int bottomOfBottomDivider = mBottomSelectionDividerBottom;
            int topOfBottomDivider = bottomOfBottomDivider - mSelectionDividerHeight;
            mSelectionDivider.setBounds(0, topOfBottomDivider, getRight(), bottomOfBottomDivider);
            mSelectionDivider.draw(canvas);
        }

    }

    public String getDisplayValue(int value) {

        if (!mWrapSelectorWheelPreferred && ((value > mMaxIndex) || (value < mMinIndex))) {
            return "";
        } else {
            final int validValue = mWrapSelectorWheelPreferred ? getWrappedSelectorIndex(value) : value;
            return (mDisplayValueFormatter == null) ? Integer.toString(validValue) : mDisplayValueFormatter.format(validValue);
        }
    }

    //*********************************************************************************************/
    //********************************  SCROLLING  ************************************************/
    //*********************************************************************************************/



    @SuppressLint("ClickableViewAccessibility")
    @Override
    public boolean onTouchEvent(MotionEvent event) {

        if (mVelocityTracker == null) {
            mVelocityTracker = VelocityTracker.obtain();
        }
        mVelocityTracker.addMovement(event);

        final int action = event.getActionMasked();
        switch (action) {
            case MotionEvent.ACTION_DOWN:
                mLastY = event.getY();
                getParent().requestDisallowInterceptTouchEvent(true);
                if (!mOverScroller.isFinished()) {
                    mOverScroller.forceFinished(true);
                }

                break;

            case MotionEvent.ACTION_MOVE:
                int deltaY = Math.round(event.getY() - mLastY);
                if (!mIsDragging && Math.abs(deltaY) > mTouchSlop) {

                    if (deltaY > 0) {
                        deltaY -= mTouchSlop;
                    } else {
                        deltaY += mTouchSlop;
                    }
                    setScrollState(OnScrollListener.SCROLL_STATE_TOUCH_SCROLL);
                    mIsDragging = true;
                }

                if (mIsDragging) {
                    scrollBy(0, deltaY);
                    invalidate();
                    mLastY = event.getY();
                }
                break;

            case MotionEvent.ACTION_UP:
                if (mIsDragging) {
                    mIsDragging = false;

                    mVelocityTracker.computeCurrentVelocity(1000, mMaximumFlingVelocity);
                    final int velocity = (int) mVelocityTracker.getYVelocity();

                    if (Math.abs(velocity) > mMinimumFlingVelocity) {
                        mPreviousScrollerY = 0;
                        mOverScroller.fling(getScrollX(), getScrollY(), 0, velocity, 0, 0, Integer.MIN_VALUE, Integer.MAX_VALUE, 0, (int)(getItemHeight() * 0.7)
                        );
                        invalidateOnAnimation();
                        setScrollState(OnScrollListener.SCROLL_STATE_FLING);
                    } else {
                        alignItem();
                    }
                } else {
                    //click event
                    final int selectorIndexOffset = Math.round(event.getY() / mItemHeight - mWheelVisibleItemMiddleIndex);
                    changeValueBySteps(selectorIndexOffset);
                }

                mVelocityTracker.recycle();
                mVelocityTracker = null;
                break;

            case MotionEvent.ACTION_CANCEL:
                mIsDragging = false;
                mVelocityTracker.recycle();
                mVelocityTracker = null;
                break;
        }
        return true;
    }


    private void changeValueBySteps(int steps) {
        mPreviousScrollerY = 0;
        mOverScroller.startScroll(0, 0, 0, -mItemHeight * steps, SNAP_SCROLL_DURATION);
        invalidate();
    }

    @Override
    public void computeScroll() {
        super.computeScroll();
        if (mOverScroller.computeScrollOffset()) {
            //final int x = mOverScroller.getCurrX();   //TODO - checked
            final int y = mOverScroller.getCurrY();
            if (mPreviousScrollerY == 0) {
                mPreviousScrollerY = mOverScroller.getStartY();
            }
            //scrollBy(x, y - mPreviousScrollerY);  //TODO - checked
            scrollBy(0, y - mPreviousScrollerY);
            mPreviousScrollerY = y;
            invalidate();
        } else if (!mIsDragging) {
            alignItem();

            /*mPreviousScrollerY = 0;
            int deltaY = mInitialFirstItemOffset - mCurrentFirstItemOffset;
            if (Math.abs(deltaY) > mItemHeight / 2) {
                if  (deltaY > 0) {
                    deltaY -= mItemHeight;
                } else {
                    deltaY += mItemHeight;
                }
            }

            if (deltaY != 0) {
                //mOverScroller.startScroll(getScrollX(), getScrollY(), 0, deltaY, 800);
                mOverScroller.startScroll(0, 0, 0, deltaY, 800);
                invalidateOnAnimation();
            }*/

            setScrollState(OnScrollListener.SCROLL_STATE_IDLE);
        }
    }

    private void alignItem() {
        int deltaY = mInitialFirstItemOffset - mCurrentFirstItemOffset;
        if (deltaY != 0) {
            mPreviousScrollerY = 0;
            if (Math.abs(deltaY) > mItemHeight / 2) {
                deltaY += (deltaY > 0) ? -mItemHeight : mItemHeight;
            }
            mOverScroller.startScroll(0, 0, 0, deltaY, 800);
            invalidateOnAnimation();
        }
    }

    /*
    * private boolean ensureScrollWheelAdjusted() {
        // adjust to the closest value
        int deltaY = mInitialScrollOffset - mCurrentScrollOffset;
        if (deltaY != 0) {
            mPreviousScrollerY = 0;
            if (Math.abs(deltaY) > mSelectorElementHeight / 2) {
                deltaY += (deltaY > 0) ? -mSelectorElementHeight : mSelectorElementHeight;
            }
            mAdjustScroller.startScroll(0, 0, 0, deltaY, SELECTOR_ADJUSTMENT_DURATION_MILLIS);
            invalidate();
            return true;
        }
        return false;
    }*/

    private void setScrollState(int scrollState) {
        if (mScrollState != scrollState) {
            mScrollState = scrollState;
            if (mOnScrollListener != null) mOnScrollListener.onScrollStateChange(this, scrollState);
        }
    }

    @SuppressLint("ObsoleteSdkInt")
    private void invalidateOnAnimation() {
        if (Build.VERSION.SDK_INT >= 16) {
            postInvalidateOnAnimation();
        } else {
            invalidate();
        }
    }


    @Override
    public void scrollBy(int x, int y) {

        if (y == 0) return;

        final int gap = mTextGapHeight;

        if (!mWrapSelectorWheelPreferred && (y > 0) && (mCurrSelectedItemIndex <= mMinIndex || (mMinValidIndex != null && mCurrSelectedItemIndex <= mMinValidIndex))) {
            if ((mCurrentFirstItemOffset + y - mInitialFirstItemOffset) < (gap / 2)) {
                mCurrentFirstItemOffset += y;
            } else {
                mCurrentFirstItemOffset = mInitialFirstItemOffset + (gap / 2);
                if (!mOverScroller.isFinished() && !mIsDragging) {
                    mOverScroller.abortAnimation();
                }
            }
            return;
        }

        if (!mWrapSelectorWheelPreferred && (y < 0) && (mCurrSelectedItemIndex >= mMaxIndex || (mMaxValidIndex != null && mCurrSelectedItemIndex >= mMaxValidIndex))) {
            if ((mCurrentFirstItemOffset + y - mInitialFirstItemOffset) > -(gap / 2))
                mCurrentFirstItemOffset += y;
            else {
                mCurrentFirstItemOffset = mInitialFirstItemOffset - (gap / 2);
                if (!mOverScroller.isFinished() && !mIsDragging) {
                    mOverScroller.abortAnimation();
                }
            }
            return;
        }

        mCurrentFirstItemOffset += y;
        while (mCurrentFirstItemOffset - mInitialFirstItemOffset > gap) {
            mCurrentFirstItemOffset -= mItemHeight;
            decrementSelectorIndexes();
            if (!mWrapSelectorWheelPreferred && (mCurrSelectedItemIndex <= mMinIndex || (mMinValidIndex != null && mCurrSelectedItemIndex <= mMinValidIndex))) {
                mCurrentFirstItemOffset = mInitialFirstItemOffset;
            }
        }
        while (mCurrentFirstItemOffset - mInitialFirstItemOffset < -gap) {
            mCurrentFirstItemOffset += mItemHeight;
            incrementSelectorIndexes();
            if (!mWrapSelectorWheelPreferred && (mCurrSelectedItemIndex >= mMaxIndex || (mMaxValidIndex != null && mCurrSelectedItemIndex >= mMaxValidIndex))) {
                mCurrentFirstItemOffset = mInitialFirstItemOffset;
            }
        }



    }



    private void incrementSelectorIndexes() {
        final int size = mSelectorItemIndices.size();
        for (int i=0; i < (size - 1); i++) {
            mSelectorItemIndices.set(i, mSelectorItemIndices.get(i + 1));
            mSelectorItemValidStatus.set(i, mSelectorItemValidStatus.get(i + 1));
        }
        int nextScrollSelectorIndex = mSelectorItemIndices.get(size - 2) + 1;
        if (mWrapSelectorWheelPreferred && nextScrollSelectorIndex > mMaxIndex) {
            nextScrollSelectorIndex = mMinIndex;
        }
        mSelectorItemIndices.set(size - 1, nextScrollSelectorIndex);
        mSelectorItemValidStatus.set(size - 1, isValidValue(nextScrollSelectorIndex));
        onSelectionChanged(mSelectorItemIndices.get(mWheelMiddleItemIndex));
    }

    private void decrementSelectorIndexes() {
        final int size = mSelectorItemIndices.size();
        for (int i = (size - 1); i > 0; i--) {
            mSelectorItemIndices.set(i, mSelectorItemIndices.get(i - 1));
            mSelectorItemValidStatus.set(i, mSelectorItemValidStatus.get(i - 1));
        }
        int nextScrollSelectorIndex = mSelectorItemIndices.get(1) - 1;
        if (mWrapSelectorWheelPreferred && nextScrollSelectorIndex < mMinIndex) {
            nextScrollSelectorIndex = mMaxIndex;
        }
        mSelectorItemIndices.set(0, nextScrollSelectorIndex);
        mSelectorItemValidStatus.set(0, isValidValue(nextScrollSelectorIndex));
        onSelectionChanged(mSelectorItemIndices.get(mWheelMiddleItemIndex));
    }

    private void onSelectionChanged(int current) {
        if (mCurrSelectedItemIndex != current) {
            int previous = mCurrSelectedItemIndex;
            mCurrSelectedItemIndex = current;
            if (mOnValueChangeListener != null) mOnValueChangeListener.onValueChange(this, previous, current);
        }
    }






    //*********************************************************************************************/
    //*********************************************************************************************/
    //*********************************************************************************************/

    public void setSelection(int value) {
        if (mCurrSelectedItemIndex != value) {
            mCurrSelectedItemIndex = value;
            mSelectorItemIndices.clear();

            for (int i=0; i<mSelectorItemCount; i++) {
                int selectorIndex = value - mWheelMiddleItemIndex + i;
                if (mWrapSelectorWheelPreferred) {
                    selectorIndex = getWrappedSelectorIndex(selectorIndex);
                }
                mSelectorItemIndices.add(selectorIndex);
            }
            //onSelectionChanged(mSelectorItemIndices.get(mWheelMiddleItemIndex));
            invalidate();
        }
    }

    public void setOnValueChangedListener(OnValueChangeListener l) {
        mOnValueChangeListener = l;
    }

    public void setOnScrollListener(OnScrollListener l ) {
        mOnScrollListener = l;
    }

    public void setDisplayValueFormatter(DisplayValueFormatter formatter) {
        this.mDisplayValueFormatter = formatter;
        invalidate();
    }

    public void smoothScrollTo(int position) {

        //Validate requested position position
        final int realPosition;
        if (mWrapSelectorWheelPreferred) {
            realPosition = getWrappedSelectorIndex(position);
        } else if (mMaxValidIndex == null && position > mMaxIndex) {
            realPosition = mMaxIndex;
        } else if (mMaxValidIndex != null && position > mMaxValidIndex) {
            realPosition = mMaxValidIndex;
        } else if (mMinValidIndex == null && position < mMinIndex) {
            realPosition = mMinIndex;
        }else if (mMinValidIndex != null && position < mMinValidIndex) {
            realPosition = mMinValidIndex;
        } else {
            realPosition = position;
        }

        changeValueBySteps(realPosition - mCurrSelectedItemIndex);
    }

    public void smoothIncrement() {
        changeValueBySteps(1);
    }

    public void smoothDecrement() {
        changeValueBySteps(-1);
    }

    /**
     * Set a custom typeface object for the text
     *
     * @param typeface the custom typeface object
     */
    public void setTypeface(Typeface typeface) {
        if (mTextPaint.getTypeface() != typeface) {
            mTextPaint.setTypeface(typeface);
            invalidate();
        }
    }

    /**
     * Sets whether the selector wheel shown during flinging/scrolling should
     * wrap around the {@link android.widget.NumberPicker#getMinValue()} and
     * {@link android.widget.NumberPicker#getMaxValue()} values.

     * @param wrapSelectorWheel Whether to wrap.
     */
    public void setWrapSelectorWheel(boolean wrapSelectorWheel) {
        if (mWrapSelectorWheelPreferred != wrapSelectorWheel) {
            mWrapSelectorWheelPreferred = wrapSelectorWheel;
            invalidate();
        }
    }

    /**
     * Gets whether the selector wheel wraps when reaching the min/max value.
     *
     * @return True if the selector wheel wraps.
     *
     * @see .getMinValue
     * @see .getMaxValue
     */
    public boolean isWrapSelectorWheel() {
        return mWrapSelectorWheelPreferred;
    }


    /**
     * Set how many visible item show in the picker
     */
    public void setWheelItemCount(int count) {
        //TODO Use this in constructor !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if (mSelectorItemCount != (count + 2)) {
            mSelectorItemCount = count + 2;
            mWheelMiddleItemIndex = (mSelectorItemCount - 1) / 2;
            mSelectorVisibleItemCount = mSelectorItemCount - 2;
            mWheelVisibleItemMiddleIndex = (mSelectorVisibleItemCount - 1) / 2;
            mSelectorItemIndices = new ArrayList<>(mSelectorItemCount);
            mSelectorItemValidStatus = new ArrayList<>(mSelectorItemCount);
            reset();
            invalidate();
        }
    }

    /**
     * Set color for current selected item
     */
    public void setSelectedTextColor(int color) {
        if (mSelectedTextColor != color) {
            mSelectedTextColor = color;
            invalidate();
        }
    }

    public void setSuggestedMaximumLengthText(String suggestedMaximumLengthText) {
        if (!TextUtils.equals(this.suggestedMaximumLengthText, suggestedMaximumLengthText)) {
            this.suggestedMaximumLengthText = suggestedMaximumLengthText;
            requestLayout();
            invalidate();
        }
    }

    public void setMaxValue(int max) {
        if (mMaxIndex != max) {
            mMaxIndex = max;
            invalidate();
        }
    }

    public int getMaxValue() {
        return mMaxIndex;
    }

    public void setMinValue(int min) {
        if (mMinIndex != min) {
            mMinIndex = min;
            invalidate();
        }
    }

    public int getMinValue() {
        return mMinIndex;
    }

    public void setMinValidValue(Integer minValid) {
        if (((mMinValidIndex != null) && !mMinValidIndex.equals(minValid)) || ((minValid != null) && !minValid.equals(mMinValidIndex))) {
            mMinValidIndex = minValid;
            invalidate();
        }
    }

    public Integer getMinValidValue() {
        return mMinValidIndex;
    }

    public void setMaxValidValue(Integer maxValid) {
        if (((mMaxValidIndex != null) && !mMaxValidIndex.equals(maxValid)) || ((maxValid != null) && !maxValid.equals(mMaxValidIndex))) {
            mMaxValidIndex = maxValid;
            invalidate();
        }
    }

    public Integer getMaxValidValue() {
        return mMaxValidIndex;
    }

    public int getCurrentSelection() {
        return mCurrSelectedItemIndex;
    }

    public void reset() {
        initializeSelectorWheelIndices();
        initializeSelectorWheel();
        invalidate();
    }

    private boolean isValidValue(int value) {
        return !((mMinValidIndex != null && value < mMinValidIndex) || (mMaxValidIndex != null && value > mMaxValidIndex));
    }

}
