package com.webssa.library.attachment;

import android.content.Context;
import android.net.Uri;

import com.webssa.library.attachment.model.AttachmentImportModel;
import com.webssa.library.attachment.model.AttachmentType;
import com.webssa.library.attachment.util.AttachmentUtils;
import com.webssa.library.attachment.util.FileUtils;
import com.webssa.library.core.error.AppError;
import com.webssa.library.core.error.LocalError;
import com.webssa.library.util.AsyncTaskCompat;

import org.jetbrains.annotations.Contract;

import java.io.File;
import java.io.IOException;
import java.util.Date;

import androidx.annotation.NonNull;

/**
 * This is the base class for all importers for all types of attachments
 * It declares setter for attachment source,
 * the import() method which is the actual import launcher,
 * and the callback interface for passing results.
 *
 * All other attachment type-specific setting may be declared
 * in concrete importer implementations
 *
 * Created by stanislav.perchenko on 11/7/2016.
 */

public abstract class BaseAttachmentImporter {

    private final Context mContext;
    private final Uri sourceUri;
    private OnImportResultListener callback;


    //----  Result data  ----
    private Exception errorOccurred;
    private AttachmentImportModel attachment;
    private File[] coverOptions;




    public BaseAttachmentImporter(Context ctx, Uri sourceUri) {
        if (sourceUri == null) throw new IllegalArgumentException("A valid source Uri must be provided");
        this.sourceUri = sourceUri;
        mContext = ctx;
    }

    public abstract BaseAttachmentImporter downscaleToMaxPixelCount(int maxPixelCount);


    protected Context getContext() {
        return mContext;
    }



    @Contract("null -> fail")
    public final void startImport(@NonNull OnImportResultListener callback) {
        if (callback == null) {
            throw new IllegalArgumentException("A valid instance of the OnImportResultListener must be provided");
        } else {
            this.callback = callback;
        }

        new AsyncTaskCompat<Void, Void, Boolean>() {


            @Override
            protected Boolean doInBackground(Void... voids) {
                try {
                    onImportInBackground(sourceUri);
                    return true;
                } catch (Exception e) {
                    errorOccurred = e;
                    return false;
                }
            }

            @Override
            protected void onPostExecute(Boolean result) {
                handleResult(result);
            }

            @Override
            protected void onCancelled(Boolean aBoolean) {
                errorOccurred = new Exception("Import task was cancelled");
                handleResult(false);
            }
        }.safeExecute();

    }

    protected abstract void onImportInBackground(Uri sourceUri) throws Exception;

    protected void setResult(File importedAttachment, String mimeType, int durationMillis, String hashName, AttachmentType aType) {
        Date d = new Date(importedAttachment.lastModified());
        this.attachment = AttachmentImportModel.create(importedAttachment.getAbsolutePath(), mimeType, durationMillis, hashName, aType, d);
    }

    protected void setCoverOptions(File[] coverOptions) {
        this.coverOptions = coverOptions;
    }


    private void handleResult(boolean result) {
        if (result) {
            handleSuccess();
        } else {
            handleError();
        }
    }

    private void handleSuccess() {
        if (attachment == null) {
            throw new IllegalStateException("The setResult() method must be called by a concrete implementation at the end of importing process");
        }

        callback.onImport(attachment, coverOptions);
    }

    private void handleError() {
        AppError error = LocalError.Companion.create(mContext, errorOccurred);
        callback.onError(error);
    }






    protected File getAttachmentWorkDirectory(Context ctx) {
        try {
            return FileUtils.getFinalCacheDirectory(ctx, AttachmentUtils.ATTACHMENT_WORK_DIRECTORY);
        } catch(IOException e) {
            return null;
        }
    }

    protected File getAttachmentCacheDirectory(Context ctx) {
        try {
            return FileUtils.getFinalCacheDirectory(ctx, AttachmentUtils.DIRECTORY_ATTACHMENTS);
        } catch(IOException e) {
            return null;
        }
    }


}

