package com.webssa.library.attachment.imports;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.net.Uri;
import android.util.Base64;
import android.util.Log;

import com.webssa.library.attachment.BaseAttachmentImporter;
import com.webssa.library.attachment.model.AttachmentType;
import com.webssa.library.attachment.util.AttachmentUtils;
import com.webssa.library.attachment.util.FileUtils;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;

/**
 * Created by stanislav.perchenko on 11/7/2016.
 */

public class ImageAttachmentImporter extends BaseAttachmentImporter {

    private int maxPixelCount;


    public ImageAttachmentImporter(Context ctx, Uri sourceUri) {
        super(ctx, sourceUri);
    }

    @Override
    public ImageAttachmentImporter downscaleToMaxPixelCount(int maxPixelCount) {
        this.maxPixelCount = maxPixelCount;
        return this;
    }


    @Override
    public void onImportInBackground(Uri uri) throws Exception {
        File originalImage = FileUtils.getNewRandomFile(getAttachmentCacheDirectory(getContext()));
        if (originalImage == null) {
            throw new Exception("Error create cache file for importing content");
        }


        InputStream iStream = null;
        BufferedOutputStream tmpOs = null;
        try {
            tmpOs = new BufferedOutputStream(new FileOutputStream(originalImage));
            iStream = (uri.toString().startsWith("content:"))
                    ? new BufferedInputStream(getContext().getContentResolver().openInputStream(uri))
                    : new BufferedInputStream(new FileInputStream(uri.getPath()));

            //--- Copy content from a URI to the temporary file and calculate MD5 ---
            MessageDigest digester = MessageDigest.getInstance("MD5");
            byte[] transBuff = new byte[2048];
            int nBytes;
            while ((nBytes = iStream.read(transBuff)) > 0) {
                digester.update(transBuff, 0, nBytes);
                tmpOs.write(transBuff, 0, nBytes);
            }
            tmpOs.flush();


            //---  Check and resize image in necessary ---
            File finalImage;
            BitmapFactory.Options opts = new BitmapFactory.Options();
            File downscaledFile = downscaleImportedImageFile(originalImage, opts);
            if (downscaledFile != null) {
                finalImage = FileUtils.getNewRandomFile(getAttachmentCacheDirectory(getContext()));
                digester = copyFileAndRecalculateHash(downscaledFile, finalImage);
            } else {
                finalImage = originalImage;
            }

            //--- Decode TMP file into a Bitmap and test ---
            String mimeType = opts.outMimeType;
            if (mimeType == null) {
                throw new Exception("Importing resource is probably not an image");
            }


            //--- Get MD5 Base64 and get final file name based on it ---
            String strMD5Base64 = Base64.encodeToString(digester.digest(), Base64.NO_WRAP | Base64.NO_PADDING | Base64.URL_SAFE);
            String relativeHashName = strMD5Base64 + "." + mimeType.substring(mimeType.lastIndexOf('/') + 1);


            setResult(finalImage, mimeType, 0, relativeHashName, AttachmentType.IMAGE);
            setCoverOptions(new File[]{originalImage});

        } finally {
            if (iStream != null) iStream.close();
            if (tmpOs != null) tmpOs.close();
        }

    }

    public static final String TAG = "ImageAttImporter";

    /**
     *
     * @param importedTmpFile

     * @return new File with downscaled image. null if an image was not downscaled
     */
    private File downscaleImportedImageFile(File importedTmpFile, BitmapFactory.Options opts) throws Exception {

        opts.inJustDecodeBounds = true;
        BitmapFactory.decodeFile(importedTmpFile.getAbsolutePath(), opts);
        int realPixelCount = opts.outWidth * opts.outHeight;
        if((maxPixelCount > 0) && (realPixelCount > maxPixelCount)) {
            // Downscaling bitmap



            Bitmap iBmp = AttachmentUtils.decodeBitmapMaxRezSafely(importedTmpFile.getAbsolutePath());
            final int availPixelCount = iBmp.getWidth() * iBmp.getHeight();

            Log.d(TAG, String.format("Orig pixels=%d, Aval pixels=%d, downscale = %.1f, MAX px count=%d", realPixelCount, availPixelCount, (float)realPixelCount/availPixelCount, maxPixelCount));


            //---  Scale-down bitmap if necessary  ---
            Bitmap oBmp;
            if (availPixelCount > maxPixelCount) {
                float scale = (float)Math.sqrt(( (double)maxPixelCount/availPixelCount ));
                Log.d(TAG, "Additional downscaling bitmap in memory to Pixel Count - "+Math.round(iBmp.getWidth()*iBmp.getHeight()*scale));
                oBmp = AttachmentUtils.scaleBitmapPreservingAspect(iBmp, scale);
                if (oBmp != iBmp) {
                    iBmp.recycle();
                }
                System.gc();
                Thread.sleep(50);
            } else {
                oBmp = iBmp;
            }

            //---  Fix rotation if necessary  ---
            Bitmap rotFixedBmp = AttachmentUtils.getBitmapInCorrectOrientation(oBmp, importedTmpFile.getAbsolutePath());
            if (rotFixedBmp != oBmp) {
                Log.d(TAG, "Recycle intermediate bitmap");
                oBmp.recycle();
            } else {
                Log.d(TAG, "No need to Recycle intermediate bitmap");
            }





            //---  Compressing prepared bitmap to a new tmp file  ---
            File f1 = FileUtils.getNewRandomFile(getAttachmentCacheDirectory(getContext()));
            OutputStream tmpOs = null;
            try {
                tmpOs = new BufferedOutputStream(new FileOutputStream(f1));
                rotFixedBmp.compress(Bitmap.CompressFormat.PNG, 0, tmpOs);
                rotFixedBmp.recycle();
                tmpOs.flush();
            } catch (IOException e) {
                throw new Exception("Error importing image - "+e.getMessage(), e);
            } finally {
                if (tmpOs != null) tmpOs.close();
            }
            return f1;
        } else {
            return null;
        }
    }

    private MessageDigest copyFileAndRecalculateHash(File srcFile, File dstFile) throws Exception {
        dstFile.delete();
        OutputStream oStream = null;
        InputStream iStream = null;
        try {
            oStream = new BufferedOutputStream(new FileOutputStream(dstFile.getAbsolutePath()));
            iStream = new BufferedInputStream(new FileInputStream(srcFile.getAbsolutePath()));

            MessageDigest digester = MessageDigest.getInstance("MD5");

            byte[] transBuff = new byte[1024];
            int nBytes;
            while ((nBytes = iStream.read(transBuff)) > 0) {
                digester.update(transBuff, 0, nBytes);
                oStream.write(transBuff, 0, nBytes);
            }
            oStream.flush();
            return digester;
        } catch (Exception e) {
            throw new Exception("Error importing image (calculate HASH) - "+e.getMessage(), e);
        } finally {
            if (oStream != null) oStream.close();
            if (iStream != null) iStream.close();
        }
    }
}

