package com.webssa.library.attachment.presenter;

import android.Manifest;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.graphics.Rect;
import android.media.MediaScannerConnection;
import android.net.Uri;
import android.provider.MediaStore;
import android.widget.Toast;

import com.webssa.library.attachment.ImportAttachment;
import com.webssa.library.attachment.OnImportResultListener;
import com.webssa.library.attachment.model.AttachmentImportModel;
import com.webssa.library.attachment.model.AttachmentType;
import com.webssa.library.attachment.model.CoverImportModel;
import com.webssa.library.attachment.util.AttachmentUtils;
import com.webssa.library.attachment.util.FileUtils;
import com.webssa.library.attachment.util.PrepareCoverImageTask;
import com.webssa.library.core.R;
import com.webssa.library.core.error.AppError;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import androidx.annotation.NonNull;
import androidx.core.app.ActivityCompat;

public class AttachAvatarPresenter extends BasePresenter<AttachAvatarView> {

    private static final int REQUEST_TAKE_AVATAR = 101;
    private static final int REQUEST_PICK_AVATAR = 102;
    private static final int REQUEST_PERMISSIONS_PICTURE_FROM_CAMERA = 103;
    private static final int REQUEST_PERMISSIONS_PICTURE_FROM_GALLERY = 104;

    private boolean liveAttachment;

    public AttachAvatarPresenter(AttachAvatarView attachAvatarView, boolean liveAttachment) {
        super(attachAvatarView);
        this.liveAttachment = liveAttachment;
    }

    @Override
    public void initializeView() {
        boolean started = (liveAttachment) ? tryStartCameraForImage() :  tryStartGallery();

        if (!started) {
            showErrorSafely(R.string.err_import_image);
            getView().close();
        }
    }

    public boolean onRequestPermissionsResult(int requestCode, int[] grantResults) {
        switch (requestCode) {
            case REQUEST_PERMISSIONS_PICTURE_FROM_CAMERA:
                if(reTryStartCameraWithPermissions) {
                    reTryStartCameraWithPermissions = false;
                    for(int result : grantResults) {
                        if(result != PackageManager.PERMISSION_GRANTED) {
                            showErrorSafely(R.string.err_permission_not_granted);
                            closeViewSafely();
                            return true;
                        }
                    }
                    tryStartCameraForImage();
                }
                return true;
            case REQUEST_PERMISSIONS_PICTURE_FROM_GALLERY:
                if (reTryStartGalleryWithPermissions) {
                    reTryStartGalleryWithPermissions = false;
                    for (int result : grantResults) {
                        if(result != PackageManager.PERMISSION_GRANTED) {
                            showErrorSafely(R.string.err_permission_not_granted);
                            closeViewSafely();
                            return true;
                        }
                    }
                    tryStartGallery();
                }
                return true;
            default:
                return false;
        }

    }

    /**
     * This method is to process calling children activities result.
     * It must be calles from the Activity's onActivityReenter
     * @param requestCode
     * @param resultCode
     * @param data
     * @return true if result was consumed (no need to call super Activity method)
     */
    public boolean onActivityResult(int requestCode, int resultCode, Intent data) {
        Uri[] selectedItems;
        switch (requestCode) {
            case REQUEST_TAKE_AVATAR:
                selectedItems = AttachmentUtils.getUrisFromResult(takePictureIntent);
                if (resultCode == Activity.RESULT_OK && selectedItems.length > 0) {
                    importImageAsync(selectedItems[0]);
                } else {
                    showErrorSafely(R.string.err_image_not_taken);
                    closeViewSafely();
                }
                return true;
            case REQUEST_PICK_AVATAR:
                selectedItems = AttachmentUtils.getUrisFromResult(data);
                if (resultCode == Activity.RESULT_OK && selectedItems.length > 0) {
                    importImageAsync(selectedItems[0]);
                } else {
                    showErrorSafely(R.string.err_image_not_picked);
                    closeViewSafely();
                }
                return true;
        }
        return false;
    }

    private void importImageAsync(Uri src) {
        getView().showProgressWithMessage(getView().getContext().getString(R.string.progress_import));
        ImportAttachment
                .from(src)
                .setContext(getView().getContext())
                .typeOff(AttachmentType.IMAGE)
                .downscaleToMaxPixelCount(0)
                .startImport(new OnImportResultListener() {
                    @Override
                    public void onImport(@NonNull AttachmentImportModel localAttachment, File[] coverImageOptions) {
                        dismissProgressSafely();
                        setImportedResult(coverImageOptions[0]);
                        notifyViewImageImportedSafely();
                    }

                    @Override
                    public void onError(@NonNull AppError error) {
                        dismissProgressSafely();
                        showErrorSafely(R.string.err_import_image);
                        closeViewSafely();
                    }
                });
    }


    /**
     * This field is to be used in the onActivityResult()
     * As the camera activity does not return Uri.
     */
    private Intent takePictureIntent;

    private boolean reTryStartCameraWithPermissions;

    private boolean tryStartCameraForImage() {
        Context ctx = getView().getContext();
        reTryStartCameraWithPermissions = false;
        List<String> notGrantedPermissions = new ArrayList<>(3);
        if(ActivityCompat.checkSelfPermission(ctx, Manifest.permission.READ_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            notGrantedPermissions.add(Manifest.permission.READ_EXTERNAL_STORAGE);
        }
        if(ActivityCompat.checkSelfPermission(ctx, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            notGrantedPermissions.add(Manifest.permission.WRITE_EXTERNAL_STORAGE);
        }
        if(ActivityCompat.checkSelfPermission(ctx, Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED) {
            notGrantedPermissions.add(Manifest.permission.CAMERA);
        }
        if(notGrantedPermissions.size() > 0) {
            reTryStartCameraWithPermissions = true;
            getView().requestPermissionsForPresenter(notGrantedPermissions.toArray(new String[notGrantedPermissions.size()]), REQUEST_PERMISSIONS_PICTURE_FROM_CAMERA);
            return true;
        }


        takePictureIntent = new Intent(MediaStore.ACTION_IMAGE_CAPTURE);
        if (takePictureIntent.resolveActivity(ctx.getPackageManager()) != null) {
            try {
                FileUtils.createImageFileForCamera(getView().getContext(), (MediaScannerConnection.OnScanCompletedListener) (s, uri) -> {
                    takePictureIntent.putExtra(MediaStore.EXTRA_OUTPUT, uri);
                    getView().getActivity().startActivityForResult(takePictureIntent, REQUEST_TAKE_AVATAR);
                });
                return true;
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        return false;
    }


    private boolean reTryStartGalleryWithPermissions;

    private boolean tryStartGallery() {
        Context ctx = getView().getContext();
        reTryStartGalleryWithPermissions = false;
        List<String> notGrantedPermissions = new ArrayList<>(2);
        if (ActivityCompat.checkSelfPermission(ctx, Manifest.permission.READ_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            notGrantedPermissions.add(Manifest.permission.READ_EXTERNAL_STORAGE);
        }
        if (ActivityCompat.checkSelfPermission(ctx, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            notGrantedPermissions.add(Manifest.permission.WRITE_EXTERNAL_STORAGE);
        }

        if (notGrantedPermissions.size() > 0) {
            reTryStartGalleryWithPermissions = true;
            getView().requestPermissionsForPresenter(notGrantedPermissions.toArray(new String[notGrantedPermissions.size()]), REQUEST_PERMISSIONS_PICTURE_FROM_GALLERY);
            return true;
        }

        Intent intent = AttachmentUtils.getIntentPickImagesFromGallery(getView().getContext(), false);
        if (intent != null) {
            ((Activity) getView().getContext()).startActivityForResult(Intent.createChooser(intent, "Select pictures"), REQUEST_PICK_AVATAR);
            return true;
        }
        return false;
    }

    private void dismissProgressSafely() {
        AttachAvatarView v = getView();
        if (!isReleased() && v != null) {
            v.dismissProgressWithMessage();
        }
    }

    private void showErrorSafely(int resId) {
        AttachAvatarView v = getView();
        if (!isReleased() && v != null) {
            Toast.makeText(v.getContext(), v.getContext().getString(resId), Toast.LENGTH_SHORT).show();
        }
    }

    private void closeViewSafely() {
        AttachAvatarView v = getView();
        if (!isReleased() && v != null) {
            v.close();
        }
    }

    private void finishWithResultSafely() {
        AttachAvatarView v = getView();
        if (!isReleased() && v != null) {
            if (coverImageOption != null) coverImageOption.delete();
            v.finishWithResult(resultCoverImage);
        }
    }

    private void notifyViewImageImportedSafely() {
        AttachAvatarView v = getView();
        if (!isReleased() && v != null) {
            v.onAttachmentImported(coverImageOption);
        }
    }


    public void prepareCoverImageAndFinish(File coverFileToCrop, Rect selectedRegion) {
        new PrepareCoverImageTask(getView().getContext(), coverFileToCrop, selectedRegion, 3_500_000) {
            @Override
            protected void onPreExecute() {
                getView().showProgressWithMessage("Preparing...");
            }

            @Override
            protected void onPostExecute(CoverImportModel coverImportModel) {
                resultCoverImage = coverImportModel;
                dismissProgressSafely();
                finishWithResultSafely();
            }
        }.safeExecute();
    }


    public void clearAndRecycle() {
        reTryStartCameraWithPermissions = false;
        if (resultCoverImage != null) {
            new File(resultCoverImage.localFile()).delete();
            resultCoverImage = null;
        }
        if (coverImageOption != null) coverImageOption.delete();
    }

    private CoverImportModel resultCoverImage;
    private File coverImageOption;

    private void setImportedResult(File coverImageOption) {
        this.coverImageOption = coverImageOption;
    }


}
