package com.webssa.library.attachment.util;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.graphics.Rect;

import com.webssa.library.attachment.model.CoverImportModel;
import com.webssa.library.security.HashCalculator;
import com.webssa.library.util.AsyncTaskCompat;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import androidx.annotation.NonNull;

/**
 * Cropping parameters which must be passed to the task:
 * [0] - distance from the Left edge, relative to the width
 * [1] - distance from the Top edge, relative to the height
 * [2] - distance from the Right edge, relative to the width
 * [3] - distance from the Bottom edge, relative to the height
 * Created by stanislav.perchenko on 11/8/2016.
 */

public class PrepareCoverImageTask extends AsyncTaskCompat<Void, Void, CoverImportModel> {

    private Context context;
    private final File imgFile;
    private int maxPixelNumber;
    private Rect region;


    public PrepareCoverImageTask(Context context, @NonNull File f, Rect region, int maxPixelNumber) {
        this.context = context;
        this.imgFile = f;
        this.maxPixelNumber = maxPixelNumber;
        this.region = region;
    }

    @Override
    protected CoverImportModel doInBackground(Void... params) {

        BitmapFactory.Options opts = AttachmentUtils.decodeBitmapForOptions(imgFile.getAbsolutePath());
        final int origW = opts.outWidth;
        final int origH = opts.outHeight;



        if (region == null) {
            region = new Rect(0, 0, origW, origH);
        }
        if (region.left < 0) {
            region.right -= region.left;
            region.left = 0;
            if (region.right > origW) region.right = origW;
        }
        if (region.top < 0) {
            region.bottom -= region.top;
            region.top = 0;
            if (region.bottom > origH) region.right = origH;
        }
        if (region.right > origW) {
            region.left -= (region.right - origW);
            region.right = origW;
            if (region.left < 0) region.left = 0;
        }
        if (region.bottom > origH) {
            region.top -= (region.bottom - origH);
            region.bottom = origH;
            if (region.top < 0) region.top = 0;
        }

        //--- Decode bitmap from file ---
        Bitmap decodedBmp = AttachmentUtils.decodeBitmapMaxRezSafely(imgFile.getAbsolutePath());

        //--- Check if decoded image was downscaled ---
        if ((origW != decodedBmp.getWidth()) || (origH != decodedBmp.getHeight())) {
            float preDscl = Math.min(((float) decodedBmp.getWidth()/origW), ((float) decodedBmp.getHeight() / origH));
            region.left = (int) (region.left * preDscl + 0.5f);
            region.top = (int) (region.top * preDscl + 0.5f);
            region.right = (int) (region.right * preDscl + 0.5f);
            region.bottom = (int) (region.bottom * preDscl + 0.5f);
        }




        //---  Define downscale matrix if necessary ---
        Matrix dsMatrix = null;
        double croppedPixels = region.width()*region.height();
        if (croppedPixels/maxPixelNumber > 1.1) {
            float scale = (float)Math.sqrt(( (double)maxPixelNumber/croppedPixels ));
            dsMatrix = new Matrix();
            dsMatrix.preScale(scale, scale);
            /*region.left = (int) (region.left * scale + 0.5f);
            region.top = (int) (region.top * scale + 0.5f);
            region.right = (int) (region.right * scale + 0.5f);
            region.bottom = (int) (region.bottom * scale + 0.5f);*/
        }


        Bitmap finBmp = null;
        try {
            finBmp = Bitmap.createBitmap(decodedBmp, region.left, region.top, region.width(), region.height(), dsMatrix, (dsMatrix != null));
            if (finBmp != decodedBmp) decodedBmp.recycle();

            //---- Compress image to byte array in memory  ----
            byte[] compressedBytes;
            ByteArrayOutputStream bos = null;
            try {
                bos = new ByteArrayOutputStream(finBmp.getByteCount() / 2);
                finBmp.compress(Bitmap.CompressFormat.PNG, 0, bos);
                compressedBytes = bos.toByteArray();
            } finally {
                if (bos != null) bos.close();
            }

            //----  Calculate hash  ----
            String hashName;
            String destFilePath;
            InputStream is = null;
            try {
                is = new ByteArrayInputStream(compressedBytes);
                hashName = HashCalculator.createForSHA1().process(is).getBase64Hash() + ".png";
            } finally {
                if (is != null) is.close();
            }

            //----  Save compressed data to file  ----
            destFilePath = saveCompressedData(hashName, compressedBytes);

            //----  Decode save file back for checking and get MIME type  ---
            String mimeType = AttachmentUtils.decodeImageFileForMimeType(destFilePath);

            //----  Instantiate result model and return  ----
            return CoverImportModel.create(destFilePath, mimeType, hashName, finBmp.getWidth(), finBmp.getHeight());
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        } finally {
            if (finBmp != null) finBmp.recycle();
        }








        //----  Crop image  ----
       /* Bitmap croppedBitmap = Bitmap.createBitmap(origBitmap, region.left, region.top, region.width(), region.height());
        if (origBitmap != croppedBitmap) origBitmap.recycle();
        try {

            //---- Downscale image  ----
            int numPixels = croppedBitmap.getWidth()*croppedBitmap.getHeight();
            if ((maxPixelNumber > 0) && (numPixels > maxPixelNumber)) {

                Bitmap tmpDownscaled = AttachmentUtils.scaleBitmapPreservingAspect(croppedBitmap, scale);
                if (tmpDownscaled != croppedBitmap) {
                    croppedBitmap.recycle();
                    croppedBitmap = tmpDownscaled;
                }
            }

            finalWidth = croppedBitmap.getWidth();
            finalHeight= croppedBitmap.getHeight();







        } catch (Exception e) {
            Crashlytics.logException(e);
            return null;
        } finally {
            croppedBitmap.recycle();
        }*/
    }


    /**
     * Saves array of bytes to file and returns the absolute path of the created files
     * @param relativeFileName
     * @param data
     * @return
     */
    private String saveCompressedData(String relativeFileName, byte[] data) throws IOException {
        InputStream is = null;
        try {
            is = new ByteArrayInputStream(data);
            File fDest = new File(FileUtils.getFinalCacheDirectory(context, AttachmentUtils.DIRECTORY_ATTACHMENTS), relativeFileName);
            if (fDest.exists()) fDest.delete();
            FileUtils.streamToFile(fDest, is);
            return fDest.getAbsolutePath();
        } finally {
            if (is != null) is.close();
        }
    }

}