﻿using Microsoft.EntityFrameworkCore;
using PDA.Database.CustomExceptions;
using PDA.Database.EntityModels;
using PDA.Database.Interfaces;
using PDA.Database.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PDA.Database.Repositories
{
    public class ShiftRepository : IShiftRepository
    {
        private readonly PDADbContext _context;
        public ShiftRepository(PDADbContext context)
        {
            _context = context;
        }

        public async Task<Shift> AcceptShift(int shiftId)
        {
            var shift = await _context.Shifts.FirstOrDefaultAsync(x => x.ShiftId == shiftId);
            if (shift == null)
                throw new ShiftNotFoundException();
            if (shift.StateId != (int)ShiftStateEnum.New)
                throw new ShiftNotAcceptableException();
            shift.StateId = (int)ShiftStateEnum.Accepted;
            shift.ActualTimeStart = DateTimeOffset.UtcNow;
            await _context.SaveChangesAsync();
            return await _context.Shifts.Include(x => x.Visits).ThenInclude(v => v.Site).FirstOrDefaultAsync(x => x.ShiftId == shiftId);
        }

        public async Task<Shift> CloseShift(int shiftId, DateTimeOffset? closeTime, string userId)
        {
           var shift = await _context.Shifts.FirstOrDefaultAsync(x => x.ShiftId == shiftId && x.UserId == userId);
            if (shift == null)
                throw new ShiftNotFoundException();
            if (shift.StateId != (int)ShiftStateEnum.Accepted)
                throw new ShiftNotClosableException();
            shift.StateId = (int)ShiftStateEnum.Closed;
            shift.ActualTimeClosed = closeTime == null ? DateTimeOffset.UtcNow : closeTime;
            await _context.SaveChangesAsync();
            return await _context.Shifts.FirstOrDefaultAsync(x => x.ShiftId == shiftId && x.UserId == userId);
        }
        public async Task CloseShift(Shift shift, DateTimeOffset? closeTime, bool completedVisit)
        {
            shift.NumVisitsCompleted = shift.NumVisitsCompleted.HasValue ? shift.NumVisitsCompleted : 0;
            shift.NumVisitsFailed = shift.NumVisitsFailed.HasValue ? shift.NumVisitsFailed : 0;
            if (completedVisit)
                shift.NumVisitsCompleted++;
            else
                shift.NumVisitsFailed++;
            if (shift.NumVisits == shift.NumVisitsCompleted + shift.NumVisitsFailed)
            {
                shift.StateId = (int)ShiftStateEnum.Closed;
                shift.ActualTimeClosed = closeTime == null ? DateTimeOffset.UtcNow : closeTime;
            }
            await _context.SaveChangesAsync();
        }
        public async Task<Shift> FindNewShiftForDate(string userId)
        {
            var shifts =  _context.Shifts.Where(x => x.UserId == userId && x.ExecutionDate.Date == DateTime.UtcNow.Date);
           var shift =  await shifts.FirstOrDefaultAsync(x => x.StateId == (int)ShiftStateEnum.New);
            if (shift == null)
                throw new ShiftNotFoundException();
            if (shift.UserId != userId)
                throw new ShiftNotUsersException();
            return shift;
        }

        public async Task<Shift> GetByShiftId(int shiftId, string userId)
        {
            var shift = await _context.Shifts.FirstOrDefaultAsync(x => x.ShiftId == shiftId);
            if (shift == null)
                throw new ShiftNotFoundException();
            if (shift.UserId != userId)
                throw new ShiftNotUsersException();
            return shift;
        }

        public async Task<Shift> RejectShift(int shiftId, string userId)
        {
            var shift = await _context.Shifts.FirstOrDefaultAsync(x => x.ShiftId == shiftId && x.UserId == userId);
            if (shift == null)
                throw new ShiftNotFoundException();
            if (shift.StateId != (int)ShiftStateEnum.New)
                throw new ShiftNotRejectableException();
            shift.StateId = (int)ShiftStateEnum.Closed;
            shift.ActualTimeClosed = DateTimeOffset.UtcNow;
            await _context.SaveChangesAsync();
            return await _context.Shifts.Include(x => x.Visits).ThenInclude(v => v.Site).FirstOrDefaultAsync(x => x.ShiftId == shiftId && x.UserId == userId);
        }
    }
}
