﻿using Microsoft.EntityFrameworkCore;
using PDA.Database.CustomExceptions;
using PDA.Database.EntityModels;
using PDA.Database.Interfaces;
using PDA.Database.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace PDA.Database.Repositories
{
    public class VisitRepository : IVisitRepository
    {
        private readonly PDADbContext _context;
        private readonly IShiftRepository _shiftRepository;
        private readonly INoteRepository _noteRepository;
        private readonly IGoodsRepository _goodsRepository;
        public VisitRepository(PDADbContext context, IShiftRepository shiftRepository, INoteRepository noteRepository, IGoodsRepository goodsRepository)
        {
            _context = context;
            _shiftRepository = shiftRepository;
            _noteRepository = noteRepository;
            _goodsRepository = goodsRepository;
        }

        public async Task<Visit> CompleteVisit(int visitId, DateTimeOffset? delayedCompleteTime, Note note,  string userId, 
            List<string> imageUrls, List<CollectedGood> goods,  string customerName, string customerSig, int clearBags, bool isWhiteSacks)
        {
            var visit = await _context.Visits.Include(x => x.CollectedGoods)
                .ThenInclude(x => x.Price)
                .Include(x => x.Note)
                .ThenInclude(x => x.NoteAttachments)
                .FirstOrDefaultAsync(x => x.VisitId == visitId);
            if (visit == null)
                throw new VisitNotFoundException();
            var parentShift = await _shiftRepository.GetByShiftId(visit.ShiftId, userId);
            if (visit.StateId != (int)VisitStateEnum.InProgress || parentShift.StateId != (int)ShiftStateEnum.Accepted)
                throw new VisitNotFailableException();
            visit.StateId = (int)VisitStateEnum.Completed;
            visit.CompleteTime = delayedCompleteTime.HasValue ? delayedCompleteTime : DateTimeOffset.UtcNow;
            visit.CustomerName = customerName;
            visit.CustomerSignature = customerSig;
            visit.ClearBags = clearBags;
            visit.IsWhiteSacks = isWhiteSacks;
            if (visit.Note == null && note != null)
            {
                note.ShiftId = parentShift.ShiftId;
                note.VisitId = visitId;
                 await _noteRepository.SaveNoteFromVisit(note, userId);
                if (imageUrls.Count > 0)
                    for (var i = 0; i < imageUrls.Count; i++)
                        await _noteRepository.SaveAttahment(imageUrls[i], note.NoteId, i);
            }
            await _goodsRepository.SaveVisitGoods(goods, visitId);
            await _shiftRepository.CloseShift(parentShift, visit.CompleteTime, true);
            visit.CollectedGoods = await _goodsRepository.GetVisitGoodsAsync(visitId);
            CalculateVisitTotal(visit);
            await _context.SaveChangesAsync();
            return visit;
        }

        public async Task<Visit> FailVisit(int visitId, DateTimeOffset? delayedFailTime, string noteText, List<string> imageUrls, int emergencyId, string userId)
        {
            var visit = await _context.Visits.Include(x => x.Note).FirstOrDefaultAsync(x => x.VisitId == visitId);
            if (visit == null)
                throw new VisitNotFoundException();
            var parentShift = await _shiftRepository.GetByShiftId(visit.ShiftId, userId);
            if (visit.StateId != (int)VisitStateEnum.InProgress || parentShift.StateId != (int)ShiftStateEnum.Accepted)
                throw new VisitNotFailableException();
            visit.CompleteTime = delayedFailTime.HasValue ? delayedFailTime : DateTimeOffset.UtcNow;
            visit.StateId = (int)VisitStateEnum.Failed;
            if (visit.Note == null)
            {
                var note = new Note
                {
                    TimeCreated = delayedFailTime.HasValue ? delayedFailTime.Value : DateTimeOffset.UtcNow,
                    VisitId = visitId,
                    ShiftId = parentShift.ShiftId,
                    Text = noteText,
                    EmergencyId = emergencyId
                };
                await _noteRepository.SaveNoteFromVisit(note, userId);
                if (imageUrls.Count > 0)
                    for(var i = 0; i < imageUrls.Count; i++)
                        await _noteRepository.SaveAttahment(imageUrls[i], note.NoteId, i);
            }
            await _shiftRepository.CloseShift(parentShift, visit.CompleteTime, false);
            await _context.SaveChangesAsync();
            return visit;
        }

        public async Task<Visit> GetVisitById(int id) => await _context.Visits.FirstOrDefaultAsync(x => x.VisitId == id);
      
        public async Task<Visit> GetVisitState(int visitId, string userId)
        {
            var visit = await _context.Visits.Include(x => x.Shift).FirstOrDefaultAsync(x => x.VisitId == visitId && x.Shift.UserId == userId);
            if (visit == null)
                throw new VisitNotFoundException();
            return visit;
        }
        public async Task<Visit> StartVisit(int visitId, DateTimeOffset? delayedStartTime, string userId)
        {
            var visit = await _context.Visits.FirstOrDefaultAsync(x => x.VisitId == visitId);
            if (visit == null)
                throw new VisitNotFoundException();
            var parentShift = await _shiftRepository.GetByShiftId(visit.ShiftId, userId);
            if (visit.StateId != (int)VisitStateEnum.New || parentShift.StateId != (int)ShiftStateEnum.Accepted)
                throw new VisitNotStartableException();
            visit.StateId = (int)VisitStateEnum.InProgress;
            if (delayedStartTime.HasValue)
                visit.StartTime = delayedStartTime;
            else
                visit.StartTime = DateTimeOffset.UtcNow;
            await _context.SaveChangesAsync();
            return visit;

        }

        public async Task<Visit> AddVisit(int shiftId, int siteId, DateTime? suggestTime, string userId)
        {
            var parentShift = await _shiftRepository.GetByShiftId(shiftId, userId);

            var maxOrder = await _context.Visits.Where(v => v.ShiftId == shiftId).MaxAsync(v => v.Order);
            if (!suggestTime.HasValue)
            {
                suggestTime = DateTime.UtcNow;
                suggestTime = new DateTime(suggestTime.Value.Year, suggestTime.Value.Month, suggestTime.Value.Day,
                    suggestTime.Value.Hour, suggestTime.Value.Minute, 0, DateTimeKind.Utc);
            }

            var order = ++maxOrder;
            var visit = new Visit
            {
                ShiftId = shiftId,
                SiteId = siteId,
                SuggestTime = suggestTime.Value,
                StateId = (int) VisitStateEnum.New,
                Order = order
            };

            await _context.Visits.AddAsync(visit);
            parentShift.NumVisits++;
            await _context.SaveChangesAsync();

            var resultVisit =
                await _context.Visits.FirstAsync(v => v.ShiftId == shiftId && v.SiteId == siteId && v.Order == order);

            return resultVisit;
        }

        private void CalculateVisitTotal(Visit visit)
        {
            visit.TotalPayedScale = 1;
            visit.TotalPayedScale = visit.CollectedGoods.Max(x => x.Price.Scale);
            visit.TotalPayedAmount = 0;
            foreach(var good in visit.CollectedGoods)
            {
                var difference = visit.TotalPayedScale - good.Price.Scale;
                var scaled = 10;
                visit.TotalPayedAmount = difference > 0 ? (good.Price.Value * Convert.ToInt32(Math.Pow(scaled, (double)difference))) * good.Weight
                    + visit.TotalPayedAmount : good.Price.Value * good.Weight + visit.TotalPayedAmount;
            }
        }
    }
}
