﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Threading.Tasks;
using AutoMapper;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http.Features;
using Microsoft.AspNetCore.HttpsPolicy;
using Microsoft.AspNetCore.Identity;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using PDA.API.AppConfiguration;
using PDA.API.Filters;
using PDA.API.Mapping;
using PDA.BL.Interfaces;
using PDA.BL.Models;
using PDA.BL.Models.Interfaces;
using PDA.BL.Models.Models;
using PDA.BL.Services;
using PDA.Database.EntityModels;
using PDA.Database.Interfaces;
using PDA.Database.Models;
using PDA.Database.Repositories;
using Swashbuckle.AspNetCore.Swagger;

namespace PDA
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            services.AddMvc().SetCompatibilityVersion(CompatibilityVersion.Version_2_2);
            services.AddDbContext<PDADbContext>(options => options.UseSqlServer(Configuration.GetConnectionString("DefaultConnection"), b => b.MigrationsAssembly("PDA.API")));
            services.AddIdentity<User, IdentityRole>().AddEntityFrameworkStores<PDADbContext>();
            services.AddScoped<IAuthorizationConfiguration, AuthorizationConfiguration>();
            services.AddScoped<IDriverService, DriverService>();
            services.AddScoped<IDriverRepository, DriverRepository>();
            services.AddScoped<ITokenService, TokenService>();
            services.AddScoped<ITokenRepository, TokenRepository>();
            services.AddScoped<IShiftService, ShiftService>();
            services.AddScoped<IShiftRepository, ShiftRepository>();
            services.AddScoped<IVisitService, VisitService>();
            services.AddScoped<IVisitRepository, VisitRepository>();
            services.AddScoped<INoteService, NoteService>();
            services.AddScoped<INoteRepository, NoteRepository>();
            services.AddScoped<ISecurityConfiguration, SecurityConfiguration>();
            services.AddScoped<IGoodsService, GoodsService>();
            services.AddScoped<IGoodsRepository, GoodsRepository>();
            services.AddScoped<IEmergencyRepository, EmergencyRepository>();
            services.AddScoped<IEmergencyService, EmergencyService>();
            services.AddScoped<IUploadFileService, UploadFileService>();
            services.AddScoped<IAzureBlobStorageService, AzureBlobStorageService>();
            services.AddScoped<ISiteRepository, SiteRepository>();
            services.AddScoped<AuthorizationTokenAttribute>();
            services.Configure<AzureBlobStorageConfiguration>(Configuration.GetSection("AzureBlobStorageConfiguration"));
            services.Configure<MobileAppSettings>(Configuration.GetSection("MobileAppSettings"));
            services.AddSwaggerGen(c =>
            {
                c.SwaggerDoc("v1", new Info { Title = "PDA Api", Version = "v1" });
                c.OperationFilter<UserHeaderFilter>();
                c.OperationFilter<AuthorizationHeaderFilter>();
            });
            var mappingConfig = new MapperConfiguration(mc =>
            {
                mc.AddProfile(new MappingProfile());
            });
            services.Configure<FormOptions>(options =>
            {
                options.MultipartBodyLengthLimit = long.MaxValue;
            });
            IMapper mapper = mappingConfig.CreateMapper();
            services.AddSingleton(mapper);
            services.AddCors();
         
        }
       

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IHostingEnvironment env)
        {
            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }
            else
            {
                // The default HSTS value is 30 days. You may want to change this for production scenarios, see https://aka.ms/aspnetcore-hsts.
                app.UseHsts();
            }

            app.UseHttpsRedirection();
            app.UseSwagger();
            app.UseSwaggerUI(c =>
            {
                c.SwaggerEndpoint("/swagger/v1/swagger.json", "PDA Api V1");
                c.RoutePrefix = string.Empty;
            });
            app.UseCors(builder => builder
              .AllowAnyOrigin()
              .WithMethods("GET", "HEAD", "POST", "DEBUG", "PUT", "DELETE", "PATCH", "OPTIONS"));
          
            app.UseMvc(routes =>
            {
                routes.MapRoute(
                    name: "default",
                    template: "{controller}/{action=Index}/{id?}");
            });
        }
    }
}
